@@ -1,4 +1,4 @@
-// Copyright (c) 1997, 1998, 1999, 2001, 2002, 2004, 2005, 2008, 2009  Per M.A. Bothner.
+// Copyright (c) 1997, 1998, 1999, 2001, 2002, 2004, 2005, 2008, 2009, 2011  Per M.A. Bothner.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.bytecode;
@@ -128,24 +128,18 @@ public class ClassType extends ObjectType
       }
     /* #endif */
     String name = getName();
+    if (enclosingMember instanceof ClassType)
+      {
+        String enclosingName = ((ClassType) enclosingMember).getName();
+        int enclosingLength;
+        if (enclosingName != null && name.startsWith(enclosingName)
+            && name.length() > (enclosingLength = enclosingName.length()) + 1
+            && name.charAt(enclosingLength) == '$')
+          return name.substring(enclosingLength+1);
+      }
     int dot = name.lastIndexOf('.');
     if (dot > 0)
       name = name.substring(dot+1);
-    int dollar = name.lastIndexOf('$');
-    if (dollar >= 0)
-      {
-        int len = name.length();
-        int start = dollar + 1;
-        while (start < len)
-          {
-            char ch = name.charAt(start);
-            if (ch >= '0' && ch <= '9')
-              start++;
-            else
-              break;
-          }
-        name = name.substring(start);
-      }
     return name;
   }
 
@@ -389,6 +383,24 @@ public class ClassType extends ObjectType
     SourceFileAttr.setSourceFile(this, name);
   }
 
+    TypeVariable[] typeParameters;
+
+    public TypeVariable[] getTypeParameters() {
+	TypeVariable[] params = typeParameters;
+	if (params == null && (flags & EXISTING_CLASS) != 0
+	    && getReflectClass() != null) {
+	    java.lang.reflect.TypeVariable[] rparams
+		= reflectClass.getTypeParameters();
+	    int nparams = rparams.length;
+	    params = new TypeVariable[nparams];
+	    for (int i = 0;  i < nparams;  i++) {
+		params[i] = TypeVariable.make(rparams[i]);
+	    }
+	    typeParameters = params;
+	}
+	return params;
+    }
+
   /**
    * Set the superclass of the is class.
    * @param name name of super class, or null if this is "Object".
@@ -476,6 +488,12 @@ public class ClassType extends ObjectType
     else access_flags &= ~Access.INTERFACE|Access.ABSTRACT;
   }
 
+  public final boolean isFinal ()
+  { return (getModifiers() & Access.FINAL) != 0; }
+
+  public final boolean isAnnotation ()
+  { return (getModifiers() & Access.ANNOTATION) != 0; }
+
   public ClassType () { }
 
   public ClassType (String class_name)
@@ -615,7 +633,7 @@ public class ClassType extends ObjectType
         java.lang.reflect.Field field = fields[i];
         if ("this$0".equals(field.getName()))
           flags |= HAS_OUTER_LINK;
-        addField(field.getName(), Type.make(field.getType()),
+        addField(field.getName(), Type.make(field.getType(), field.getGenericType()),
                  field.getModifiers());
       }
     flags |= ADD_FIELDS_DONE;
@@ -679,11 +697,12 @@ public class ClassType extends ObjectType
   {
     int modifiers = method.getModifiers();
     Class[] paramTypes = method.getParameterTypes();
+    java.lang.reflect.Type[] gparamTypes = method.getGenericParameterTypes();
     int j = paramTypes.length;
     Type[] args = new Type[j];
     while (--j >= 0)
-      args[j] = Type.make(paramTypes[j]);
-    Type rtype = Type.make(method.getReturnType());
+	args[j] = Type.make(paramTypes[j], gparamTypes[j]);
+    Type rtype = Type.make(method.getReturnType(), method.getGenericReturnType());
     return addMethod(method.getName(), modifiers, args, rtype);
   }
 
@@ -907,14 +926,15 @@ public class ClassType extends ObjectType
     return null;
   }
 
-  /** Get a method with matching name and number of arguments. */
-  public synchronized Method getDeclaredMethod(String name, int argCount)
+  synchronized Method getDeclaredMethod(String name, boolean mustBeStatic, int argCount)
   {
     Method result = null;
     int needOuterLinkArg = "<init>".equals(name) && hasOuterLink() ? 1 : 0;
     for (Method method = getDeclaredMethods();
 	 method != null;  method = method.next)
       {
+        if (mustBeStatic && ! method.getStaticFlag())
+          continue;
 	if (name.equals(method.getName())
 	    && argCount + needOuterLinkArg == method.getParameterTypes().length)
 	  {
@@ -927,7 +947,22 @@ public class ClassType extends ObjectType
       }
     return result;
   }
+  /** Get a method with matching name and number of arguments. */
+  public Method getDeclaredMethod(String name, int argCount)
+  {
+    return getDeclaredMethod(name, false, argCount);
+  }
 
+  /** Get a static method with matching name and number of arguments. */
+  public Method getDeclaredStaticMethod(String name, int argCount)
+  {
+    return getDeclaredMethod(name, true, argCount);
+  }
+
+  /** Looks for a method matching the name and types.
+   * Note looks for an exact match, unless a type is null,
+   * not necessarily the best match.
+   */
   public synchronized Method getMethod(String name, Type[] arg_types)
   {
     ClassType cl = this;
@@ -961,6 +996,11 @@ public class ClassType extends ObjectType
     return null;
   }
 
+  public Method getDefaultConstructor ()
+  {
+    return getDeclaredMethod("<init>", Type.typeArray0);
+  }
+
   /** Use reflection to add all the declared methods of this class.
    * Does not add constructors nor private or package-private methods.
    * Does not check for duplicate (already-known) methods.
@@ -1267,6 +1307,8 @@ public class ClassType extends ObjectType
     if ((this == toStringType && other == javalangStringType)
 	|| (this == javalangStringType && other == toStringType))
       return true;
+    if (other == Type.javalangObjectType)
+      return true;
     ClassType baseClass = this;
     while (baseClass != null)
       {
@@ -1296,9 +1338,11 @@ public class ClassType extends ObjectType
     if (cother == toStringType)
       return this == Type.javalangObjectType ? 1 : -1;
     if (this.isInterface())
-      return cother == Type.javalangObjectType ? -1 : -2;
+      return cother.isAnnotation() || cother.isFinal() ? -3
+        : cother == Type.javalangObjectType ? -1 : -2;
     if (cother.isInterface())
-      return this == Type.javalangObjectType ? 1 : -2;
+      return isAnnotation() || isFinal() ? -3
+        : this == Type.javalangObjectType ? 1 : -2;
     return -3;
   }
 
