@@ -1362,6 +1362,8 @@ public class CodeAttr extends Attribute implements AttrContainer
 
   public void emitStore (Variable var)
   {
+    if (! reachableHere())
+      return;
     int offset = var.offset;
     if (offset < 0 || !var.isSimple ())
       throw new Error ("attempting to store in unassigned "+var
@@ -1378,6 +1380,10 @@ public class CodeAttr extends Attribute implements AttrContainer
   }
 
 
+  /** Emit an instruction to increment a variable by some amount.
+   * If the increment is zero, do nothing.
+   * The variable must contain an integral value - except if increment is zero.
+   */
   public void emitInc (Variable var, short inc)
   {
     if (var.dead ())
@@ -1386,12 +1392,15 @@ public class CodeAttr extends Attribute implements AttrContainer
     if (offset < 0 || !var.isSimple ())
       throw new Error ("attempting to increment unassigned variable"+var.getName()
 		       +" simple:"+var.isSimple()+", offset: "+offset);
+
+    if (inc == 0)
+      return;
+
     reserve(6);
     if (var.getType().getImplementationType().promote() != Type.intType)
       throw new Error("attempting to increment non-int variable");
 
     boolean wide = offset > 255 || inc > 255 || inc < -256;
-
     if (wide)
     {
       put1(196); // wide
@@ -1467,6 +1476,8 @@ public class CodeAttr extends Attribute implements AttrContainer
 
   public void emitInvokeMethod (Method method, int opcode)
   {
+    if (! reachableHere())
+      return;
     reserve(opcode == 185 ? 5 : 3);
     int arg_count = method.arg_types.length;
     boolean is_invokestatic = opcode == 184;
@@ -1883,7 +1894,6 @@ public class CodeAttr extends Attribute implements AttrContainer
   /** Compile end of conditional. */
   public final void emitFi ()
   {
-    boolean make_unreachable = false;
     if (! if_stack.doing_else)
       { // There was no 'else' clause.
 	if (reachableHere ()
@@ -1906,14 +1916,11 @@ public class CodeAttr extends Attribute implements AttrContainer
 	  throw new Error("at PC "+PC+": SP at end of 'then' was " +
 			  then_clause_stack_size
 			  + " while SP at end of 'else' was " + SP);
+        setReachable(true);
       }
-    else if (unreachable_here)
-      make_unreachable = true;
 
     if (if_stack.end_label != null)
       if_stack.end_label.define (this);
-    if (make_unreachable)
-      setUnreachable();
     // Pop the if_stack.
     if_stack = if_stack.previous;
   }
@@ -2002,7 +2009,7 @@ public class CodeAttr extends Attribute implements AttrContainer
   public static boolean castNeeded (Type top, Type required)
   {
     if (top instanceof UninitializedType)
-      top = ((UninitializedType) top).getImplementationType();
+      top = ((UninitializedType) top).getRawType();
     for (;;)
       {
         if (top == required)
@@ -2074,6 +2081,8 @@ public class CodeAttr extends Attribute implements AttrContainer
 
   final void emitRawReturn ()
   {
+    if (! reachableHere())
+      return;
     if (getMethod().getReturnType().size == 0)
       {
 	reserve(1);
@@ -2261,6 +2270,7 @@ public class CodeAttr extends Attribute implements AttrContainer
     ClassType type = var == null ? null : (ClassType) var.getType();
     try_stack.try_type = type;
     addHandler(try_stack.start_try, try_stack.end_try, type);
+    setReachable(true);
     if (var != null)
       emitStore(var);
   }
