@@ -2,7 +2,6 @@
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.bytecode;
-import java.io.*;
 
 /** An abstracted "variable", inherited by Field and Variable. */
 
@@ -45,7 +44,7 @@ public class Location {
     this.type = type;
   }
 
-  public final String getSignature () { return type.getSignature (); }
+    public final String getSignature () { return type.getRawType().getSignature (); }
 
   public void setSignature (int signature_index, ConstantPool constants)
   {
