@@ -30,6 +30,7 @@ public class ApplyExp extends Expression
   public final Expression[] getArgs() { return args; }
   public final int getArgCount() { return args.length; }
   public void setFunction(Expression func) { this.func = func; }
+  public void setFunction(Procedure proc) { this.func = new QuoteExp(proc); }
   public void setArgs(Expression[] args) { this.args = args; }
   public Expression getArg(int i) { return args[i]; }
   public void setArg(int i, Expression arg) { args[i] = arg; }
@@ -37,20 +38,32 @@ public class ApplyExp extends Expression
   public final void setTailCall(boolean tailCall)
   { setFlag(tailCall, TAILCALL); }
 
+  public ApplyExp setFuncArgs (Expression func, Expression[] args)
+  {
+    setFunction(func);
+    setArgs(args);
+    setFlag(false, Expression.VALIDATED);
+    return this;
+  }
+
+  public ApplyExp setFuncArgs (Procedure proc, Expression[] args)
+  {
+    return setFuncArgs(new QuoteExp(proc), args);
+  }
+
   /** If getFunction() is constant, return its value; otherwise null. */
   public final Object getFunctionValue()
   {
     return func instanceof QuoteExp ? ((QuoteExp) func).getValue() : null;
   }
 
-  public ApplyExp (Expression f, Expression[] a) { func = f; args = a; }
+  public ApplyExp (Expression f, Expression... a) { func = f; args = a; }
 
-  public ApplyExp (Procedure p, Expression[] a) { func = new QuoteExp(p); args = a; }
+  public ApplyExp (Procedure p, Expression... a) { this(new QuoteExp(p), a); }
 
-  public ApplyExp (Method m, Expression[] a)
+  public ApplyExp (Method m, Expression... a)
   {
-    func = new QuoteExp(new PrimProcedure(m));
-    args = a;
+    this(new QuoteExp(new PrimProcedure(m)), a);
   }
 
   protected boolean mustCompile () { return false; }
@@ -65,17 +78,19 @@ public class ApplyExp extends Expression
     ((Procedure) proc).checkN(vals, ctx);
   }
 
-  public static void compileToArray(Expression[] args, Compilation comp)
+  private static void compileToArray(Expression[] args, int start, Compilation comp)
   {
     CodeAttr code = comp.getCode();
-    if (args.length == 0)
+    int argslength = args.length;
+    int nargs = argslength - start;
+    if (nargs == 0)
       {
 	code.emitGetStatic(Compilation.noArgsField);
 	return;
       }
-    code.emitPushInt(args.length);
+    code.emitPushInt(nargs);
     code.emitNewArray(Type.pointer_type);
-    for (int i = 0; i < args.length; ++i)
+    for (int i = start; i < argslength; ++i)
       {
 	Expression arg = args[i];
 	if (comp.usingCPStyle()
@@ -137,10 +152,11 @@ public class ApplyExp extends Expression
         ReferenceExp func_ref = (ReferenceExp) exp_func;
         owner = func_ref.contextDecl();
         Declaration func_decl = func_ref.binding;
+        Expression dvalue;
         while (func_decl != null && func_decl.isAlias()
-               && func_decl.value instanceof ReferenceExp)
+               && (dvalue = func_decl.getValueRaw()) instanceof ReferenceExp)
           {
-            func_ref = (ReferenceExp) func_decl.value;
+            func_ref = (ReferenceExp) dvalue;
             if (owner != null || func_decl.needsContext() || func_ref.binding == null)
               break;
             func_decl = func_ref.binding;
@@ -184,6 +200,13 @@ public class ApplyExp extends Expression
     gnu.bytecode.CodeAttr code = comp.getCode();
     Method method;
 
+    // Check for tail-recursion.
+    boolean tail_recurse
+      = exp.isTailCall()
+      && func_lambda != null && func_lambda == comp.curLambda
+      // No keyword or optional arguments.
+      && func_lambda.opt_args == 0 && func_lambda.keywords == null;
+
     if (func_lambda != null)
       {
 	if ((func_lambda.max_args >= 0 && args_length > func_lambda.max_args)
@@ -192,9 +215,13 @@ public class ApplyExp extends Expression
 	  throw new Error ("internal error - wrong number of parameters for "
 			   + func_lambda);
 	int conv = func_lambda.getCallConvention();
+        if (func_lambda.primMethods==null && ! func_lambda.isClassGenerated()
+            && ! func_lambda.getInlineOnly())
+            func_lambda.allocMethod(func_lambda.outerLambda(), comp);
         // Mostly duplicates logic with LambdaExp.validateApply.
         // See comment there.
 	if (comp.inlineOk(func_lambda)
+            && ! tail_recurse
 	    && (conv <= Compilation.CALL_WITH_CONSUMER
 		|| (conv == Compilation.CALL_WITH_TAILCALLS
 		    && ! exp.isTailCall()))
@@ -280,16 +307,10 @@ public class ApplyExp extends Expression
 	return;
       }
     */
-
-    // Check for tail-recursion.
-    boolean tail_recurse
-      = exp.isTailCall()
-      && func_lambda != null && func_lambda == comp.curLambda;
-
     if (func_lambda != null && func_lambda.getInlineOnly() && !tail_recurse
 	&& func_lambda.min_args == args_length)
       {
-        pushArgs(func_lambda, exp.args, null, comp);
+        pushArgs(func_lambda, exp.args, exp.args.length, null, comp);
         if (func_lambda.getFlag(LambdaExp.METHODS_COMPILED))
           {
             popParams(code, func_lambda, null, false);
@@ -329,7 +350,7 @@ public class ApplyExp extends Expression
 	  }
 	else
 	  {
-	    compileToArray (exp.args, comp);
+            compileToArray (exp.args, 0, comp);
 	    comp.loadCallContext();
 	    code.emitInvoke(Compilation.typeProcedure
 			    .getDeclaredMethod("checkN", 2));
@@ -356,20 +377,23 @@ public class ApplyExp extends Expression
       exp_func.compile (comp, new StackTarget(Compilation.typeProcedure));
 
     boolean toArray
-      = (tail_recurse ? func_lambda.min_args != func_lambda.max_args
+      = (tail_recurse ? func_lambda.max_args < 0
          : args_length > 4);
     int[] incValues = null; // Increments if we can use iinc.
-    if (toArray)
+    if (tail_recurse)
       {
-	compileToArray(exp.args, comp);
-	method = Compilation.applyNmethod;
-      }
-    else if (tail_recurse)
-      {
-        incValues = new int[exp.args.length];
-        pushArgs(func_lambda, exp.args, incValues, comp);
+        int fixed = func_lambda.min_args;
+        incValues = new int[fixed];
+        pushArgs(func_lambda, exp.args, fixed, incValues, comp);
+        if (toArray)
+          compileToArray(exp.args, fixed, comp);
         method = null;
       }
+    else if (toArray)
+      {
+        compileToArray(exp.args, 0, comp);
+	method = Compilation.applyNmethod;
+      }
     else
       {
 	for (int i = 0; i < args_length; ++i)
@@ -382,7 +406,8 @@ public class ApplyExp extends Expression
       }
     if (! code.reachableHere())
       {
-        comp.error('e', "unreachable code");
+        if (comp.warnUnreachable())
+          comp.error('w', "unreachable code");
         return;
       }
     if (tail_recurse)
@@ -436,7 +461,7 @@ public class ApplyExp extends Expression
     out.writeSpaceFill();
     printLineColumn(out);
     func.print(out);
-    for (int i = 0; i < args.length; ++i)
+    for (int i = 0; args != null && i < args.length; ++i)
       {
 	out.writeSpaceLinear();
 	args[i].print(out);
@@ -445,11 +470,11 @@ public class ApplyExp extends Expression
   }
 
   /** Only used for inline- and tail-calls. */
-  private static void pushArgs (LambdaExp lexp, Expression[] args,
+  private static void pushArgs (LambdaExp lexp,
+                                Expression[] args, int args_length,
                                 int[] incValues, Compilation comp)
   {
     Declaration param = lexp.firstDecl();
-    int args_length = args.length;
     for (int i = 0; i < args_length; ++i)
       {
         Expression arg = args[i];
@@ -474,27 +499,19 @@ public class ApplyExp extends Expression
       vars = vars.nextVar();
     if (vars != null && vars.getName() == "$ctx")
       vars = vars.nextVar();
-    if (vars != null && vars.getName() == "argsArray")
-      {
-	if (toArray)
-	  {
-	    popParams (code, 0, 1, null, decls, vars);
-	    return;
-	  }
-        vars = vars.nextVar();
-      }
-    popParams (code, 0, lexp.min_args, incValues, decls, vars);
+    popParams (code, 0, lexp.min_args, toArray, incValues, decls, vars);
   }
 
   // Recursive helper function.
-  private static void popParams (CodeAttr code, int paramNo, int count,
+  private static void popParams (CodeAttr code, int paramNo,
+                                 int count, boolean toArray,
                                  int[] incValues,
                                  Declaration decl, Variable vars)
   {
     if (count > 0)
       {
         count--;
-	popParams (code, paramNo+1, count, incValues, decl.nextDecl(),
+	popParams (code, paramNo+1, count, toArray, incValues, decl.nextDecl(),
                    decl.getVariable() == null ? vars : vars.nextVar());
         if (! decl.ignorable())
           {
@@ -504,11 +521,10 @@ public class ApplyExp extends Expression
               code.emitStore(vars);
           }
       }
+    else if (toArray)
+      code.emitStore(vars); // rest array
   }
 
-  /** Cache for getType(). */
-  protected Type type;
-
   public final gnu.bytecode.Type getTypeRaw()
   {
     return type;
@@ -549,13 +565,9 @@ public class ApplyExp extends Expression
     return afunc;
   }
 
-  public final gnu.bytecode.Type getType()
+  protected gnu.bytecode.Type calculateType()
   {
-    if (type != null)
-      return type;
     Expression afunc = derefFunc(func);
-    // In case of cycles.
-    type = Type.objectType;
     if (afunc instanceof QuoteExp)
       {
         Object value = ((QuoteExp) afunc).getValue();
@@ -622,7 +634,7 @@ public class ApplyExp extends Expression
       }
     try
       {
-	return new QuoteExp(proc.applyN(vals), type);
+	return new QuoteExp(proc.applyN(vals), type).setLine(this);
       }
     catch (Throwable ex)
       {
