@@ -1,10 +1,11 @@
-// Copyright (c) 2008  Per M.A. Bothner.
+// Copyright (c) 2008, 2011  Per M.A. Bothner.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.expr;
 import gnu.bytecode.*;
 import gnu.mapping.*;
 import java.util.*;
+import java.lang.annotation.ElementType;
 
 public class ClassExp extends LambdaExp
 {
@@ -17,12 +18,13 @@ public class ClassExp extends LambdaExp
   public static final int INTERFACE_SPECIFIED = 2 * LambdaExp.NEXT_AVAIL_FLAG;
   public static final int CLASS_SPECIFIED = 4 * LambdaExp.NEXT_AVAIL_FLAG;
   public static final int HAS_SUBCLASS = 8 * LambdaExp.NEXT_AVAIL_FLAG;
+  public static final int IS_PACKAGE_MEMBER = 16 * LambdaExp.NEXT_AVAIL_FLAG;
 
   /** True if there is at least one explicit "<init>" ("*init*"} method. */
   boolean explicitInit;
 
   /** The class of instances of this class.
-   * Same as super.type unless isMakingClassPair(), in which case super.type
+   * Same as compiledType unless isMakingClassPair(), in which case super.type
    * is an interface, and instanceType is a class implementing the interface.
    * Using an interface plus a class gives us true multiple inheritance. */
   ClassType instanceType;
@@ -32,7 +34,7 @@ public class ClassExp extends LambdaExp
   /** True if we should make a pair of an interface and a class. */
   public boolean isMakingClassPair()
   {
-    return type != instanceType;
+    return compiledType != instanceType;
   }
 
   /** The ClassType generated for this class.
@@ -40,12 +42,18 @@ public class ClassExp extends LambdaExp
    * (viewed as an expression) is a class/type object, so getType returns
    * the type of a type.
    */
-  public Type getType()
+  protected Type calculateType()
   { return simple ? Compilation.typeClass : Compilation.typeClassType; }
 
   /** The ClassType generated for this class.
    */
-  public ClassType getClassType() { return type; }
+  public ClassType getClassType() { return compiledType; }
+
+  public void setClassType(ClassType type)
+  {
+    this.compiledType = type;
+    this.instanceType = type;
+  }
 
   /** List of base classes and implemented interfaces. */
   public Expression[] supers;
@@ -61,14 +69,10 @@ public class ClassExp extends LambdaExp
    * static initializations. */
   public LambdaExp clinitMethod;
 
-  public ClassExp ()
-  {
-  }
-
-  public ClassExp (boolean simple)
+  public ClassExp (boolean simple, ClassType type)
   {
     this.simple = simple;
-    instanceType = type = new ClassType();
+    setClassType(type);
   }
 
   protected boolean mustCompile () { return true; }
@@ -83,7 +87,7 @@ public class ClassExp extends LambdaExp
 
   public void compilePushClass (Compilation comp, Target target)
   {
-    ClassType new_class = type;
+    ClassType new_class = compiledType;
 
     gnu.bytecode.CodeAttr code = comp.getCode();
     comp.loadClassRef(new_class);
@@ -124,7 +128,7 @@ public class ClassExp extends LambdaExp
 
   protected ClassType getCompiledClassType(Compilation comp)
   {
-    return type;
+    return compiledType;
   }
 
   public void setTypes(Compilation comp)
@@ -136,13 +140,17 @@ public class ClassExp extends LambdaExp
     for (int i = 0;  i < nsupers;  i++)
       {
 	Type st = Language.getDefaultLanguage().getTypeFor(supers[i]);
-	if (! (st instanceof ClassType))
+        ClassType t;
+        if (st instanceof ClassType)
+            t = (ClassType) st;
+        else if (st instanceof ParameterizedType)
+            t = ((ParameterizedType) st).getRawType();
+        else
           {
             comp.setLine(supers[i]);
             comp.error('e', "invalid super type");
             continue;
           }
-	ClassType t = (ClassType) st;
 	int modifiers;
 	try
 	  {
@@ -167,21 +175,23 @@ public class ClassExp extends LambdaExp
     if (superType != null && (flags & INTERFACE_SPECIFIED) != 0)
       comp.error('e', "cannot be interface since has superclass");
     if (! simple && superType == null && (flags & CLASS_SPECIFIED) == 0
+        // If the class is module-local and has no sub-classes, we can
+        // optimze away the need for a pair-class.
         && (getFlag(HAS_SUBCLASS)
             || (nameDecl != null && nameDecl.isPublic())))
       {
         PairClassType ptype = new PairClassType();//(PairClassType) type;
-        type = ptype;
+        compiledType = ptype;
         ptype.setInterface(true);
         ptype.instanceType = instanceType;
-        ClassType[] interfaces = { type };
+        ClassType[] interfaces = { compiledType };
         // Can do better.  FIXME.
         instanceType.setSuper(Type.pointer_type);
         instanceType.setInterfaces(interfaces);
       }
     else if (getFlag(INTERFACE_SPECIFIED))
       instanceType.setInterface(true);
-    type.setSuper(superType == null ? Type.pointer_type : superType);
+    compiledType.setSuper(superType == null ? Type.pointer_type : superType);
 
     ClassType[] interfaces;
     if (j == nsupers)
@@ -191,86 +201,95 @@ public class ClassExp extends LambdaExp
 	interfaces = new ClassType[j];
 	System.arraycopy(superTypes, 0, interfaces, 0, j);
       }
-    type.setInterfaces(interfaces);
+    compiledType.setInterfaces(interfaces);
 
-    if (type.getName() == null)
+    if (compiledType.getName() == null)
+      compiledType.setName(getClassName(comp));
+
+    // Don't add class twice - i.e. if ModuleExp.USE_DEFINED_CLASS is set.
+    if (compiledType != comp.mainClass)
+	comp.addClass(compiledType);
+    if (isMakingClassPair())
       {
-	String name;
-        if (classNameSpecifier != null)
-          name = classNameSpecifier;
-        else
+        instanceType.setName(compiledType.getName()+"$class");
+        comp.addClass(instanceType);
+      }
+  }
+
+  public String getClassName (Compilation comp)
+  {
+    String name;
+    if (classNameSpecifier != null)
+      name = classNameSpecifier;
+    else
+      {
+        name = getName();
+        if (name != null)
           {
-            name = getName();
-            if (name != null)
-              {
-                int nlen = name.length();
-                if (nlen > 2
-                    && name.charAt(0) == '<' && name.charAt(nlen-1) == '>')
-                  name = name.substring(1, nlen-1);
-              }
-          }
-        if (name == null)
-          {
-	    StringBuffer nbuf = new StringBuffer(100);
-            comp.getModule().classFor(comp);
-            nbuf.append(comp.mainClass.getName());
-            nbuf.append('$');
-            int len = nbuf.length();
-            for (int i = 0;  ; i++)
-              {
-                nbuf.append(i);
-                name = nbuf.toString();
-                if (comp.findNamedClass(name) == null)
-                  break;
-                nbuf.setLength(len);
-              }
-          }
-	else if (! isSimple() || this instanceof ObjectExp)
-	  name = comp.generateClassName(name);
-	else
-	  {
-	    int start = 0;
-	    StringBuffer nbuf = new StringBuffer(100);
-	    for (;;)
-	      {
-		int dot = name.indexOf('.', start);
-		if (dot < 0)
-		  break;
-		nbuf.append(Compilation
-			    .mangleNameIfNeeded(name.substring(start, dot)));
-		start = dot + 1;
-                if (start < name.length())
-                  nbuf.append('.');
-	      }
-	    if (start == 0)
-	      {
-		String mainName = comp.mainClass == null ? null
-		  : comp.mainClass.getName();
-		int dot = mainName == null ? -1 : mainName.lastIndexOf('.');
-		if (dot > 0)
-		  nbuf.append(mainName.substring(0, dot + 1));
-		else if (comp.classPrefix != null)
-		  nbuf.append(comp.classPrefix);
-	      }
-            else if (start == 1 && start < name.length())
-              {
-                nbuf.setLength(0);
-                nbuf.append(comp.mainClass.getName());
-                nbuf.append('$');
-              }
-	    if (start < name.length())
-	      nbuf.append(Compilation
-			  .mangleNameIfNeeded(name.substring(start)));
-	    name = nbuf.toString();
-	  }
-	type.setName(name);
-        comp.addClass(type);
-        if (isMakingClassPair())
-          {
-            instanceType.setName(type.getName()+"$class");
-            comp.addClass(instanceType);
+            int nlen = name.length();
+            if (nlen > 2
+                && name.charAt(0) == '<' && name.charAt(nlen-1) == '>')
+              name = name.substring(1, nlen-1);
           }
       }
+    if (name == null)
+      {
+        StringBuffer nbuf = new StringBuffer(100);
+        comp.getModule().classFor(comp);
+        nbuf.append(comp.mainClass.getName());
+        nbuf.append('$');
+        int len = nbuf.length();
+        for (int i = 0;  ; i++)
+          {
+            nbuf.append(i);
+            name = nbuf.toString();
+            if (comp.findNamedClass(name) == null)
+              break;
+            nbuf.setLength(len);
+          }
+      }
+    else if (! isSimple() || this instanceof ObjectExp)
+      name = comp.generateClassName(name);
+    else
+      {
+        int start = 0;
+        StringBuffer nbuf = new StringBuffer(100);
+        for (;;)
+          {
+            int dot = name.indexOf('.', start);
+            if (dot < 0)
+              break;
+            nbuf.append(Compilation
+                        .mangleNameIfNeeded(name.substring(start, dot)));
+            start = dot + 1;
+            if (start < name.length())
+              nbuf.append('.');
+          }
+        if (start == 0)
+          {
+            setFlag(IS_PACKAGE_MEMBER);
+            String mainName = comp.mainClass == null ? null
+              : comp.mainClass.getName();
+            int dot = mainName == null ? -1 : mainName.lastIndexOf('.');
+            if (dot > 0)
+              nbuf.append(mainName.substring(0, dot + 1));
+            else if (comp.classPrefix != null)
+              nbuf.append(comp.classPrefix);
+          }
+        else if (start == 1 && start < name.length())
+          {
+            nbuf.setLength(0);
+            nbuf.append(comp.mainClass.getName());
+            nbuf.append('$');
+          }
+        else
+          setFlag(IS_PACKAGE_MEMBER);
+        if (start < name.length())
+          nbuf.append(Compilation
+                      .mangleNameIfNeeded(name.substring(start)));
+        name = nbuf.toString();
+      }
+    return name;
   }
 
   boolean partsDeclared;
@@ -295,10 +314,10 @@ public class ClassExp extends LambdaExp
 	      {
 		flags |= Access.ABSTRACT;
 		Type ftype = decl.getType().getImplementationType();
-		type.addMethod(slotToMethodName("get", decl.getName()),
+		compiledType.addMethod(slotToMethodName("get", decl.getName()),
 			       flags, Type.typeArray0, ftype);
 		Type[] stypes = { ftype };
-		type.addMethod(slotToMethodName("set",decl.getName()),
+		compiledType.addMethod(slotToMethodName("set",decl.getName()),
 			       flags, stypes, Type.voidType);
 	      }
 	    else
@@ -326,7 +345,7 @@ public class ClassExp extends LambdaExp
             explicitInit = true;
             if (child.isAbstract())
               comp.error('e', "*init* method cannot be abstract", child);
-            if (type instanceof PairClassType)
+            if (compiledType instanceof PairClassType)
               comp.error('e', "'*init*' methods only supported for simple classes");
           }
         // Setting child.outer isn't normally needed.  The exception is
@@ -338,9 +357,9 @@ public class ClassExp extends LambdaExp
              && child.nameDecl != null // only if error
              && ! child.nameDecl.getFlag(Declaration.STATIC_SPECIFIED))
 	    || ! isMakingClassPair())
-	  child.addMethodFor(type, comp, null);
+	  child.addMethodFor(compiledType, comp, null);
 	if (isMakingClassPair())
-	  child.addMethodFor(instanceType, comp, type);
+	  child.addMethodFor(instanceType, comp, compiledType);
       }
     if (! explicitInit && ! instanceType.isInterface())
       Compilation.getConstructor(instanceType, this);
@@ -435,11 +454,11 @@ public class ClassExp extends LambdaExp
         LambdaExp outer = outerLambda();
         Member enclosing = null;
         if (outer instanceof ClassExp)
-          enclosing = outer.type;
+          enclosing = outer.compiledType;
         else if (outer != null && ! (outer instanceof ModuleExp))
           enclosing = saveMethod;
-        else if (outer instanceof ModuleExp && type.getName().indexOf('$') > 0)
-         enclosing = outer.type;
+        else if (outer instanceof ModuleExp && ! getFlag(IS_PACKAGE_MEMBER))
+          enclosing = outer.compiledType;
         if (enclosing != null)
           {
             new_class.setEnclosingMember(enclosing);
@@ -448,12 +467,12 @@ public class ClassExp extends LambdaExp
           }
         if (instanceType != new_class)
           {
-            instanceType.setEnclosingMember(type);
-            type.addMemberClass(instanceType);
+            instanceType.setEnclosingMember(compiledType);
+            compiledType.addMemberClass(instanceType);
           }
-              
-	usedSuperClasses(type, comp);
-	if (type != instanceType)
+
+	usedSuperClasses(compiledType, comp);
+	if (compiledType != instanceType)
 	  usedSuperClasses(instanceType, comp);
 
 	String filename = getFileName();
@@ -466,10 +485,18 @@ public class ClassExp extends LambdaExp
 	allocFrame(comp);
 	CodeAttr code;
 
+        if (nameDecl != null)
+          nameDecl.compileAnnotations(compiledType, ElementType.TYPE);
+        for (Declaration decl = firstDecl(); decl != null;
+             decl = decl.nextDecl())
+          {
+            decl.compileAnnotations(decl.field, ElementType.FIELD);
+          }
+
 	for (LambdaExp child = firstChild;  child != null;
              child = child.nextSibling)
 	  {
-            if (child.isAbstract())
+            if (child.isAbstract() || child.isNative())
               continue;
 	    Method save_method = comp.method;
 	    LambdaExp save_lambda = comp.curLambda;
@@ -483,6 +510,8 @@ public class ClassExp extends LambdaExp
             if (childDecl == null
                 || ! childDecl.getFlag(Declaration.STATIC_SPECIFIED))
               child.declareThis(comp.curClass);
+            if (childDecl != null)
+              childDecl.compileAnnotations(comp.method, ElementType.METHOD);
 	    comp.curClass = instanceType;
 	    comp.curLambda = child;
 	    comp.method.initCode();
@@ -500,34 +529,13 @@ public class ClassExp extends LambdaExp
                   }
 
                 // Extract "first" expression to see if it is special.
-                Expression bodyFirst = child.body;
-                while (bodyFirst instanceof BeginExp)
-                  {
-                    BeginExp bbody = (BeginExp) bodyFirst;
-                    if (bbody.length == 0)
-                      bodyFirst = null;
-                    else
-                      bodyFirst = bbody.exps[0];
-                  }
-
+                Expression bodyFirst = child.getBodyFirstExpression();
                 // See if bodyFirst is a this(...) or super(...) call.
-                ClassType calledInit = null;
-                Object value;  Expression exp;
-                if (bodyFirst instanceof ApplyExp
-                    && (exp = ((ApplyExp) bodyFirst).func) instanceof QuoteExp
-                    && (value = ((QuoteExp) exp).getValue()) instanceof PrimProcedure)
-                  {
-                    PrimProcedure pproc = (PrimProcedure) value;
-                    if (pproc.isSpecial()
-                        && ("<init>".equals(pproc.method.getName())))
-                      calledInit = pproc.method.getDeclaringClass();
-                  }
+                ClassType calledInit = checkForInitCall(bodyFirst);
                 ClassType superClass = instanceType.getSuperclass();
                 if (calledInit != null)
                   {
                     bodyFirst.compileWithPosition(comp, Target.Ignore);
-                    if (calledInit != instanceType && calledInit != superClass)
-                      comp.error('e', "call to <init> for not this or super class");
                   }
                 else if (superClass != null)
                   {
@@ -552,6 +560,23 @@ public class ClassExp extends LambdaExp
               }
 	    child.compileEnd(comp);
 	    child.generateApplyMethods(comp);
+        // Check to see if child overrides a superclass method and has
+        // a covariant return type. If so, generate a bridge method
+        // matching the superclass method's return type.
+        {
+          Method thisMethod = comp.method;
+          Type[] ptypes = thisMethod.getParameterTypes();
+          Type rtype = thisMethod.getReturnType();
+          ClassType superClass = instanceType.getSuperclass();
+          Method superMethod = superClass.getMethod(child.getName(),
+                                                    ptypes);
+          if (superMethod != null &&
+              superMethod.getReturnType().compare(rtype) == 1)
+          {
+            generateBridgeMethod(comp, thisMethod, ptypes,
+                                 superMethod.getReturnType());
+          }
+        }
 	    comp.method = save_method;
 	    comp.curClass = new_class;
 	    comp.curLambda = save_lambda;
@@ -571,7 +596,7 @@ public class ClassExp extends LambdaExp
           }
         else
           {
-            methods = type.getAbstractMethods();
+            methods = compiledType.getAbstractMethods();
             nmethods = methods.length;
           }
 	for (int i = 0;  i < nmethods;  i++)
@@ -621,14 +646,24 @@ public class ClassExp extends LambdaExp
 	    else
 	      {
 		Vector vec = new Vector();
-		getImplMethods(type, mname, ptypes, vec);
+		getImplMethods(compiledType, mname, ptypes, vec);
 		if (vec.size() != 1)
 		  {
-		    // FIXME - need better error message!
-		    String msg = vec.size() == 0
-		      ? "missing implementation for "
-		      : "ambiguous implementation for ";
-		    comp.error('e', msg+meth);
+            Method impl = vec.size() != 0 ? null :
+              findMethodForBridge(mname, ptypes, rtype);
+
+            if (impl != null)
+            {
+              generateBridgeMethod(comp, impl, ptypes, rtype);
+            }
+            else
+            {
+              // FIXME - need better error message!
+              String msg = vec.size() == 0
+                ? "missing implementation for "
+                : "ambiguous implementation for ";
+              comp.error('e', msg+meth);
+            }
 		  }
 		else
 		  {
@@ -657,6 +692,63 @@ public class ClassExp extends LambdaExp
       }
   }
 
+  /**
+   * Finds a like-named method suitable for bridging the given
+   * arg/return types (i.e. a method whose arg types and return types
+   * are subclasses of those given here.
+   */
+  protected Method findMethodForBridge(String mname, Type[] ptypes,
+                                       Type rtype)
+  {
+    Method result = null;
+    for (Method method = compiledType.getDeclaredMethods();
+         method != null; method = method.getNext())
+    {
+      if (mname.equals(method.getName()) &&
+          method.getReturnType().isSubtype(rtype) &&
+          Type.isMoreSpecific(method.getParameterTypes(), ptypes))
+        result = method;
+    }
+    return result;
+  }
+
+  /**
+   * Given an existing method and a desired bridge method signature,
+   * generates an appropriate bridge method.
+   */
+  public final void generateBridgeMethod(Compilation comp,
+                                         Method src_method,
+                                         Type[] bridge_arg_types,
+                                         Type bridge_return_type)
+  {
+    ClassType save_class = comp.curClass;
+    Method save_method = comp.method;
+    try
+    {
+      comp.curClass = getCompiledClassType(comp);
+      comp.method = comp.curClass.addMethod
+        (src_method.getName(),
+         Access.PUBLIC|Access.BRIDGE|Access.SYNTHETIC,
+         bridge_arg_types, bridge_return_type);
+      Type[] src_arg_types = src_method.getParameterTypes();
+
+      CodeAttr code = comp.method.startCode();
+      code.emitLoad(code.getArg(0));
+      for (int i = 0; i < src_arg_types.length; ++i)
+      {
+        code.emitLoad(code.getArg(i+1));
+        code.emitCheckcast(src_arg_types[i]);
+      }
+      code.emitInvokeVirtual(src_method);
+      code.emitReturn();
+    }
+    finally
+    {
+      comp.method = save_method;
+      comp.curClass = save_class;
+    }
+  }
+
   protected <R,D> R visit (ExpVisitor<R,D> visitor, D d)
   {
     Compilation comp = visitor.getCompilation();
@@ -666,7 +758,7 @@ public class ClassExp extends LambdaExp
     try
 
       {
-	comp.curClass = type;
+	comp.curClass = compiledType;
 	return visitor.visitClassExp(this, d);
       }
     finally
@@ -690,7 +782,7 @@ public class ClassExp extends LambdaExp
               {
                 Declaration firstParam = child.firstDecl();
                 if (firstParam != null && firstParam.isThisParameter())
-                  firstParam.setType(type);
+                  firstParam.setType(compiledType);
               }
             visitor.visitLambdaExp(child, d);
           }
@@ -711,6 +803,13 @@ public class ClassExp extends LambdaExp
     code.emitCheckcast(superClass.getOuterLinkType());
   }
 
+  void checkDefaultSuperConstructor(ClassType superClass, Compilation comp)
+  {
+    if (superClass.getDeclaredMethod("<init>", 0) == null)
+      comp.error('e', ("super class "+superClass.getName()
+                       +" does not have a default constructor"));
+  }
+
   static void invokeDefaultSuperConstructor (ClassType superClass,
                                              Compilation comp,
                                              LambdaExp lexp)
@@ -718,19 +817,17 @@ public class ClassExp extends LambdaExp
     CodeAttr code = comp.getCode();
     Method superConstructor
       = superClass.getDeclaredMethod("<init>", 0);
-    if (superConstructor == null)
-      comp.error('e', "super class does not have a default constructor");
-    else
+    // InlineCalls catches the missing superConstructor case, using
+    // checkDefaultSuperConstructor.
+    assert superConstructor != null;
+    code.emitPushThis();
+    if (superClass.hasOuterLink() && lexp instanceof ClassExp)
       {
-        code.emitPushThis();
-        if (superClass.hasOuterLink() && lexp instanceof ClassExp)
-          {
-            ClassExp clExp = (ClassExp) lexp;
-            Expression superExp = clExp.supers[clExp.superClassIndex];
-            loadSuperStaticLink(superExp, superClass, comp);
-          }
-        code.emitInvokeSpecial(superConstructor);
+        ClassExp clExp = (ClassExp) lexp;
+        Expression superExp = clExp.supers[clExp.superClassIndex];
+        loadSuperStaticLink(superExp, superClass, comp);
       }
+    code.emitInvokeSpecial(superConstructor);
   }
 
   public void print (OutPort out)
@@ -756,19 +853,14 @@ public class ClassExp extends LambdaExp
           }
         out.endLogicalBlock("");
       }
-    out.print('(');
     Special prevMode = null;
-    int i = 0;
     int key_args = keywords == null ? 0 : keywords.length;
     //int opt_args = defaultArgs == null ? 0 : defaultArgs.length - key_args;
     for (Declaration decl = firstDecl();  decl != null; decl = decl.nextDecl())
       {
-	if (i > 0)
-	  out.print(' ');
+        out.writeSpaceFill();
         decl.printInfo(out);
-	i++;
       }
-    out.print(") ");
     for (LambdaExp child = firstChild;  child != null;
 	 child = child.nextSibling)
       {
@@ -794,7 +886,7 @@ public class ClassExp extends LambdaExp
    */
   public static String slotToMethodName(String prefix, String sname)
   {
-    if (! Compilation.isValidJavaName(sname))
+    if (! Language.isValidJavaName(sname))
       sname = Compilation.mangleName(sname, false);
     int slen = sname.length();
     StringBuffer sbuf = new StringBuffer(slen+3);
