@@ -11,6 +11,7 @@ import gnu.text.*;
 import java.util.zip.*;
 import java.util.Stack;
 import gnu.kawa.functions.Convert;
+import gnu.kawa.reflect.LazyType;
 
 /** State for a single expression or module.
  * For each top-level thing (expression or file) we compile or evaluate
@@ -110,6 +111,14 @@ public class Compilation implements SourceLocator
   public static boolean debugPrintFinalExpr;
 
   public static Options options = new Options();
+  public static Options.OptionInfo fullTailCallsVariable =
+    options.add("full-tailcalls",
+                Options.BOOLEAN_OPTION, Boolean.TRUE,
+		"support full tailcalls");
+  public static Options.OptionInfo warnUnreachable =
+    options.add("warn-unreachable",
+                Options.BOOLEAN_OPTION, Boolean.TRUE,
+		"warn if this code can never be executed");
   public static Options.OptionInfo warnUndefinedVariable =
     options.add("warn-undefined-variable",
                 Options.BOOLEAN_OPTION, Boolean.TRUE,
@@ -122,12 +131,20 @@ public class Compilation implements SourceLocator
     options.add("warn-invoke-unknown-method",
                 Options.BOOLEAN_OPTION, warnUnknownMember,
 		"warn if invoke calls an unknown method (subsumed by warn-unknown-member)");
+  public static Options.OptionInfo warnUnused =
+    options.add("warn-unused",
+                Options.BOOLEAN_OPTION, Boolean.TRUE,
+                "warn if a variable is usused or code never executed");
   public static Options.OptionInfo warnAsError =
     options.add("warn-as-error", Options.BOOLEAN_OPTION, Boolean.FALSE,
 		"Make all warnings into errors");
 
   public Options currentOptions = new Options(options);
 
+  public boolean warnUnreachable ()
+  {
+    return currentOptions.getBoolean(warnUnreachable);
+  }
   public boolean warnUndefinedVariable ()
   {
     return currentOptions.getBoolean(warnUndefinedVariable);
@@ -140,6 +157,10 @@ public class Compilation implements SourceLocator
   {
     return currentOptions.getBoolean(warnInvokeUnknownMethod);
   }
+  public boolean warnUnused()
+  {
+    return currentOptions.getBoolean(warnUnused);
+  }
   public boolean warnAsError ()
   {
     return currentOptions.getBoolean(warnAsError);
@@ -174,7 +195,7 @@ public class Compilation implements SourceLocator
     ;
 
   /** The default calling convention.
-   * One of the following CALL_WITHG_xxx values. */
+   * One of the following CALL_WITH_xxx values. */
   public static int defaultCallConvention;
   public static final int CALL_WITH_UNSPECIFIED = 0;
   /** Plain calling convention, using regular Java parameters and returns. */
@@ -186,10 +207,21 @@ public class Compilation implements SourceLocator
   /** Support for full continuations.  Not implemented. */
   public static final int CALL_WITH_CONTINUATIONS = 4;
 
+    public int currentCallConvention() {
+        Object ft = currentOptions.getLocal("full-tailcalls");
+        if (ft instanceof Boolean)
+            return ((Boolean) ft).booleanValue()
+                ? Compilation.CALL_WITH_TAILCALLS
+                : Compilation.CALL_WITH_RETURN;
+        return defaultCallConvention;
+    }
+
   public boolean usingCPStyle()
-  { return defaultCallConvention == CALL_WITH_CONTINUATIONS; }
+  { return currentCallConvention() == CALL_WITH_CONTINUATIONS; }
   public boolean usingTailCalls()
-  { return defaultCallConvention >= CALL_WITH_TAILCALLS; }
+  { return currentCallConvention() >= CALL_WITH_TAILCALLS; }
+  public boolean usingCallContext()
+  { return currentCallConvention() >= CALL_WITH_CONSUMER; }
 
   public static final int MODULE_NONSTATIC = -1;
   public static final int MODULE_STATIC_DEFAULT = 0;
@@ -235,10 +267,13 @@ public class Compilation implements SourceLocator
   static public ClassType javaStringType = typeString;
   static public ClassType scmKeywordType = ClassType.make("gnu.expr.Keyword");
   static public ClassType scmSequenceType = ClassType.make("gnu.lists.Sequence");
+  public static final ClassType typeList = ClassType.make("java.util.List");
   static public ClassType scmListType = ClassType.make("gnu.lists.LList");
   static public ClassType typePair = ClassType.make("gnu.lists.Pair");
+  public static final ClassType typeConstVector = ClassType.make("gnu.lists.ConstVector");
   public static final ArrayType objArrayType = ArrayType.make(typeObject);
   static public ClassType typeRunnable = ClassType.make("java.lang.Runnable");
+  static public ClassType typeRunnableModule = ClassType.make("gnu.expr.RunnableModule");
   public static ClassType typeType = ClassType.make("gnu.bytecode.Type");
   public static ClassType typeObjectType
     = ClassType.make("gnu.bytecode.ObjectType");
@@ -252,6 +287,10 @@ public class Compilation implements SourceLocator
     = ClassType.make("gnu.mapping.Environment");
   static public ClassType typeLocation
     = ClassType.make("gnu.mapping.Location");
+  static public ClassType typeFieldLocation
+    = ClassType.make("gnu.kawa.reflect.FieldLocation");
+  static public ClassType typeStaticFieldLocation
+    = ClassType.make("gnu.kawa.reflect.StaticFieldLocation");
   static public ClassType typeSymbol
     = ClassType.make("gnu.mapping.Symbol");
   static public final Method getSymbolValueMethod
@@ -354,8 +393,6 @@ public class Compilation implements SourceLocator
     = ClassType.make("gnu.mapping.ProcedureN");
   public static ClassType typeModuleBody
     = ClassType.make("gnu.expr.ModuleBody");
-  public static ClassType typeModuleWithContext
-    = ClassType.make("gnu.expr.ModuleWithContext");
   public static ClassType typeApplet = ClassType.make("java.applet.Applet");
   public static ClassType typeServlet = ClassType.make("gnu.kawa.servlet.KawaServlet");
 
@@ -376,6 +413,8 @@ public class Compilation implements SourceLocator
   = ClassType.make("gnu.mapping.MethodProc");
   public static ClassType typeModuleMethod
   = ClassType.make("gnu.expr.ModuleMethod");
+  public static ClassType typeModuleMethodWithContext
+  = ClassType.make("gnu.expr.ModuleMethodWithContext");
   //  public static Field numArgsCallFrameField = typeCallFrame.getDeclaredField("numArgs");
   public static Field argsCallContextField
     = typeCallContext.getDeclaredField("values");
@@ -428,9 +467,7 @@ public class Compilation implements SourceLocator
 
   public final ClassType getModuleType()
   {
-    return (defaultCallConvention >= Compilation.CALL_WITH_CONSUMER
-	    ? typeModuleWithContext
-	    : typeModuleBody);
+    return typeModuleBody;
   }
 
   /** Emit code to "evaluate" a compile-time constant.
@@ -486,6 +523,28 @@ public class Compilation implements SourceLocator
     return inlineOk;
   }
 
+    /** Should this inlineable method by inlined?
+     * Usually it's best to not inline a module-level function, since that
+     * makes stack traces less helpful, and increases the risk of
+     * methods getting too big.
+     */
+    static boolean avoidInline(LambdaExp proc) {
+        return proc.outer instanceof ModuleExp && proc.nameDecl != null;
+    }
+
+  public boolean isApplyFunction (Expression exp)
+  {
+    return false;
+  }
+
+  /** A simple apply function maps actual arguments to formals directly.
+   * E.g. no distribution of multiple values.
+   */
+  public boolean isSimpleApplyFunction (Expression exp)
+  {
+    return false;
+  }
+
   public void compileConstant (Object value, Target target)
   {
     if (target instanceof IgnoreTarget)
@@ -513,6 +572,8 @@ public class Compilation implements SourceLocator
     if (target instanceof StackTarget)
       {
 	Type type = ((StackTarget) target).getType();
+	if (type instanceof LazyType)
+	  type = ((LazyType) type).getValueType();
 	if (type instanceof PrimType)
 	  {
 	    try
@@ -553,7 +614,7 @@ public class Compilation implements SourceLocator
 		  }
 		if (sig1 == 'Z')
 		  {
-		    boolean val = PrimType.booleanValue(value);
+		    boolean val = getLanguage().isTrue(value);
 		    code.emitPushInt(val ? 1 : 0);
 		    return;
 		  }
@@ -586,7 +647,7 @@ public class Compilation implements SourceLocator
                   sbuf.append(value.getClass().getName());
 		sbuf.append(") to ");
 	      }
-	    sbuf.append(type.getName());
+	    sbuf.append(type);
             error('w', sbuf.toString());
          }
       }
@@ -596,6 +657,25 @@ public class Compilation implements SourceLocator
                             : Type.make(value.getClass()));
   }
 
+  public void emitPushBoolean(boolean value)
+  {
+    getCode().emitGetStatic(value ? Compilation.trueConstant
+                            : Compilation.falseConstant);
+  }
+
+  /** Generate code to test if an object is considered true.
+   * Assume the object has been pushed on the JVM stack.
+   * Generate code to push true or false as appropriate. */
+  public void emitCoerceToBoolean()
+  {
+    CodeAttr code = getCode();
+    emitPushBoolean(false);
+    code.emitIfNEq();
+    code.emitPushInt(1);
+    code.emitElse();
+    code.emitPushInt(0);
+    code.emitFi();
+  }
 
   private void dumpInitializers (Initializer inits)
   {
@@ -704,26 +784,7 @@ public class Compilation implements SourceLocator
 
   public static String mangleName (String name)
   {
-    return mangleName(name, -1);
-  }
-
-  public static String mangleNameIfNeeded (String name)
-  {
-    if (name == null || isValidJavaName(name))
-      return name;
-    else
-      return mangleName(name, 0);
-  }
-
-  public static boolean isValidJavaName(String name)
-  {
-    int len = name.length();
-    if (len == 0 || ! Character.isJavaIdentifierStart(name.charAt(0)))
-      return false;
-    for (int i = len;  --i > 0; )
-      if (! Character.isJavaIdentifierPart(name.charAt(i)))
-	return false;
-    return true;
+    return Language.mangleName(name, -1);
   }
 
   /** Convert a string to a safe Java identifier.
@@ -731,109 +792,12 @@ public class Compilation implements SourceLocator
    */
   public static String mangleName (String name, boolean reversible)
   {
-    return mangleName(name, reversible ? 1 : -1);
+    return Language.mangleName(name, reversible ? 1 : -1);
   }
 
-  /** Convert a string to a safe Java identifier.
-   * @param kind -1 - non-reversible;
-   *  0: reversible, except that '$' is not mapped;
-   *  1: reversible
-   */
-  public static String mangleName (String name, int kind)
+  public static String mangleNameIfNeeded (String name)
   {
-    boolean reversible = kind >= 0;
-    int len = name.length ();
-    if (len == 6 && name.equals("*init*")) // Constructor methods.
-      return "<init>";
-    StringBuffer mangled = new StringBuffer (len);
-    boolean upcaseNext = false;
-    for (int i = 0;  i < len;  i++)
-      {
-	char ch = name.charAt(i);
-	if (upcaseNext)
-	  {
-	    ch = Character.toTitleCase(ch);
-	    upcaseNext = false;
-	  }
-	if (Character.isDigit(ch))
-	  {
-	    if (i == 0)
-	      mangled.append("$N");
-	    mangled.append(ch);
-	  }
-	else if (Character.isLetter(ch) || ch == '_')
-	  mangled.append(ch);
-	else if (ch == '$')
-	  mangled.append(kind > 1 ? "$$" : "$");
-	else
-	  {
-	    switch (ch)
-	      {
-	      case '+':  mangled.append("$Pl");  break;
-	      case '-':
-		if (reversible)
-		  mangled.append("$Mn");
-		else
-		  {
-		    char next = i + 1 < len ? name.charAt(i+1) : '\0';
-		    if (next == '>')
-		      {
-			mangled.append("$To$");
-			i++;
-		      }
-		    else if (! Character.isLowerCase(next))
-		      mangled.append("$Mn");
-		  }
-		break;
-	      case '*':  mangled.append("$St");  break;
-	      case '/':  mangled.append("$Sl");  break;
-	      case '=':  mangled.append("$Eq");  break;
-	      case '<':  mangled.append("$Ls");  break;
-	      case '>':  mangled.append("$Gr");  break;
-	      case '@':  mangled.append("$At");  break;
-	      case '~':  mangled.append("$Tl");  break;
-	      case '%':  mangled.append("$Pc");  break;
-	      case '.':  mangled.append("$Dt");  break;
-	      case ',':  mangled.append("$Cm");  break;
-	      case '(':  mangled.append("$LP");  break;
-	      case ')':  mangled.append("$RP");  break;
-	      case '[':  mangled.append("$LB");  break;
-	      case ']':  mangled.append("$RB");  break;
-	      case '{':  mangled.append("$LC");  break;
-	      case '}':  mangled.append("$RC");  break;
-	      case '\'': mangled.append("$Sq");  break;
-	      case '"':  mangled.append("$Dq");  break;
-	      case '&':  mangled.append("$Am");  break;
-	      case '#':  mangled.append("$Nm");  break;
-	      case '?':
-		char first = mangled.length() > 0 ? mangled.charAt(0) : '\0';
-		if (! reversible
-		    && i + 1 == len && Character.isLowerCase(first))
-		  {
-		    mangled.setCharAt(0, Character.toTitleCase(first));
-		    mangled.insert(0, "is");
-		  }
-		else
-		  mangled.append("$Qu");
-		break;
-	      case '!':  mangled.append("$Ex");  break;
-	      case ':':  mangled.append("$Cl");  break;
-	      case ';':  mangled.append("$SC");  break;
-	      case '^':  mangled.append("$Up");  break;
-	      case '|':  mangled.append("$VB");  break;
-	      default:
-		mangled.append('$');
-		mangled.append(Character.forDigit ((ch >> 12) & 15, 16));
-		mangled.append(Character.forDigit ((ch >>  8) & 15, 16));
-		mangled.append(Character.forDigit ((ch >>  4) & 15, 16));
-		mangled.append(Character.forDigit ((ch      ) & 15, 16));
-	      }
-	    if (! reversible)
-	      upcaseNext = true;
-	  }
-      }
-    String mname = mangled.toString ();
-    return mname.equals(name) ? name : mname;
+    return Language.mangleNameIfNeeded(name);
   }
 
   /** Demangle a three-character mangling starting with '$'.
@@ -991,8 +955,8 @@ public class Compilation implements SourceLocator
 	dout.flush();
       }
 
+    PushApply.pushApply(mexp, this);
     InlineCalls.inlineCalls(mexp, this);
-    PushApply.pushApply(mexp);
     ChainLambdas.chainLambdas(mexp, this);
     FindTailCalls.findTailCalls(mexp, this);
   }
@@ -1137,11 +1101,11 @@ public class Compilation implements SourceLocator
       }
     if (makeRunnable())
       type.addInterface(typeRunnable);
+    type.addInterface(typeRunnableModule);
     type.setSuper(sup);
 
-    module.type = type;
+    module.compiledType = type;
     addClass(type);
-    getConstructor(mainClass, module);
   }
 
   public final Method getConstructor (LambdaExp lexp)
@@ -1382,6 +1346,7 @@ public class Compilation implements SourceLocator
 		    Type ptype = var.getType();
 		    if (ptype != Type.objectType)
 		      {
+                        StackTarget.forceLazyIfNeeded(this, Type.objectType, ptype);
 			if (ptype instanceof TypeValue)
 			  {
 			    Label trueLabel = new Label(code),
@@ -1421,13 +1386,14 @@ public class Compilation implements SourceLocator
 		code.emitPutField(typeCallContext.getField("values"));
 	      }
 	    code.emitLoad(ctxVar);
-	    if (defaultCallConvention < Compilation.CALL_WITH_CONSUMER)
-	      code.emitLoad(code.getArg(1)); // proc
-	    else
+            boolean usingCallContext = usingCallContext();
+            if (usingCallContext)
 	      code.emitLoad(code.getArg(0)); // this (module)
+	    else
+	      code.emitLoad(code.getArg(1)); // proc
 	    code.emitPutField(procCallContextField);
 	    code.emitLoad(ctxVar);
-	    if (defaultCallConvention >= CALL_WITH_CONSUMER)
+	    if (usingCallContext)
 	      code.emitPushInt(source.getSelectorValue(this)+methodIndex);
 	    else
 	      code.emitPushInt(i);
@@ -1464,7 +1430,7 @@ public class Compilation implements SourceLocator
       return;
     ClassType save_class = curClass;
     curClass = lexp.getHeapFrameType();
-    if (! (curClass.getSuperclass().isSubtype(typeModuleWithContext)))
+    if (! (curClass.getSuperclass().isSubtype(typeModuleBody)))
       curClass = moduleClass;
     ClassType procType = typeModuleMethod;
     Method save_method = method;
@@ -1485,6 +1451,8 @@ public class Compilation implements SourceLocator
     for (int j = 0;  j < numApplyMethods;  ++j)
       {
 	LambdaExp source = (LambdaExp) lexp.applyMethods.elementAt(j);
+        if (! source.usingCallContext())
+          continue;
 	Method[] primMethods = source.primMethods;
 	int numMethods = primMethods.length;
 
@@ -1593,7 +1561,7 @@ public class Compilation implements SourceLocator
 	    code.emitInvoke(primMethod);
 	    while (--pendingIfEnds >= 0)
 	      code.emitFi();
-	    if (defaultCallConvention < Compilation.CALL_WITH_CONSUMER)
+	    if (! usingCallContext())
 	      Target.pushObject.compileFromStack(this,
 						 source.getReturnType());
             messages.swapSourceLocator(saveLoc1);
@@ -1626,9 +1594,7 @@ public class Compilation implements SourceLocator
       curClass = moduleClass;
     Method save_method = method;
     CodeAttr code = null;
-    for (int i = defaultCallConvention >= Compilation.CALL_WITH_CONSUMER
-	   ? 5 : 0;
-	 i < 6; i++)
+    for (int i = usingCallContext() ? 5 : 0; i < 6; i++)
       {
 	// If i < 5, generate the method named ("apply"+i);
 	// else generate "applyN".
@@ -1640,6 +1606,8 @@ public class Compilation implements SourceLocator
 	for (int j = 0;  j < numApplyMethods;  j++)
 	  {
 	    LambdaExp source = (LambdaExp) lexp.applyMethods.elementAt(j);
+            if (source.usingCallContext())
+              continue;
 	    // Select the subset of source.primMethods[*] that are suitable
 	    // for the current apply method.
 	    Method[] primMethods = source.primMethods;
@@ -1684,7 +1652,7 @@ public class Compilation implements SourceLocator
 		  }
 		applyArgs[0] = procType;
 		method = curClass.addMethod (mname, applyArgs,
-					     defaultCallConvention >= Compilation.CALL_WITH_CONSUMER ? (Type) Type.voidType : (Type) Type.objectType,
+					     usingCallContext() ? (Type) Type.voidType : (Type) Type.objectType,
 					     Access.PUBLIC);
 		code = method.startCode();
 
@@ -1790,7 +1758,7 @@ public class Compilation implements SourceLocator
 	    code.emitInvoke(primMethod);
 	    while (--pendingIfEnds >= 0)
 	      code.emitFi();
-	    if (defaultCallConvention < Compilation.CALL_WITH_CONSUMER)
+	    if (! usingCallContext())
 	      Target.pushObject.compileFromStack(this,
 						 source.getReturnType());
             messages.swapSourceLocator(saveLoc1);
@@ -1799,7 +1767,7 @@ public class Compilation implements SourceLocator
 	if (needThisApply)
 	  {
 	    aswitch.addDefault(code);
-	    if (defaultCallConvention >= Compilation.CALL_WITH_CONSUMER)
+	    if (usingCallContext())
 	      {
 		Method errMethod
 		  = typeModuleMethod.getDeclaredMethod("applyError", 0);
@@ -1986,8 +1954,7 @@ public class Compilation implements SourceLocator
           }
         if (wantedState >= CLASS_WRITTEN && getState() < CLASS_WRITTEN)
           { 
-            ModuleManager manager = ModuleManager.getInstance();
-            outputClass(manager.getCompilationDirectory());
+            outputClass(ModuleManager.getInstance().getCompilationDirectory());
             setState(CLASS_WRITTEN);
           }
       }
@@ -2037,25 +2004,25 @@ public class Compilation implements SourceLocator
 	generateConstructor(moduleClass, null);
       }
 
-    curClass = module.type;
+    curClass = module.compiledType;
     int arg_count;
     LambdaExp saveLambda = curLambda;
     curLambda = module;
     Type[] arg_types;
-    if (module.isHandlingTailCalls())
+    if (module.isHandlingTailCalls()) // Is this ever false?
       {
 	arg_count = 1;
 	arg_types = new Type[1];
 	arg_types[0] = typeCallContext;
       }
     else if (module.min_args != module.max_args || module.min_args > 4)
-      {
+      { // Likely dead code.
 	arg_count = 1;
 	arg_types = new Type[1];
 	arg_types[0] = new ArrayType (typeObject);
       }
     else
-      {
+      { // Likely dead code.
 	arg_count = module.min_args;
 	arg_types = new Type[arg_count];
 	for (int i = arg_count;  --i >= 0; )
@@ -2078,7 +2045,7 @@ public class Compilation implements SourceLocator
     code = getCode();
     // if (usingCPStyle())   code.addParamLocals();
 
-    thisDecl = method.getStaticFlag() ? null : module.declareThis(module.type);
+    thisDecl = method.getStaticFlag() ? null : module.declareThis(module.compiledType);
     module.closureEnv = module.thisVariable;
     module.heapFrame = module.isStatic() ? null : module.thisVariable;
     module.allocChildClasses(this);
@@ -2127,14 +2094,17 @@ public class Compilation implements SourceLocator
 
 	if (staticModule)
 	  {
-	    generateConstructor (module);
+            if (! module.getFlag(ModuleExp.USE_DEFINED_CLASS))
+              generateConstructor (module);
 
 	    code.emitNew(moduleClass);
 	    code.emitDup(moduleClass);
 	    code.emitInvokeSpecial(moduleClass.constructor);
+            // The $instance field needs to be Public so
+            // ModuleContext.findInstance can find it.
 	    moduleInstanceMainField
 	      = moduleClass.addField("$instance", moduleClass,
-				     Access.STATIC|Access.PUBLIC|Access.FINAL);
+				     Access.STATIC|Access.PUBLIC);
 	    code.emitPutStatic(moduleInstanceMainField);
 	  }
         Initializer init;
@@ -2151,8 +2121,8 @@ public class Compilation implements SourceLocator
 	  }
 	code.emitReturn();
 
-	if (moduleClass != mainClass
-	    && ! staticModule && ! generateMain && ! immediate)
+	if (moduleClass != mainClass && ! staticModule
+            && curClass.getSuperclass().getDeclaredMethod("run", 0) == null)
 	  {
 	    // Compare the run methods in ModuleBody.
 	    method = curClass.addMethod("run", Access.PUBLIC,
@@ -2269,7 +2239,13 @@ public class Compilation implements SourceLocator
 	    code.emitDup(curClass);
 	    code.emitInvokeSpecial(curClass.constructor);
 	  }
-	code.emitInvokeVirtual(typeModuleBody.getDeclaredMethod("runAsMain", 0));
+        Method runAsMainMethod = null;
+        ClassType superClass = curClass.getSuperclass();
+        if (superClass != typeModuleBody)
+           runAsMainMethod = superClass.getDeclaredMethod("runAsMain", 0);
+        if (runAsMainMethod == null)
+            runAsMainMethod = typeModuleBody.getDeclaredMethod("runAsMain", 1);
+        code.emitInvoke(runAsMainMethod);
 	code.emitReturn();
       }
 
@@ -2340,16 +2316,15 @@ public class Compilation implements SourceLocator
                 {
                   // If the source path was relative, emit it as relative.
                   // But make it relative to the compilation directory,
-                  // to allows sources to be moved along with binaries.
-                  char sep = File.separatorChar;
-                  String path = manager.getCompilationDirectory();
-                  path = path + mainPrefix.replace('.', sep);
-                  path = Path.toURL(path).toString();
+                  // to allow sources to be moved along with binaries.
+                  String path = Path.toURL(manager.getCompilationDirectory())
+                      + mainPrefix.replace('.', '/');
                   int plen = path.length();
-                  if (plen > 0 && path.charAt(plen-1) != sep)
-                    path = path + sep;
-                  moduleSource
-                    = Path.relativize(mi.getSourceAbsPathname(), path);
+                  if (plen > 0 && path.charAt(plen-1) != '/')
+                    path = path + '/';
+                  String sourcePath =
+                      Path.toURL(mi.getSourceAbsPathname()).toString();
+                  moduleSource = Path.relativize(sourcePath, path);
                 }
               catch (Throwable ex)
                 {
@@ -2441,7 +2416,7 @@ public class Compilation implements SourceLocator
    */
   public void mustCompileHere ()
   {
-    if (! mustCompile && ! ModuleExp.compilerAvailable)
+    if (! mustCompile && ! ModuleExp.compilerAvailable())
       error('e', "this expression must be compiled, but compiler is unavailable");
     mustCompile = true;
   }
@@ -2517,7 +2492,7 @@ public class Compilation implements SourceLocator
   {
     if (! mustCompile
         && (scope.mustCompile()
-            || (ModuleExp.compilerAvailable
+            || (ModuleExp.compilerAvailable()
                 // We set mustCompile if we see a LambdaExp - not because
                 // we must but because it is usually desirable.
                 && scope instanceof LambdaExp
@@ -2562,6 +2537,21 @@ public class Compilation implements SourceLocator
       loader.addClass((ClassType) type);
   }
 
+    /** Set moudle name - which sets name of generated class. */
+    public void setModuleName(String name) {
+        getModule().setName(name);
+    }
+
+    /** Generate and set unique module name suitable for an interactive session. */
+    public void setInteractiveName() {
+        setModuleName(ModuleManager.getInstance().getNewInteractiveName());
+    }
+
+    /** Generate and set unique module name suitable for a call to eval. */
+    public void setEvalName() {
+        setModuleName(ModuleManager.getInstance().getNewEvalName());
+    }
+
   public SourceMessages getMessages() { return messages; }
   public void setMessages (SourceMessages messages)
   { this.messages = messages; }
@@ -2656,34 +2646,45 @@ public class Compilation implements SourceLocator
 
   public void letStart ()
   {
-    pushScope(new LetExp(null));
+    pushScope(new LetExp());
   }
 
   public Declaration letVariable (Object name, Type type, Expression init)
   {
-    LetExp let = (LetExp) current_scope;
-    Declaration decl = let.addDeclaration(name, type);
-    decl.noteValue(init);
+    Declaration decl = new Declaration(name, type);
+    letVariable(decl, init);
     return decl;
   }
 
+  public void letVariable (Declaration decl, Expression init)
+  {
+    LetExp let = (LetExp) current_scope;
+    let.add(decl);
+    decl.setInitValue(init);
+  }
+
   public void letEnter ()
   {
     LetExp let = (LetExp) current_scope;
-    int ndecls = let.countDecls();
-    Expression[] inits = new Expression[ndecls];
-    int i = 0;
+    // Set a flag which letDone uses to check if letEnter has been called.
+    let.setFlag(Expression.VALIDATED);
     for (Declaration decl = let.firstDecl();
-	 decl != null;
-	 decl = decl.nextDecl())
-      inits[i++] = decl.getValue();
-    let.inits = inits;
+	 decl != null;  decl = decl.nextDecl())
+      {
+        Expression init = decl.getInitValue();
+        if (init != QuoteExp.undefined_exp)
+          decl.noteValueFromLet(let);
+      }
     lexical.push(let);
   }
 
   public LetExp letDone (Expression body)
   {
     LetExp let = (LetExp) current_scope;
+    if (let.getFlag(Expression.VALIDATED))
+      let.setFlag(false, Expression.VALIDATED);
+    else
+      letEnter();
     let.body = body;
     pop(let);
     return let;
@@ -2699,12 +2700,14 @@ public class Compilation implements SourceLocator
    * (We could make this implied by the first loopVaribale call ???) */
   public void loopStart()
   {
+    if (exprStack == null)
+      exprStack = new Stack<Expression>();
     LambdaExp loopLambda = new LambdaExp();
-    Expression[] inits = { loopLambda };
-    LetExp let = new LetExp(inits);
+    LetExp let = new LetExp();
     String fname = "%do%loop";
     Declaration fdecl = let.addDeclaration(fname);
-    fdecl.noteValue(loopLambda);
+    fdecl.setInitValue(loopLambda);
+    fdecl.noteValueFromLet(let);
     loopLambda.setName(fname);
     let.outer = current_scope;
     loopLambda.outer = let;
@@ -2716,8 +2719,6 @@ public class Compilation implements SourceLocator
     checkLoop();
     LambdaExp loopLambda = (LambdaExp) current_scope;
     Declaration decl = loopLambda.addDeclaration(name, type);
-    if (exprStack == null)
-      exprStack = new Stack<Expression>();
     exprStack.push(init);
     loopLambda.min_args++;
     return decl;
@@ -2775,6 +2776,11 @@ public class Compilation implements SourceLocator
     return loopRepeat(args);
   }
 
+  public QuoteExp makeQuoteExp (Object value)
+  {
+    return QuoteExp.getInstance(value, this);
+  }
+
   /**
    * Convenience method to make an Expression that coerces a value.
    * @param value to be coerced
