@@ -1,10 +1,16 @@
-// Copyright (c) 2003, 2009, 2010  Per M.A. Bothner.
+// Copyright (c) 2003, 2009, 2010, 2011  Per M.A. Bothner.
 // This is free software;  for terms and warranty disclaimer see ./COPYING.
 
 package gnu.expr;
 import gnu.bytecode.*;
 import gnu.mapping.*;
 import gnu.text.SourceLocator;
+import gnu.math.*;
+import java.util.ArrayList;
+import java.util.List;
+import java.lang.reflect.Proxy;
+import java.lang.annotation.ElementType;
+import gnu.kawa.functions.*;
 
 /**
  * The static information associated with a local variable binding.
@@ -71,9 +77,95 @@ public class Declaration
   public final Type getType()
   {
     if (type == null)
-      setType(Type.objectType);
+      {
+        Type t = Type.objectType;
+        type = t;  // To protect against possible cycles.
+        if (! hasUnknownValue() && nvalues > 0)
+          {
+            int arithCount = 0;
+            for (int i = 0;  i < nvalues;  i++)
+              {
+                Expression vi = values[i].getValue(this);
+                boolean arithOp = false;
+                if (vi != null && ! (vi.getFlag(Expression.VALIDATED)))
+                  {
+                    // Special handling of arithmetic assignments like
+                    // '(set! D (+ D V))', where D is this Declaration.
+                    // We can then use the type of V, instead of the unknown
+                    // type of (+ D V), to inductively determine the type of D.
+                    Expression oparg = checkArithStepOp(vi);
+                    if (oparg != null)
+                      {
+                        // oparg is the expression D above.
+                        arithOp = true;
+                        arithCount++;
+                        vi = oparg;
+                      }
+                  }
+                Declaration d;
+                if ((vi != null && vi.getFlag(Expression.VALIDATED))
+                    || vi instanceof LambdaExp
+                    || vi instanceof QuoteExp
+                    || (vi instanceof ReferenceExp
+                        && (d = ((ReferenceExp) vi).getBinding()) != null
+                        && d.type != null))
+                  {
+                    Type vt = vi.getType();
+                    if (i == 0)
+                      t = vt;
+                    else if (arithOp)
+                      t = CompileArith.combineType(t, vt);
+                    else
+                      t = Language.unionType(t, vt);
+                  }
+                else
+                  { 
+                    t = Type.objectType;
+                    break;
+                  }
+              }
+          }
+        setType(t);
+      }
     return type;
   }
+
+  private Expression checkArithStepOp (Expression exp)
+  {
+    if (! (exp instanceof ApplyExp))
+      return null;
+    ApplyExp aexp = (ApplyExp) exp;
+    Expression func = aexp.getFunction();
+    Compilation comp = Compilation.getCurrent();
+    boolean isApplyFunc = comp.isSimpleApplyFunction(func);
+    if (aexp.getArgCount() == (isApplyFunc ? 3 : 2))
+      {
+        if (isApplyFunc)
+          func = aexp.getArg(0);
+        Declaration fdecl;
+        if (func instanceof ReferenceExp
+            && (fdecl = ((ReferenceExp) func).getBinding()) != null
+            && fdecl.getValue() != null)
+          {
+            Object proc = fdecl.getValue().valueIfConstant();
+            if (proc instanceof AddOp || proc instanceof MultiplyOp)
+              {
+                Expression arg1 = aexp.getArg(isApplyFunc ? 1 : 0);
+                Expression arg2 = aexp.getArg(isApplyFunc ? 2 : 1);
+                if (arg1 instanceof ReferenceExp
+                    && ((ReferenceExp) arg1).getBinding() == this)
+                  return arg2;
+                if (arg2 instanceof ReferenceExp
+                    && ((ReferenceExp) arg2).getBinding() == this)
+                  return arg1;
+              }
+          }
+      }
+    return null;
+  }
+
+  public Expression getTypeExpRaw () { return typeExp; }
+
   public final void setType(Type type)
   {
     this.type = type;
@@ -95,6 +187,12 @@ public class Declaration
     if (var != null) var.setType(t);
   }
 
+  public final void setType (Expression typeExp, Type type)
+  {
+    this.typeExp = typeExp;
+    this.type = type;
+  }
+
   public final String getName()
   {
     return symbol == null ? null : symbol instanceof Symbol ? ((Symbol) symbol).getName()
@@ -109,10 +207,10 @@ public class Declaration
   public final void setSymbol(Object symbol) { this.symbol = symbol; }
 
   /* Declarations in a ScopeExp are linked together in a linked list. */
-  Declaration next;
+  private Declaration next;
 
   public final Declaration nextDecl() { return next; }
-  public final void setNext(Declaration next) {  this.next = next; }
+  public final void setNext(Declaration next) { this.next = next; }
 
   /** Index in evalFrame for this scope, if interpreting. */
   int evalIndex;
@@ -164,9 +262,10 @@ public class Declaration
     if (target instanceof IgnoreTarget)
       return;
     Declaration owner = access == null ? null : access.contextDecl();
-    if (isAlias() && value instanceof ReferenceExp)
+    Expression dvalue = getValueRaw();
+    if (isAlias() && dvalue instanceof ReferenceExp)
       {
-        ReferenceExp rexp = (ReferenceExp) value;
+        ReferenceExp rexp = (ReferenceExp) dvalue;
         Declaration orig = rexp.binding;
         if (orig != null
             && ((flags & ReferenceExp.DONT_DEREFERENCE) == 0
@@ -197,12 +296,12 @@ public class Declaration
         boolean immediate = comp.immediate;
         if (field.getStaticFlag())
           {
-            ltype = ClassType.make("gnu.kawa.reflect.StaticFieldLocation");
+            ltype = Compilation.typeStaticFieldLocation;
             meth = ltype.getDeclaredMethod("make", immediate ? 1 : 2);
           }
         else
           {
-            ltype = ClassType.make("gnu.kawa.reflect.FieldLocation");
+            ltype = Compilation.typeFieldLocation;
             meth = ltype.getDeclaredMethod("make", immediate ? 2 : 3);
 
             loadOwningObject(owner, comp);
@@ -217,9 +316,25 @@ public class Declaration
         code.emitInvokeStatic(meth);
         rtype = ltype;
       }
+    else if (getFlag(ALLOCATE_ON_STACK))
+      {
+        int SP = code.getSP();
+        if (access.getFlag(ReferenceExp.ALLOCATE_ON_STACK_LAST))
+          ;
+        else if (SP == evalIndex)
+          code.emitDup();
+        else if (SP == evalIndex+1)
+          {
+            code.emitSwap();
+            code.emitDupX();
+          }
+        else
+          throw new InternalError("allocate-on-stack mismatch");
+      }
     else
       {
         Object val;
+        Expression value = getValueRaw();
         if (field != null)
           {
             comp.usedClass(field.getDeclaringClass());
@@ -251,8 +366,7 @@ public class Declaration
             return;
           }
         else if (value != QuoteExp.undefined_exp && ignorable()
-                 && ! (value instanceof LambdaExp
-                       && ((LambdaExp) value).outer instanceof ModuleExp))
+                 && ! (value instanceof LambdaExp))
           {
             value.compile(comp, target);
             return;
@@ -266,7 +380,7 @@ public class Declaration
                 && (cl = (ClassExp) context).isMakingClassPair())
               {
                 String getName = ClassExp.slotToMethodName("get", getName());
-                Method getter = cl.type.getDeclaredMethod(getName, 0);
+                Method getter = cl.compiledType.getDeclaredMethod(getName, 0);
                 cl.loadHeapFrame(comp);
                 code.emitInvoke(getter);
               }
@@ -351,74 +465,6 @@ public class Declaration
       }
   }
 
-  /** If non-null, the single expression used to set this variable.
-   * If the variable can be set more than once, then value is null. */
-  protected Expression value = QuoteExp.undefined_exp;
-
-  /** The value of this <code>Declaration</code>, if known.
-   * Usually the expression used to initialize the <code>Declaration</code>,
-   * or null if the <code>Declaration</code> can be assigned a different
-   * value after initialization.  Note that this is the semantic value: If the
-   * <code>INDIRECT_LOCATION</code> is set, then <code>getValue</code> is the
-   * value <em>after</em> de-referencing the resulting <code>Location</code>.
-   * An exception is if <code>isAlias()</code>; in that case
-   * <code>getValue()</code> is an expression yielding a <code>Location</code>
-   * which needs to be de-referenced to get this <code>Declaration</code>'s
-   * actual value.
-   */
-  public final Expression getValue()
-  {
-    if (value == QuoteExp.undefined_exp)
-      {
-        if (field != null
-            && ((field.getModifiers() & Access.STATIC+Access.FINAL)
-                == Access.STATIC+Access.FINAL)
-            && ! isIndirectBinding())
-          {
-            try
-              {
-                value = new QuoteExp(field.getReflectField().get(null));
-              }
-            catch (Throwable ex)
-              {
-              }
-          }
-      }
-    else if (value instanceof QuoteExp && getFlag(TYPE_SPECIFIED)
-             && value.getType() != type)
-      {
-        try
-          {
-            Object val = ((QuoteExp) value).getValue();
-            Type t = getType();
-            value = new QuoteExp(t.coerceFromObject(val), t);
-          }
-        catch (Throwable ex)
-          {
-          }
-      }
-    return value;
-  }
-
-  /** Set the value associated with this Declaration.
-   * Most code should use noteValue instead. */
-  public final void setValue(Expression value) { this.value = value; }
-
-  /** If getValue() is a constant, return the constant value, otherwise null. */
-  public final Object getConstantValue()
-  {
-    Object v = getValue();
-    if (! (v instanceof QuoteExp) || v == QuoteExp.undefined_exp)
-      return null;
-    return ((QuoteExp) v).getValue();
-  }
-
-  public final boolean hasConstantValue ()
-  {
-    Object v = getValue();
-    return (v instanceof QuoteExp) && v != QuoteExp.undefined_exp;
-  }
-
   boolean shouldEarlyInit ()
   {
     return getFlag(EARLY_INIT) || isCompiletimeConstant ();
@@ -517,13 +563,22 @@ public class Declaration
   public static final long TRANSIENT_ACCESS = 0x100000000l;
   public static final long ENUM_ACCESS = 0x200000000l;
   public static final long FINAL_ACCESS = 0x400000000l;
+  public static final long ABSTRACT_ACCESS = 0x800000000l;
   public static final long CLASS_ACCESS_FLAGS =
-    PRIVATE_ACCESS|PROTECTED_ACCESS|ENUM_ACCESS|FINAL_ACCESS;
+    PRIVATE_ACCESS|PROTECTED_ACCESS|ENUM_ACCESS|FINAL_ACCESS|ABSTRACT_ACCESS;
   public static final long FIELD_ACCESS_FLAGS = PRIVATE_ACCESS|PROTECTED_ACCESS|
     PUBLIC_ACCESS|PACKAGE_ACCESS|VOLATILE_ACCESS|TRANSIENT_ACCESS|
     ENUM_ACCESS|FINAL_ACCESS;
   public static final long METHOD_ACCESS_FLAGS = PRIVATE_ACCESS
     |PROTECTED_ACCESS|PUBLIC_ACCESS|PACKAGE_ACCESS|FINAL_ACCESS;
+  public static final long MAYBE_UNINITIALIZED_ACCESS = 0x1000000000l;
+  /** Allocate variable on JVM stack as an optimization.
+   * This means load is implemented as a dup instruction.
+   * (This is no faster on decent JVMs, but the bytecode is more compact.)
+   * Note this may cause an InternalError if this is loaded when the
+   * JVM stack has grown since the variable was initialized.
+   */
+  public static final long ALLOCATE_ON_STACK = 0x2000000000l;
 
   protected long flags = IS_SIMPLE;
 
@@ -548,6 +603,10 @@ public class Declaration
 
   public final boolean isPrivate() { return (flags & PRIVATE) != 0; }
 
+    public final boolean isModuleLocal() {
+        return ! isPublic() && ! needsExternalAccess();
+    }
+
   public final void setPrivate(boolean isPrivate)
   {
     setFlag(isPrivate, PRIVATE);
@@ -592,6 +651,16 @@ public class Declaration
 
   public final void setProcedureDecl (boolean val) { setFlag(val, PROCEDURE); }
 
+  public final boolean isClassMethod ()
+  {
+    return (flags & FIELD_OR_METHOD+PROCEDURE) == FIELD_OR_METHOD+PROCEDURE;
+  }
+
+  public final boolean isClassField ()
+  {
+    return (flags & FIELD_OR_METHOD+PROCEDURE) == FIELD_OR_METHOD;
+  }
+
   public final boolean isNamespaceDecl ()
   {
     return (flags & IS_NAMESPACE_PREFIX) != 0;
@@ -660,6 +729,11 @@ public class Declaration
     return symbol == ThisExp.THIS_NAME;
   }
 
+  public boolean mayBeAccessedUninitialized ()
+  {
+    return getFlag(MAYBE_UNINITIALIZED_ACCESS);
+  }
+
   /** True if we never need to access this declaration. */
   // rename to isAccessed?
   public boolean ignorable()
@@ -677,16 +751,6 @@ public class Declaration
     return ! lexp.isHandlingTailCalls() || lexp.getInlineOnly();
   }
 
-  /** Does this variable need to be initialized or is default ok
-   */
-  public boolean needsInit()
-  {
-    // This is a kludge.  Ideally, we should do some data-flow analysis.
-    // But at least it makes sure require'd variables are not initialized.
-    return ! ignorable()
-      && ! (value == QuoteExp.nullExp && base != null);
-  }
-
   public boolean isStatic()
   {
     if (field != null)
@@ -711,30 +775,37 @@ public class Declaration
     return decl == null || decl.getFlag(IS_UNKNOWN);
   }
 
+  int numReferences;
+
   /** List of ApplyExp where this declaration is the function called.
    * The applications are chained using their nextCall fields.
    * The chain is not built if STATIC_SPECIFIED. */
   public ApplyExp firstCall;
 
-  public void noteValue (Expression value)
+  /** Add a call to the list headed by {@code firstCall}. */
+  public void addCaller (ApplyExp exp)
   {
-    // We allow assigning a real value after undefined ...
-    if (this.value == QuoteExp.undefined_exp)
-      {
-	if (value instanceof LambdaExp)
-	  ((LambdaExp) value).nameDecl = this;
-	this.value = value;
-      }
-    else if (this.value != value)
-      {
-	if (this.value instanceof LambdaExp) 
-	  ((LambdaExp) this.value).nameDecl = null;
-	this.value = null;
-      }
+    exp.nextCall = firstCall;
+    firstCall = exp;
   }
 
-  protected Declaration()
+  /** Clear the list of callers headed by {@code firstCall}. */
+  public void clearCallList ()
   {
+    for (ApplyExp exp = firstCall; exp != null; )
+      {
+        ApplyExp next = exp.nextCall;
+        exp.nextCall = null;
+        exp = next;
+      }
+    firstCall = null;
+  }
+
+  public Declaration (Object name, Type type)
+  {
+    setName(name);
+    if (type != null)
+      setType(type);
   }
 
   public Declaration (Variable var)
@@ -745,13 +816,7 @@ public class Declaration
 
   public Declaration (Object name)
   {
-    setName(name);
-  }
-
-  public Declaration (Object name, Type type)
-  {
-    setName(name);
-    setType(type);
+    this(name, (Type) null);
   }
 
   public Declaration (Object name, Field field)
@@ -761,6 +826,53 @@ public class Declaration
     setSimple(false);
   }
 
+  List<Expression> annotations;
+
+  /** The number of annotations associated with the declaration. */
+  public int numAnnotations ()
+  {
+    return annotations == null ? 0 : annotations.size();
+  }
+
+  /** Indexed get of one of the annotations associated with this declaration. */
+  public Expression getAnnotation (int i)
+  {
+    return annotations.get(i);
+  }
+
+  /** Replace one of the annotations associated with this declaration. */
+  public void setAnnotation (int i, Expression ann)
+  {
+    annotations.set(i, ann);
+  }
+
+  /** Add an annotation to the set of our annotations.
+   * @param exp A constant-valued expression that evaluates to an Annotation.
+   */
+  public void addAnnotation (Expression exp)
+  {
+    if (annotations == null)
+      annotations = new ArrayList<Expression>();
+    annotations.add(exp);
+  }
+
+  public void compileAnnotations (AttrContainer container, ElementType etype)
+  {
+    if (container == null)
+      return;
+    int n = numAnnotations();
+    for (int i = 0;  i < n;  i++)
+      {
+        Object ann = getAnnotation(i).valueIfConstant();
+        if (ann != null)
+          {
+            AnnotationEntry ae = (AnnotationEntry) Proxy.getInvocationHandler(ann);
+            if (container != null && ae.hasTarget(etype))
+              RuntimeAnnotationsAttr.maybeAddAnnotation(container, ae);
+          }
+      }
+  }
+
   Method makeLocationMethod = null;
 
   /** Create a Location object, given that isIndirectBinding().
@@ -868,7 +980,20 @@ public class Declaration
   {
     StringBuffer sbuf = new StringBuffer();
     printInfo(sbuf);
+    out.startLogicalBlock("", "", 2);
     out.print(sbuf.toString());
+    int numAnnotations = numAnnotations();
+    if (numAnnotations > 0)
+      {
+        out.writeSpaceLinear();
+        out.print("Annotations:");
+        for (int i = 0;  i < numAnnotations;  i++)
+          {
+            out.writeSpaceLinear();
+            annotations.get(i).print(out);
+          }
+      }
+    out.endLogicalBlock("");
   }
 
   public void printInfo(StringBuffer sbuf)
@@ -906,13 +1031,13 @@ public class Declaration
     if (ignorable())
       sbuf.append("(ignorable)");
     Expression tx = typeExp;
-    Type t = getType();
+    Type t = type;
     if (tx != null && ! (tx instanceof QuoteExp))
       {
 	sbuf.append("::");
         sbuf.append(tx);
       }
-    else if (type != null && t != Type.pointer_type)
+    else if (t != null && t != Type.pointer_type)
       {
 	sbuf.append("::");
 	sbuf.append(t.getName());
@@ -994,7 +1119,9 @@ public class Declaration
       ftype = Compilation.typeLocation;
     if (! ignorable())
       {
-        String fname = getName();
+        String dname = getName();
+        String fname = dname;
+        boolean haveName = fname != null;
         int nlength;
         if (fname==null)
           {
@@ -1005,15 +1132,56 @@ public class Declaration
           {
             fname = Compilation.mangleNameIfNeeded(fname);
             if (getFlag(IS_UNKNOWN))
-              fname = UNKNOWN_PREFIX + fname;
+              {
+                fname = UNKNOWN_PREFIX + fname;
+                haveName = false;
+              }
             if (external_access && ! getFlag(Declaration.MODULE_REFERENCE))
-              fname = PRIVATE_PREFIX + fname;
-            nlength = fname.length();
+              {
+                fname = PRIVATE_PREFIX + fname; 
+                haveName = false;
+              }
+           nlength = fname.length();
           }
         int counter = 0;
         while (frameType.getDeclaredField(fname) != null)
           fname = fname.substring(0, nlength) + '$' + (++ counter);
         field = frameType.addField (fname, ftype, fflags);
+        if (haveName)
+          {
+            Object fsymbol = getSymbol();
+            String uri, prefix;
+            boolean haveUri, havePrefix; 
+            // If name is a non-simple Symbol (i.e. with uri or prefix)
+            // or the field name doesn't demangle to name, then emit
+            // a SourceName annotation so we can recover the correct name.
+            if (fsymbol instanceof Symbol)
+              {
+                uri = ((Symbol) fsymbol).getNamespaceURI();
+                prefix = ((Symbol) fsymbol).getPrefix();
+                if (uri == null)
+                  uri = "";
+                haveUri = ! "".equals(uri);
+                havePrefix = ! "".equals(prefix);
+              }
+            else
+              {
+                uri = prefix = "";
+                haveUri = havePrefix = false;
+              }
+            // FIXME should optimize if uri == module.getNamespaceUri()
+            if (haveUri || havePrefix
+                || ! Compilation.demangleName(fname, true).equals(dname))
+              {
+                AnnotationEntry ae = new AnnotationEntry(ClassType.make("gnu.expr.SourceName"));
+                ae.addMember("name", dname, Type.javalangStringType);
+                if (haveUri)
+                  ae.addMember("uri", uri, Type.javalangStringType);
+                if (havePrefix)
+                  ae.addMember("prefix", prefix, Type.javalangStringType);
+                RuntimeAnnotationsAttr.maybeAddAnnotation(field, ae);
+              }
+          }
         if (value instanceof QuoteExp)
           {
             Object val = ((QuoteExp) value).getValue();
@@ -1150,4 +1318,232 @@ public class Declaration
       }
     return null;
   }
+
+  /** Get the "initial value" expression.
+   * This is used for the initializing value in a LetExp,
+   * a parameter's default value, or for pattern-matching.
+   */
+  public Expression getInitValue() { return initValue; }
+  public void setInitValue(Expression init) { this.initValue = init; }
+  private Expression initValue;
+
+  ValueSource values[];
+  int nvalues;
+  static final ValueSource unknownValueInstance =
+    new ValueSource(ValueSource.UNKNOWN_KIND, null, 0);
+  static final ValueSource[] unknownValueValues = { unknownValueInstance };
+
+  public boolean hasUnknownValue () { return values == Declaration.unknownValueValues; }
+
+  /** The value of this <code>Declaration</code>, if known.
+   * Usually the expression used to initialize the <code>Declaration</code>,
+   * or null if the <code>Declaration</code> can be assigned a different
+   * value after initialization.  Note that this is the semantic value: If the
+   * <code>INDIRECT_LOCATION</code> is set, then <code>getValue</code> is the
+   * value <em>after</em> de-referencing the resulting <code>Location</code>.
+   * An exception is if <code>isAlias()</code>; in that case
+   * <code>getValue()</code> is an expression yielding a <code>Location</code>
+   * which needs to be de-referenced to get this <code>Declaration</code>'s
+   * actual value.
+   */
+  public final Expression getValue()
+  {
+    if (nvalues == 0)
+      {
+        if (field != null
+            && field.getDeclaringClass().isExisting()
+            && ((field.getModifiers() & Access.STATIC+Access.FINAL)
+                == Access.STATIC+Access.FINAL)
+            && ! isIndirectBinding())
+          {
+            try
+              {
+                Expression value = new QuoteExp(field.getReflectField().get(null));
+                noteValue(value);
+                return value;
+              }
+            catch (Throwable ex)
+              {
+              }
+          }
+        return QuoteExp.undefined_exp;
+      }
+    if (nvalues == 1)
+      return values[0].getValue(this);
+    return null;
+  }
+
+  public Expression getValueRaw ()
+  {
+    if (nvalues == 0)
+      return QuoteExp.undefined_exp;
+    if (nvalues == 1)
+      return values[0].getValue(this);
+    return null;
+  }
+
+  /** Set the value associated with this Declaration.
+   * Most code should use noteValue instead. */
+  public final void setValue(Expression value)
+  {
+    values = null;
+    nvalues = 0;
+    noteValue(value);
+  }
+
+  /** If getValue() is a constant, return the constant value, otherwise null. */
+  public final Object getConstantValue()
+  {
+    Object v = getValue();
+    if (! (v instanceof QuoteExp) || v == QuoteExp.undefined_exp)
+      return null;
+    return ((QuoteExp) v).getValue();
+  }
+
+  public final boolean hasConstantValue ()
+  {
+    Object v = getValue();
+    return (v instanceof QuoteExp) && v != QuoteExp.undefined_exp;
+  }
+
+  public LambdaExp getLambdaValue ()
+  {
+    if (! isAlias() && nvalues == 1)
+      {
+        Expression val = values[0].getValue(this);
+        if (val != null && val.getClass() == LambdaExp.class)
+          return (LambdaExp) val;
+      }
+    return null;
+  }
+
+  public void noteValue (Expression value)
+  {
+    checkNameDecl(value);
+    if (value == null)
+      noteValueUnknown();
+    else if (values != unknownValueValues)
+      noteValue(new ValueSource(ValueSource.GENERAL_KIND, value, 0));
+  }
+
+  void noteValue (ValueSource value)
+  {
+    if (values == unknownValueValues)
+      throw new InternalError();
+    if (values == null)
+      values = new ValueSource[4];
+    else if (nvalues >= values.length)
+      {
+        ValueSource[] tmp = new ValueSource[2 * nvalues];
+        System.arraycopy(values, 0, tmp, 0, nvalues);
+        values = tmp;
+      }
+    values[nvalues++] = value;
+  }
+
+  public void noteValueConstant (Object value)
+  {
+    if (values != unknownValueValues)
+      {
+        noteValue(new QuoteExp(value));
+      }
+  }
+
+  public void noteValueUnknown ()
+  {
+    checkNameDecl(null);
+    values = unknownValueValues;
+    nvalues = 1;
+  }
+
+  public void noteValueFromSet (SetExp setter)
+  {
+    if (values != unknownValueValues)
+      {
+        checkNameDecl(setter.new_value);
+        setter.valueIndex = nvalues;
+        noteValue(new ValueSource(ValueSource.SET_RHS_KIND, setter, 0));
+      }
+  }
+
+  public void noteValueFromLet (LetExp letter)
+  {
+    Expression init = getInitValue();
+    if (init != QuoteExp.undefined_exp && values != unknownValueValues)
+      {
+        checkNameDecl(init);
+        noteValue(new ValueSource(ValueSource.LET_INIT_KIND, letter, 0));
+      }
+  }
+
+  public void noteValueFromApply (ApplyExp app, int index)
+  {
+    if (values != unknownValueValues)
+      noteValue(new ValueSource(ValueSource.APPLY_KIND, app, index));
+  }
+
+  private void checkNameDecl (Expression value)
+  {
+    if (nvalues == 1 && values[0].kind == ValueSource.GENERAL_KIND)
+      {
+        Expression old = values[0].base;
+        if (old == value)
+          return;
+        if (old instanceof LambdaExp)
+          ((LambdaExp) old).nameDecl = null;
+      }
+    if (nvalues == 0 && value instanceof LambdaExp)
+      ((LambdaExp) value).nameDecl = this;
+  }
+
+  public static class ValueSource
+  {
+    static final int UNKNOWN_KIND = 0;
+    static final int GENERAL_KIND = 1;
+    static final int SET_RHS_KIND = 2;
+    static final int LET_INIT_KIND = 3;
+    static final int APPLY_KIND = 4;
+    public int kind;
+    public Expression base;
+    public int index;
+
+    ValueSource (int kind, Expression base, int index)
+    {
+      this.kind = kind;
+      this.base = base;
+      this.index = index;
+    }
+
+    Expression getValue (Declaration decl)
+    {
+      switch (kind) 
+        {
+        case UNKNOWN_KIND:
+          return null;
+        case GENERAL_KIND:
+          return base;
+        case SET_RHS_KIND:
+          return ((SetExp) base).new_value;
+        case LET_INIT_KIND:
+          return decl.getInitValue();
+        case APPLY_KIND:
+          ApplyExp app = (ApplyExp) base;
+          int i = index;
+          // If a function is called via an apply-function, the latter
+          // might distribute a multiple-valued argument among multiple
+          // parameters.  Punt on that for now.
+          Compilation comp = Compilation.getCurrent();
+          Expression afunc = app.getFunction();
+          if (comp.isSimpleApplyFunction(afunc))
+              i++;
+          else if (comp.isApplyFunction(afunc))
+              return null;
+          if (i >= app.getArgCount())
+              return null;
+          return app.getArg(i);
+        default:
+          throw new Error();
+        }
+    }
+  }
 }
