@@ -16,11 +16,13 @@ import gnu.kawa.functions.Convert;
 public class LambdaExp extends ScopeExp
 {
   public Expression body;
-  /** Minumnum number of parameters.
+  /** Minimum number of parameters.
    * Does not count implicit isThisParameter(). */
   public int min_args;
   /** Maximum number of actual arguments;  -1 if variable. */
   public int max_args;
+  /** Number of optional arguments, not counting keyword arguments. */
+  public int opt_args;
 
   /** Set of visible top-level LambdaExps that need apply methods. */
   Vector applyMethods;
@@ -31,12 +33,18 @@ public class LambdaExp extends ScopeExp
   private Declaration firstArgsArrayArg;
 
   public Keyword[] keywords;
-  public Expression[] defaultArgs;
 
   /** A list of Declarations, chained using Declaration's nextCapturedVar.
     * All the Declarations are allocated in the current heapFrame. */
   Declaration capturedVars;
 
+  /** Linked link of references to sibling declarations.
+   * Chained using {@see ReferenceExp#siblingReferencesNext} links.
+   * I.e. References to declarations external to this LambdaExp,
+   * but local to the outer LambdaExp.
+   */
+  ReferenceExp siblingReferences;
+
   public void capture (Declaration decl)
   {
     if (decl.isSimple())
@@ -71,7 +79,8 @@ public class LambdaExp extends ScopeExp
   public LambdaExp nextSibling;
 
   /** A magic value to indicate there is no unique return continuation. */
-  final static ApplyExp unknownContinuation = new ApplyExp ((Expression) null, null);
+  final static ApplyExp unknownContinuation
+      = new ApplyExp ((Expression) null, (Expression[]) null);
 
   /** The unique call site that calls this lambda.
    * The value is null if no callers have been seen.
@@ -217,6 +226,10 @@ public class LambdaExp extends ScopeExp
     else flags &= ~CAN_CALL;
   }
 
+    public final void setCanAccess(boolean canRead) {
+        flags |= canRead ? CAN_READ : CAN_CALL;
+    }
+
   /** True if this is a method in an ClassExp. */
   public final boolean isClassMethod()
   { return (flags & CLASS_METHOD) != 0; }
@@ -241,54 +254,67 @@ public class LambdaExp extends ScopeExp
     return body == QuoteExp.abstractExp;
   }
 
-  /** Specify the calling convention used for this function.
-   * @return One of the CALL_WITH_xxx values in Compilation. */
-  public int getCallConvention ()
+  public boolean isNative ()
   {
-    if (isModuleBody())
-      return ((Compilation.defaultCallConvention
-	      >= Compilation.CALL_WITH_CONSUMER)
-	      ? Compilation.defaultCallConvention
-	      : Compilation.CALL_WITH_CONSUMER);
-    if (isClassMethod())
-      return Compilation.CALL_WITH_RETURN;
-    return ((Compilation.defaultCallConvention
-	     != Compilation.CALL_WITH_UNSPECIFIED)
-	    ? Compilation.defaultCallConvention
-	    : Compilation.CALL_WITH_RETURN);
+    return body == QuoteExp.nativeExp;
   }
 
+  int callConvention;
+  /** The calling convention used for this function.
+   * It is derived from Compilation's currentCallConvention.
+   * @return One of the CALL_WITH_xxx values in Compilation. */
+  public int getCallConvention() { return callConvention; }
+  public void setCallConvention(Compilation comp)
+  {
+    if (isClassMethod())
+      callConvention = Compilation.CALL_WITH_RETURN;
+    else
+      {
+        int defaultConvention = comp.currentCallConvention();
+        callConvention =
+            (defaultConvention < Compilation.CALL_WITH_CONSUMER
+             && isModuleBody())
+            ? Compilation.CALL_WITH_CONSUMER
+            : defaultConvention == Compilation.CALL_WITH_UNSPECIFIED
+            ? Compilation.CALL_WITH_RETURN
+            : defaultConvention;
+      }
+  }
+  public boolean usingCallContext()
+  { return getCallConvention() >= Compilation.CALL_WITH_CONSUMER; }
+
   public final boolean isHandlingTailCalls ()
   {
     return isModuleBody()
-      || (Compilation.defaultCallConvention >= Compilation.CALL_WITH_TAILCALLS
+      || (getCallConvention() >= Compilation.CALL_WITH_TAILCALLS
 	  && ! isClassMethod());
   }
 
   public final boolean variable_args () { return max_args < 0; }
 
-  ClassType type = Compilation.typeProcedure;
+  ClassType compiledType = Compilation.typeProcedure;
 
   /** Return the ClassType of the Procedure this is being compiled into. */
   protected ClassType getCompiledClassType(Compilation comp)
   {
-    if (type == Compilation.typeProcedure)
+    if (compiledType == Compilation.typeProcedure)
       throw new Error("internal error: getCompiledClassType");
-    return type;
+    return compiledType;
   }
 
-  public Type getType()
+  protected Type calculateType()
   {
-    return type;
+    return compiledType;
   }
 
   /** The ClassType generated for this class.
    * Only used for ClassExp (which overrides this method) or ModuleExp.
    */
-  public ClassType getClassType() { return type; }
+  public ClassType getClassType() { return compiledType; }
 
   public void setType (ClassType type)
   {
+    this.compiledType = type;
     this.type = type;
   }
 
@@ -370,6 +396,11 @@ public class LambdaExp extends ScopeExp
     return outer == null ? null : outer.currentLambda ();
   }
 
+  public LambdaExp outerLambdaOrCaller()
+  {
+    return getInlineOnly() ? inlineHome : outerLambda();
+  }
+
   /** Return the closest outer non-inlined LambdaExp. */
 
   public LambdaExp outerLambdaNotInline ()
@@ -427,13 +458,11 @@ public class LambdaExp extends ScopeExp
   {
     if (closureEnv == null && getNeedsClosureEnv())
       {
-	LambdaExp parent = outerLambda();
+        LambdaExp parent = outerLambdaOrCaller();
 	if (parent instanceof ClassExp)
 	  parent = parent.outerLambda();
-	Variable parentFrame = parent.heapFrame != null ?  parent.heapFrame
-	  : parent.closureEnv;
 	if (isClassMethod() && ! "*init*".equals(getName()))
-          closureEnv = declareThis(type);
+          closureEnv = declareThis(compiledType);
 	else if (parent.heapFrame == null && ! parent.getNeedsStaticLink()
 		 && ! (parent instanceof ModuleExp))
 	  closureEnv = null;
@@ -457,13 +486,18 @@ public class LambdaExp extends ScopeExp
 		closureEnv.setParameter(true);
 	      }
 	  }
-	else if (inlinedIn(parent))
-	  closureEnv = parentFrame;
 	else
-	  {
-	    closureEnv = new Variable("closureEnv", parentFrame.getType());
-	    getVarScope().addVariable(closureEnv);
-	  }
+          {
+            Variable parentFrame = parent.heapFrame != null ?  parent.heapFrame
+                : parent.closureEnv;
+            if (inlinedIn(parent))
+              closureEnv = parentFrame;
+            else
+              {
+                closureEnv = new Variable("closureEnv", parentFrame.getType());
+                getVarScope().addVariable(closureEnv);
+              }
+          }
       }
     return closureEnv;
   }
@@ -516,7 +550,7 @@ public class LambdaExp extends ScopeExp
             code.emitGetField(link);
             curType = (ClassType) link.getType();
           }
-        curLambda = curLambda.outerLambda();
+        curLambda = curLambda.outerLambdaOrCaller();
       }
   }
 
@@ -539,7 +573,7 @@ public class LambdaExp extends ScopeExp
     if (! getInlineOnly())
       {
 	if (comp.method.reachableHere()
-	    && (Compilation.defaultCallConvention < Compilation.CALL_WITH_TAILCALLS
+	    && (getCallConvention() < Compilation.CALL_WITH_TAILCALLS
 		|| isModuleBody() || isClassMethod() || isHandlingTailCalls()))
 	  code.emitReturn();
 	popScope(code);        // Undoes enterScope in allocParameters
@@ -562,10 +596,8 @@ public class LambdaExp extends ScopeExp
   public void generateApplyMethods(Compilation comp)
   {
     comp.generateMatchMethods(this);
-    if (Compilation.defaultCallConvention >= Compilation.CALL_WITH_CONSUMER)
-      comp.generateApplyMethodsWithContext(this);
-    else
-      comp.generateApplyMethodsWithoutContext(this);
+    comp.generateApplyMethodsWithContext(this);
+    comp.generateApplyMethodsWithoutContext(this);
   }
 
   Field allocFieldFor (Compilation comp)
@@ -664,6 +696,15 @@ public class LambdaExp extends ScopeExp
   {
     if (target instanceof IgnoreTarget)
       return;
+    if (getInlineOnly())
+      {
+        // Normally this shouldn't happen.  One case where it does
+        // is when passing an inline-only lambda as a parameter to a function
+        // that doesn't get fully inlined.  Cleaner would be to elide
+        // the ignored parameter.
+        QuoteExp.nullExp.compile(comp, target);
+        return;
+      }
     Type rtype;
     CodeAttr code = comp.getCode();
 
@@ -733,7 +774,8 @@ public class LambdaExp extends ScopeExp
       { LambdaExp outer = outerLambda();
 	rtype = Compilation.typeModuleMethod;
 	if ((flags & NO_FIELD) != 0
-	    || (comp.immediate && outer instanceof ModuleExp))
+            || (comp.immediate && outer instanceof ModuleExp
+                && comp.mainClass == comp.moduleClass))
 	  {
             if (primMethods == null)
               allocMethod(outerLambda(), comp);
@@ -778,7 +820,7 @@ public class LambdaExp extends ScopeExp
 	  return null;
 	if (exp instanceof ModuleExp
 	    || (exp instanceof ClassExp && getNeedsClosureEnv())
-	    || (exp instanceof LambdaExp
+	    || (exp instanceof LambdaExp 
 		&& ((LambdaExp) exp).heapFrame != null))
 	  return (LambdaExp) exp;
       }
@@ -806,7 +848,6 @@ public class LambdaExp extends ScopeExp
     LambdaExp outer = outerLambda();
 
     int key_args = keywords == null ? 0 : keywords.length;
-    int opt_args = defaultArgs == null ? 0 : defaultArgs.length - key_args;
     int numStubs =
       ((flags & DEFAULT_CAPTURES_ARG) != 0) ? 0 : opt_args;
     boolean varArgs = max_args < 0 || min_args + numStubs < max_args;
@@ -901,6 +942,8 @@ public class LambdaExp extends ScopeExp
       }
     if (ctype.isInterface() || isAbstract())
       mflags |= Access.ABSTRACT;
+    if (isNative())
+      mflags |= Access.NATIVE;
 
     // If a class method has unspecified parameter types, see if we
     // can "inherit" the parameter types from an inherited method.
@@ -1208,9 +1251,7 @@ public class LambdaExp extends ScopeExp
       {
 	if (! child.isClassGenerated() && ! child.getInlineOnly()
             && child.nameDecl != null)
-	  {
             child.allocMethod(this, comp);
-	  }
         if (child instanceof ClassExp)
           {
             ClassExp cl = (ClassExp) child;
@@ -1293,7 +1334,7 @@ public class LambdaExp extends ScopeExp
 	  }
 	else if (! inlinedIn(outerLambda()))
 	  {
-	    outerLambda().loadHeapFrame(comp);
+	    outerLambdaOrCaller().loadHeapFrame(comp);
 	    code.emitStore(closureEnv);
 	  }
       }
@@ -1343,12 +1384,11 @@ public class LambdaExp extends ScopeExp
     // For each non-artificial parameter, copy it from its incoming
     // location (a local variable register, or the argsArray) into
     // its home location, if they are different.
-    int i = 0;
-    int opt_i = 0;
+    Declaration param = firstDecl();
+    // i is index of current parameter, not counting this.
+    int i = param != null && param.isThisParameter() ? -1 : 0;
     int key_i = 0;
     int key_args = keywords == null ? 0 : keywords.length;
-    int opt_args = defaultArgs == null ? 0
-      : defaultArgs.length - key_args;
     if (this instanceof ModuleExp)
       return;
     // If plainArgs>=0, it is the number of arguments *not* in argsArray.
@@ -1357,7 +1397,7 @@ public class LambdaExp extends ScopeExp
     Method mainMethod = getMainMethod();
     Variable callContextSave = comp.callContextVar;
 
-    for (Declaration param = firstDecl();  param != null; param = param.nextDecl())
+    for (;  param != null; param = param.nextDecl(), i++)
       {
         comp.callContextVar
           = (getCallConvention() < Compilation.CALL_WITH_CONSUMER ? null
@@ -1375,6 +1415,8 @@ public class LambdaExp extends ScopeExp
 		defaultStart = 0;
 	      }
 	  }
+        if (param.ignorable())
+          continue;
 	if (plainArgs >= 0 || ! param.isSimple()
 	    || param.isIndirectBinding())
 	  {
@@ -1409,7 +1451,7 @@ public class LambdaExp extends ScopeExp
 		code.emitPushInt(i - plainArgs);
 		code.emitArrayLoad();
 		code.emitElse();
-		defaultArgs[defaultStart + opt_i++].compile(comp, paramType);
+		param.getInitValue().compile(comp, paramType);
 		code.emitFi();
 	      }
 	    else if (max_args < 0 && i == min_args + opt_args)
@@ -1426,7 +1468,11 @@ public class LambdaExp extends ScopeExp
 		code.emitLoad(argsArray);
 		code.emitPushInt(min_args + opt_args - plainArgs);
 		comp.compileConstant(keywords[key_i++]);
-		Expression defaultArg = defaultArgs[defaultStart + opt_i++];
+		Expression defaultArg = param.getInitValue();
+                Type boxedParamType = paramType instanceof PrimType
+                    ? ((PrimType) paramType).boxedType()
+                    : paramType;
+
 		// We can generate better code if the defaultArg expression
 		// has no side effects.  For simplicity and safety, we just
 		// special case literals, which handles most cases.
@@ -1444,7 +1490,7 @@ public class LambdaExp extends ScopeExp
 			  ("searchForKeyword",  argts,
 			   Type.objectType, Access.PUBLIC|Access.STATIC);
 		      }
-		    defaultArg.compile(comp, paramType);
+		    defaultArg.compile(comp, boxedParamType);
 		    code.emitInvokeStatic(searchForKeywordMethod4);
 		  }
 		else
@@ -1465,7 +1511,8 @@ public class LambdaExp extends ScopeExp
 		    comp.compileConstant(Special.dfault);
 		    code.emitIfEq();
 		    code.emitPop(1);
-		    defaultArg.compile(comp, paramType);
+		    defaultArg.compile(comp, boxedParamType);
+                    paramType.emitCoerceToObject(code);
 		    code.emitFi();
 		  }
 	      }
@@ -1485,14 +1532,13 @@ public class LambdaExp extends ScopeExp
             else
 	      code.emitPutField(param.field);
 	  }
-	i++;
       }
     comp.callContextVar = callContextSave;
   }
 
   void compileAsMethod (Compilation comp)
   {
-    if ((flags & METHODS_COMPILED) != 0 || isAbstract())
+    if ((flags & METHODS_COMPILED) != 0 || isAbstract() || isNative())
       return;
     flags |= METHODS_COMPILED;
     if (primMethods == null)
@@ -1524,11 +1570,6 @@ public class LambdaExp extends ScopeExp
 	if (i < numStubs)
 	  {
 	    CodeAttr code = comp.method.startCode();
-	    int toCall = i + 1;
-	    while (toCall < numStubs
-		   && defaultArgs[toCall] instanceof QuoteExp)
-	      toCall++;
-	    boolean varArgs = toCall == numStubs && restArgType != null;
 	    Declaration decl;
             Variable callContextSave = comp.callContextVar;
 	    Variable var = code.getArg(0);
@@ -1548,11 +1589,21 @@ public class LambdaExp extends ScopeExp
 		var = var.nextVar();
 	      }
 	    comp.callContextVar = ctxArg ? var : null;
+	    int toCall = i + 1;
 	    for (int j = i; j < toCall;  j++, decl = decl.nextDecl())
 	      {
 		Target paramTarget = StackTarget.getInstance(decl.getType());
-		defaultArgs[j].compile(comp, paramTarget);
+                Expression defaultArg = decl.getInitValue();
+		defaultArg.compile(comp, paramTarget);
+                // Minor optimization: Normally stub[i] calls stub[i+1],
+                // which calls stub[i+2] etc until we get to stub[numStubs].
+                // That way any given default argument expression is only
+                // compiled into a single stub.  However, if the default is a
+                // constant it makes sense to call stub[j] (where j>i+1) directly.
+                if (toCall < numStubs && defaultArg instanceof QuoteExp)
+                  toCall++;
 	      }
+	    boolean varArgs = toCall == numStubs && restArgType != null;
 	    if (varArgs)
 	      {
 		Expression arg;
@@ -1689,13 +1740,13 @@ public class LambdaExp extends ScopeExp
   {
     if (keywords != null && keywords.length > 0)
       return true;
-    if (defaultArgs != null)
+    if (opt_args != 0)
       {
-        for (int i = defaultArgs.length;  --i >= 0; )
+        for (Declaration p = firstDecl(); p != null; p = p.nextDecl())
           {
-            Expression def = defaultArgs[i];
+            Expression defaultArg = p.getInitValue();
             // Non-constant default arguments require care with scoping.
-            if (def != null && ! (def instanceof QuoteExp))
+            if (defaultArg != null && ! (defaultArg instanceof QuoteExp))
               return true;
           }
       }
@@ -1712,11 +1763,11 @@ public class LambdaExp extends ScopeExp
     ctx.writeValue(new Closure(this, ctx));
   }
 
-  Object evalDefaultArg(int index, CallContext ctx)
+  Object evalDefaultArg(Declaration param, CallContext ctx)
   {
     try
       {
-        return defaultArgs[index].eval(ctx);
+        return param.getInitValue().eval(ctx);
       }
     catch (Throwable ex)
       {
@@ -1815,9 +1866,7 @@ public class LambdaExp extends ScopeExp
     out.startLogicalBlock("(", false, ")");
     Special prevMode = null;
     int i = 0;
-    int opt_i = 0;
     int key_args = keywords == null ? 0 : keywords.length;
-    int opt_args = defaultArgs == null ? 0 : defaultArgs.length - key_args;
     Declaration decl = firstDecl();
     if (decl != null && decl.isThisParameter())
       i = -1;
@@ -1841,7 +1890,7 @@ public class LambdaExp extends ScopeExp
 	  }
 	Expression defaultArg = null;
 	if (mode == Special.optional || mode == Special.key)
-	  defaultArg = defaultArgs[opt_i++];
+	  defaultArg = decl.getInitValue();
 	if (defaultArg != null)
 	  out.print('(');
 	decl.printInfo(out);
@@ -1918,9 +1967,10 @@ public class LambdaExp extends ScopeExp
   {
     if (returnType == null)
       {
-	returnType = Type.objectType;  // To guards against cycles.
-	// body may not be set if define scan'd but not yet rewrit'ten.
-	if (body != null && ! isAbstract())
+	returnType = Type.objectType;  // To guard against cycles.
+	// body may not be set if define scan'd but not yet rewritten.
+	if (body != null && ! isAbstract() && ! isNative()
+            && body.getFlag(Expression.VALIDATED))
 	  returnType = body.getType();
       }
     return returnType;
@@ -1938,7 +1988,8 @@ public class LambdaExp extends ScopeExp
     if (returnType != null
         && returnType != Type.objectType
         && returnType != Type.voidType
-        && body != QuoteExp.abstractExp)
+        && body != QuoteExp.abstractExp
+        && body != QuoteExp.nativeExp)
       {
         Expression value = body;
 	body = Compilation.makeCoercion(value, returnType);
@@ -1950,7 +2001,7 @@ public class LambdaExp extends ScopeExp
   public final void setCoercedReturnValue (Expression type,
 					    Language language)
   {
-    if (! isAbstract())
+    if (! isAbstract() && ! isNative())
       {
         Expression value = body;
         body = Compilation.makeCoercion(value, type);
@@ -1960,6 +2011,43 @@ public class LambdaExp extends ScopeExp
     if (rtype != null)
       setReturnType(rtype);
   }
+
+  /** Get the first expression/statement in the body.
+   * It dives down into {@code BeginExp}s.
+   * Used to check for {@code invoke-special} calls in {@code @init} methods.
+   */
+  public Expression getBodyFirstExpression ()
+  {
+    Expression bodyFirst = body;
+    while (bodyFirst instanceof BeginExp)
+      {
+        BeginExp bbody = (BeginExp) bodyFirst;
+        if (bbody.length == 0)
+          bodyFirst = null;
+        else
+          bodyFirst = bbody.exps[0];
+      }
+    return bodyFirst;
+  }
+
+  /** Check if argument is a this(...) or super(...) initializtion.
+   * If so, return return the corresponding this or super class.
+   */
+  public ClassType checkForInitCall (Expression bodyFirst)
+  {
+    ClassType calledInit = null;
+    Object value;  Expression exp;
+    if (bodyFirst instanceof ApplyExp
+        && (exp = ((ApplyExp) bodyFirst).func) instanceof QuoteExp
+        && (value = ((QuoteExp) exp).getValue()) instanceof PrimProcedure)
+      {
+        PrimProcedure pproc = (PrimProcedure) value;
+        if (pproc.isSpecial()
+            && ("<init>".equals(pproc.method.getName())))
+          calledInit = pproc.method.getDeclaringClass();
+      }
+    return calledInit;
+  }
 }
 
 class Closure extends MethodProc
@@ -2025,10 +2113,8 @@ class Closure extends MethodProc
 
     Object[] evalFrame = new Object[lambda.frameSize];
     int key_args = lambda.keywords == null ? 0 : lambda.keywords.length;
-    int opt_args = lambda.defaultArgs == null ? 0
-      : lambda.defaultArgs.length - key_args;
+    int opt_args = lambda.opt_args;
     int i = 0;
-    int opt_i = 0;
     int key_i = 0;
     int min_args = lambda.min_args;
     for (Declaration decl = lambda.firstDecl(); decl != null;
@@ -2042,8 +2128,7 @@ class Closure extends MethodProc
             if (i < nargs)
               value = args[i++];
             else
-              value = lambda.evalDefaultArg(opt_i, ctx);
-            opt_i++;
+              value = lambda.evalDefaultArg(decl, ctx);
           }
 	else if (lambda.max_args < 0 && i == min_args + opt_args)
           {
@@ -2086,8 +2171,7 @@ class Closure extends MethodProc
             int key_offset = min_args + opt_args;
             value = Keyword.searchForKeyword(args, key_offset, keyword);
             if (value == Special.dfault)
-              value = lambda.evalDefaultArg(opt_i, ctx);
-            opt_i++;
+              value = lambda.evalDefaultArg(decl, ctx);
           }
         if (decl.type != null)
           {
