@@ -10,6 +10,7 @@ import gnu.text.Lexer;
 import gnu.text.SourceMessages;
 import gnu.kawa.reflect.*;
 import java.io.*;
+import java.lang.reflect.Constructor;
 import java.lang.reflect.InvocationTargetException;
 import gnu.kawa.lispexpr.ClassNamespace; // FIXME
 
@@ -62,6 +63,9 @@ public abstract class Language
   static String[][] languages =
   {
     { "scheme", ".scm", ".sc", "kawa.standard.Scheme" },
+    { "r5rs", ".scm", "kawa.standard.Scheme" },
+    { "r6rs", ".scm", "kawa.standard.Scheme" },
+    { "r7rs", ".scm", "kawa.standard.Scheme" },
     { "krl", ".krl", "gnu.kawa.brl.BRL" },
     { "brl", ".brl", "gnu.kawa.brl.BRL" },
     { "emacs", "elisp", "emacs-lisp", ".el", "gnu.jemacs.lang.ELisp" },
@@ -349,7 +353,7 @@ public abstract class Language
    */
   protected void defProcStFld(String name, String cname)
   {
-    defProcStFld(name, cname, Compilation.mangleNameIfNeeded(name));
+    defProcStFld(name, cname, mangleNameIfNeeded(name));
   }
 
   /** Enter a named function into the current environment. */
@@ -447,11 +451,154 @@ public abstract class Language
     return name;
   }
 
+  public static String mangleNameIfNeeded (String name)
+  {
+    if (name == null || isValidJavaName(name))
+      return name;
+    else
+      return mangleName(name, 0);
+  }
+
+  public static boolean isValidJavaName(String name)
+  {
+    int len = name.length();
+    if (len == 0 || ! Character.isJavaIdentifierStart(name.charAt(0)))
+      return false;
+    for (int i = len;  --i > 0; )
+      if (! Character.isJavaIdentifierPart(name.charAt(i)))
+	return false;
+    return true;
+  }
+
+  /** Convert a string to a safe Java identifier.
+   * @param kind -1 - non-reversible;
+   *  0: reversible, except that '$' is not mapped;
+   *  1: reversible
+   */
+  public static String mangleName (String name, int kind)
+  {
+    boolean reversible = kind >= 0;
+    int len = name.length ();
+    if (len == 6 && name.equals("*init*")) // Constructor methods.
+      return "<init>";
+    StringBuffer mangled = new StringBuffer (len);
+    boolean upcaseNext = false;
+    for (int i = 0;  i < len;  i++)
+      {
+	char ch = name.charAt(i);
+	if (upcaseNext)
+	  {
+	    ch = Character.toTitleCase(ch);
+	    upcaseNext = false;
+	  }
+	if (Character.isDigit(ch))
+	  {
+	    if (i == 0)
+	      mangled.append("$N");
+	    mangled.append(ch);
+	  }
+	else if (Character.isLetter(ch) || ch == '_')
+	  mangled.append(ch);
+	else if (ch == '$')
+	  mangled.append(kind > 1 ? "$$" : "$");
+	else
+	  {
+	    switch (ch)
+	      {
+	      case '+':  mangled.append("$Pl");  break;
+	      case '-':
+		if (reversible)
+		  mangled.append("$Mn");
+		else
+		  {
+		    char next = i + 1 < len ? name.charAt(i+1) : '\0';
+		    if (next == '>')
+		      {
+			mangled.append("$To$");
+			i++;
+		      }
+		    else if (! Character.isLowerCase(next))
+		      mangled.append("$Mn");
+		  }
+		break;
+	      case '*':  mangled.append("$St");  break;
+	      case '/':  mangled.append("$Sl");  break;
+	      case '=':  mangled.append("$Eq");  break;
+	      case '<':  mangled.append("$Ls");  break;
+	      case '>':  mangled.append("$Gr");  break;
+	      case '@':  mangled.append("$At");  break;
+	      case '~':  mangled.append("$Tl");  break;
+	      case '%':  mangled.append("$Pc");  break;
+	      case '.':  mangled.append("$Dt");  break;
+	      case ',':  mangled.append("$Cm");  break;
+	      case '(':  mangled.append("$LP");  break;
+	      case ')':  mangled.append("$RP");  break;
+	      case '[':  mangled.append("$LB");  break;
+	      case ']':  mangled.append("$RB");  break;
+	      case '{':  mangled.append("$LC");  break;
+	      case '}':  mangled.append("$RC");  break;
+	      case '\'': mangled.append("$Sq");  break;
+	      case '"':  mangled.append("$Dq");  break;
+	      case '&':  mangled.append("$Am");  break;
+	      case '#':  mangled.append("$Nm");  break;
+	      case '?':
+		char first = mangled.length() > 0 ? mangled.charAt(0) : '\0';
+		if (! reversible
+		    && i + 1 == len && Character.isLowerCase(first))
+		  {
+		    mangled.setCharAt(0, Character.toTitleCase(first));
+		    mangled.insert(0, "is");
+		  }
+		else
+		  mangled.append("$Qu");
+		break;
+	      case '!':  mangled.append("$Ex");  break;
+	      case ':':  mangled.append("$Cl");  break;
+	      case ';':  mangled.append("$SC");  break;
+	      case '^':  mangled.append("$Up");  break;
+	      case '|':  mangled.append("$VB");  break;
+	      default:
+		mangled.append('$');
+		mangled.append(Character.forDigit ((ch >> 12) & 15, 16));
+		mangled.append(Character.forDigit ((ch >>  8) & 15, 16));
+		mangled.append(Character.forDigit ((ch >>  4) & 15, 16));
+		mangled.append(Character.forDigit ((ch      ) & 15, 16));
+	      }
+	    if (! reversible)
+	      upcaseNext = true;
+	  }
+      }
+    String mname = mangled.toString ();
+    return mname.equals(name) ? name : mname;
+  }
+
   public abstract Lexer getLexer(InPort inp, SourceMessages messages);
 
-  public Compilation getCompilation (Lexer lexer, SourceMessages messages, NameLookup lexical)
+  public String getCompilationClass () { return "gnu.expr.Compilation"; }
+
+  private Constructor<Compilation> compilationClassConstructor;
+
+  public final Compilation getCompilation (SourceMessages messages, NameLookup lexical)
   {
-    return new Compilation(this, messages, lexical);
+    // Use reflection on the class named by getCompilationClass, rather than
+    // just allocation the instance directly in our subclasses. The reason is
+    // improved separability between compile-time code and run-time code.
+    try
+      {
+        if (compilationClassConstructor == null)
+          {
+            Class<Compilation> compilationClass = (Class<Compilation>)
+              Class.forName(getCompilationClass(), true,
+                            getClass().getClassLoader());
+            compilationClassConstructor = compilationClass.getConstructor
+              (Language.class, SourceMessages.class, NameLookup.class);
+          }
+        return compilationClassConstructor.newInstance(this, messages, lexical);
+      }
+    catch (Exception ex)
+      {
+        throw WrappedException.wrapIfNeeded(ex);
+      }
   }
 
   /** Flag to tell parse that expression will be evaluated immediately.
@@ -472,6 +619,7 @@ public abstract class Language
   public static final int PARSE_FOR_SERVLET = 32;
   /** Compilation explicitly requested, not just because of an import. */
   public static final int PARSE_EXPLICIT = 64;
+  public static final int PARSE_INTERACTIVE_MODULE = 128;
 
   public static boolean requirePedantic;
 
@@ -515,7 +663,7 @@ public abstract class Language
                           ? NameLookup.getInstance(getEnvironment(), this)
                           : new NameLookup(this));
     boolean immediate = (options & PARSE_IMMEDIATE) != 0;
-    Compilation tr = getCompilation(lexer, messages, lexical);
+    Compilation tr = getCompilation(messages, lexical);
     if (requirePedantic)
       tr.pedantic = true;
     if (! immediate)
@@ -527,6 +675,8 @@ public abstract class Language
     if ((options & PARSE_PROLOG) != 0)
       tr.setState(Compilation.PROLOG_PARSING);
     tr.pushNewModule(lexer);
+    if ((options & PARSE_INTERACTIVE_MODULE) != 0)
+      tr.setInteractiveName();
     if (info != null)
       info.setCompilation(tr);
     if (! parse(tr, options))
@@ -551,20 +701,31 @@ public abstract class Language
     return Type.make(clas);
   }
 
-  public final Type getLangTypeFor (Type type)
-  {
-    if (type.isExisting())
-      {
-        Class clas = type.getReflectClass();
-        if (clas != null)
-          return getTypeFor(clas);
-      }
-    return type;
-  }
+    public final Type getLangTypeFor (Type type) {
+	if (type instanceof ParameterizedType) {
+	    ParameterizedType ptype = (ParameterizedType) type;
+	    Type[] pargs = ptype.getTypeArgumentTypes();
+	    if (ptype.getRawType() == LazyType.lazyType && pargs.length == 1) {
+		return LazyType.getInstance(LazyType.lazyType, getLangTypeFor(pargs[0]));
+	    }
+	}
+	if (type instanceof TypeVariable) {
+	    return getLangTypeFor(((TypeVariable) type).getRawType());
+	}
+	if (type.isExisting()) {
+	    Class clas = type.getReflectClass();
+	    if (clas != null)
+		return getTypeFor(clas);
+	}
+	return type;
+    }
 
   public String formatType (Type type)
   {
-    return type.getName();
+    String s = type.getName();
+    if (s == null)
+      s = type.toString();
+    return s;
   }
 
   public static Type string2Type (String name)
@@ -622,6 +783,9 @@ public abstract class Language
     return getTypeFor(exp, true);
   }
 
+  /** Interpreting exp as a type specifier, get the actual type.
+   * (Does not get the type "of" but exp treats/casts exp "to" ("as") a type).
+   */
   public Type getTypeFor (Expression exp, boolean lenient)
   {
     if (exp instanceof QuoteExp)
@@ -689,6 +853,10 @@ public abstract class Language
 
   public static/* for now */ Type unionType (Type t1, Type t2)
   {
+    if (t1 == Type.neverReturnsType)
+      return t2;
+    if (t2 == Type.neverReturnsType)
+      return t1;
     if (t1 == Type.toStringType)
       t1 = Type.javalangStringType;
     if (t2 == Type.toStringType)
@@ -741,6 +909,17 @@ public abstract class Language
           }
 	fdname = Compilation.demangleName(fname, true).intern();
       }
+    try
+      {
+        SourceName sourceName = fld.getReflectField().getAnnotation(SourceName.class);
+        if (sourceName != null)
+          {
+            fdname = Symbol.valueOf(sourceName.name(), sourceName.uri(), sourceName.prefix());
+          }
+      }
+    catch (Throwable ex)
+      {
+      }
     if (fdname instanceof String)
       {
         String uri = mod.getNamespaceUri();
@@ -752,7 +931,7 @@ public abstract class Language
           fdname = Symbol.make(uri, sname);
       }
     Type dtype = isAlias ? Type.objectType
-      : getTypeFor(ftype.getReflectClass());
+	: getLangTypeFor(ftype);
     Declaration fdecl = mod.addDeclaration(fdname, dtype);
     boolean isStatic = (fld.getModifiers() & Access.STATIC) != 0;
     if (isAlias)
@@ -806,25 +985,6 @@ public abstract class Language
     return (getNamespaceOf(decl) & namespace) != 0;
   }
 
-  public void emitPushBoolean(boolean value, CodeAttr code)
-  {
-    code.emitGetStatic(value ? Compilation.trueConstant
-		       : Compilation.falseConstant);
-  }
-
-  /** Generate code to test if an object is considered true.
-   * Assume the object has been pushed on the JVM stack.
-   * Generate code to push true or false as appropriate. */
-  public void emitCoerceToBoolean(CodeAttr code)
-  {
-    emitPushBoolean(false, code);
-    code.emitIfNEq();
-    code.emitPushInt(1);
-    code.emitElse();
-    code.emitPushInt(0);
-    code.emitFi();
-  }
-
   public Object coerceFromObject(Class clas, Object obj)
   {
     return getTypeFor(clas).coerceFromObject(obj);
@@ -940,6 +1100,7 @@ public abstract class Language
     try
       {
 	Compilation comp = parse(port, messages, PARSE_FOR_EVAL);
+        comp.setEvalName();
 	ModuleExp.evalModule(getEnvironment(), ctx, comp, null, null);
       }
     finally
@@ -951,8 +1112,6 @@ public abstract class Language
 				 + messages.toString(20));
   }
 
-  static protected int env_counter = 0;
-
   public void runAsApplication (String[] args)
   {
     setDefaults(this);
