@@ -4,7 +4,6 @@
 package gnu.expr;
 import gnu.mapping.*;
 import gnu.bytecode.*;
-import gnu.kawa.reflect.FieldLocation;
 import gnu.text.*;
 import java.util.*;
 import gnu.kawa.util.AbstractWeakHashTable;
@@ -15,7 +14,7 @@ public class ModuleInfo
    * Must be non-null unless we're currently compiling the module,
    * in which case sourcePath and comp must both be non-null.
    */
-  private String className;
+  protected String className;
 
   Class moduleClass;
 
@@ -147,7 +146,7 @@ public class ModuleInfo
           return comp.mainLambda;
         ClassType ctype = ClassType.make(className);
         m = new ModuleExp();
-        m.type = ctype;
+        m.compiledType = ctype;
         m.setName(ctype.getName());
         m.flags |= ModuleExp.LAZY_DECLARATIONS;
         m.info = this;
@@ -185,17 +184,26 @@ public class ModuleInfo
 	  continue;
 	try
 	  {
-            if ((flags & Access.STATIC) == 0 && instance == null)
-              instance = getInstance();
-            Object fvalue = rclass.getField(fld.getName()).get(instance);
-
-            Declaration fdecl = language.declFromField(mod, fvalue, fld);
+            Type ftype = fld.getType();
+            Declaration fdecl;
             if ((flags & Access.FINAL) != 0
-                && (! (fvalue instanceof gnu.mapping.Location)
-                    || fvalue instanceof FieldLocation))
-              fdecl.noteValue(new QuoteExp(fvalue));
+                && (! ftype.isSubtype(Compilation.typeLocation)
+                    || ftype.isSubtype(Compilation.typeFieldLocation)))
+              {
+                if ((flags & Access.STATIC) == 0 && instance == null)
+                  instance = getInstance();
+                Object fvalue = rclass.getField(fld.getName()).get(instance);
+                fdecl = language.declFromField(mod, fvalue, fld);
+                fdecl.noteValue(new QuoteExp(fvalue));
+              }
             else
-              fdecl.noteValue(null);
+              {
+                // FIXME - Only used to get name - better to use an annotation.
+                Object fvalue = (flags & Access.STATIC) == 0 ? null
+                  : rclass.getField(fld.getName()).get(null);
+                fdecl = language.declFromField(mod, fvalue, fld);
+                fdecl.noteValueUnknown();
+              }
 	  }
 	catch (Exception ex)
 	  {
@@ -206,7 +214,7 @@ public class ModuleInfo
     for (Declaration fdecl = mod.firstDecl();
          fdecl != null;  fdecl = fdecl.nextDecl())
       {
-        makeDeclInModule2(mod, fdecl);
+        mod.makeDeclInModule2(fdecl);
       }
     return mod;
   }
@@ -272,39 +280,6 @@ public class ModuleInfo
     return inst;
   }
 
-  static void makeDeclInModule2 (ModuleExp mod, Declaration fdecl)
-  {
-    Object fvalue = fdecl.getConstantValue();
-    if (fvalue instanceof FieldLocation)
-      {
-	FieldLocation floc = (FieldLocation) fvalue;
-        Declaration vdecl = floc.getDeclaration();
-        ReferenceExp fref = new ReferenceExp(vdecl);
-        fdecl.setAlias(true);
-        fref.setDontDereference(true);
-        fdecl.setValue(fref);
-        if (vdecl.isProcedureDecl())
-          fdecl.setProcedureDecl(true);
-        if (vdecl.getFlag(Declaration.IS_SYNTAX))
-          fdecl.setSyntax();
-        if (! fdecl.getFlag(Declaration.STATIC_SPECIFIED))
-          {
-            ClassType vtype = floc.getDeclaringClass();
-            String vname = vtype.getName();
-            for (Declaration xdecl = mod.firstDecl();
-                 xdecl != null;  xdecl = xdecl.nextDecl())
-              {
-                if (vname.equals(xdecl.getType().getName())
-                    && xdecl.getFlag(Declaration.MODULE_REFERENCE))
-                  {
-                    fref.setContextDecl(xdecl);
-                    break;
-                  }
-              }
-          }
-      }
-  }
-
   public int getState () { return comp == null ? Compilation.CLASS_WRITTEN : comp.getState(); }
 
   public void loadByStages (int wantedState)
