@@ -5,10 +5,10 @@ public class ObjectExp extends ClassExp
 {
   public ObjectExp ()
   {
-    super(true);
+    super(true, new ClassType());
   }
 
-  public Type getType() { return type; }
+  protected Type calculateType() { return compiledType; }
 
   protected <R,D> R visit (ExpVisitor<R,D> visitor, D d)
   {
@@ -19,14 +19,14 @@ public class ObjectExp extends ClassExp
   {
     compileMembers(comp);
     CodeAttr code = comp.getCode();
-    code.emitNew(type);
+    code.emitNew(compiledType);
     code.emitDup(1);
-    Method init = Compilation.getConstructor(type, this);
+    Method init = Compilation.getConstructor(compiledType, this);
     if (closureEnvField != null)
       {
 	LambdaExp caller = outerLambda();
 	Variable closureEnv =
-	  Compilation.defaultCallConvention < Compilation.CALL_WITH_CONSUMER ? getOwningLambda().heapFrame
+          ! comp.usingCallContext() ? getOwningLambda().heapFrame
 	  : caller.heapFrame != null ? caller.heapFrame	: caller.closureEnv;
 	if (closureEnv == null)
 	  code.emitPushThis();
