@@ -313,10 +313,25 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
 
   public PrimProcedure(Method method, Language language)
   {
-    this(method, '\0', language);
+    this(method, '\0', language, null);
   }
 
-  public PrimProcedure(Method method, char mode, Language language)
+
+    static Type resolveTypeVariable(TypeVariable tvar, ParameterizedType parameterizedType ) {
+	if (parameterizedType != null) {
+	    TypeVariable[] tparams = parameterizedType.getRawType().getTypeParameters();
+	    int nparams = tparams.length;
+	    for (int i = 0;  i < nparams;  i++) {
+		if (tvar.getName().equals(tparams[i].getName())) {
+		    return parameterizedType.getTypeArgumentType(i);
+		}
+	    }
+	}
+	return tvar.getRawType();
+    }
+
+    public PrimProcedure(Method method, char mode, Language language,
+			 ParameterizedType parameterizedType)
   {
     this.mode = mode;
 
@@ -330,7 +345,11 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
     for (int i = nTypes;  --i >= 0; )
       {
 	Type javaType = pTypes[i];
-	Type langType = language.getLangTypeFor(javaType);
+	Type langType = javaType;
+	if (langType instanceof TypeVariable) {
+	    langType = resolveTypeVariable((TypeVariable) langType, parameterizedType);
+	}
+	langType = language.getLangTypeFor(langType);
 	if (javaType != langType)
 	  {
 	    if (argTypes == null)
@@ -349,7 +368,11 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
       retType = Type.objectType;
     else
       {
-        retType = language.getLangTypeFor(method.getReturnType());
+	retType = method.getReturnType();
+	if (retType instanceof TypeVariable) {	
+	    retType = resolveTypeVariable((TypeVariable) retType, parameterizedType);
+	}
+        retType = language.getLangTypeFor(retType);
 
         // Kludge - toStringType doesn't have methods.
         // It shouldn't be used as the "type" of anything -
@@ -471,7 +494,9 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
     // or if the last argument already is an array we can use it as is.
     // The tricky part is we sometimes have to distinguish these cases
     // at run-time, in which case we set createVarargsArrayIfNeeded.
-    boolean createVarargsArrayIfNeeded = false;
+    // FIXME This is needless and unreliable complexity.  We should by default
+    // create a varargs array - even if the actual argument is an array.
+   boolean createVarargsArrayIfNeeded = false;
     if (variable && (method.getModifiers() & Access.VARARGS) != 0
         && nargs > 0 && argTypes.length > 0
         && nargs == arg_count + (is_static ? 0 : 1))
@@ -566,10 +591,10 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
             // on the operand stack or in a local variable when
             // any backwards branch is taken."
             // Hence re-write:
-            //   (make Foo a1 backward_dump_containing_expression a3 ...)
+            //   (make Foo a1 backward_jump_containing_expression a3 ...)
             // to:
             //   (let ((t1 a1)
-            //         (t2 backward_dump_containing_expression)
+            //         (t2 backward_jump_containing_expression)
             //         (t3 q2) ...)
             //     (make Foo a1 t1 t2 t3 ...))
             int nargs = args.length;
@@ -579,9 +604,19 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
             for (int i = 1;  i < nargs;  i++)
               {
                 Expression argi = args[i];
-                Declaration d = comp.letVariable(null, argi.getType(), argi);
-                d.setCanRead(true);
-                xargs[i] = new ReferenceExp(d);
+                // A modest optimzation: Don't generate temporary if not needed.
+                // But this also avoids a possible VerifyError in the case
+                // of LambdaExp, since if we set the latter's nameDecl to the
+                // new temporary, loading the temporary can get confused.
+                if (! (argi instanceof QuoteExp
+                       || argi instanceof LambdaExp
+                       || argi instanceof ReferenceExp))
+                  {
+                    Declaration d = comp.letVariable(null, argi.getType(), argi);
+                    d.setCanRead(true);
+                    argi = new ReferenceExp(d);
+                  }
+                xargs[i] = argi;
               }
             comp.letEnter();
             LetExp let = comp.letDone(new ApplyExp(exp.func, xargs));
@@ -603,6 +638,8 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
     Expression[] args = exp.getArgs();
     gnu.bytecode.CodeAttr code = comp.getCode();
     Type stackType = retType;
+    if (method != null && method.getReturnType() instanceof TypeVariable)
+	stackType = method.getReturnType().getRawType();
     int startArg = 0;
     if (isConstructor())
       {
@@ -631,7 +668,6 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
       }
     else if (takesTarget() && method.getStaticFlag())
       startArg = 1;
-
     compileArgs(args, startArg, thisType, comp);
 
     if (method == null)
@@ -744,7 +780,6 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
     boolean varArgs = takesVarArgs();
     if (index < lenTypes && ! varArgs)
       return argTypes[index];
-    // if (! varArgs) ERROR;
     Type restType = argTypes[lenTypes - 1];
     if (restType instanceof ArrayType)
       return ((ArrayType) restType).getComponentType();
@@ -1046,7 +1081,6 @@ public class PrimProcedure extends MethodProc implements gnu.expr.Inlineable
     return buf.toString();
   }
 
-
   public String toString()
   {
     StringBuffer buf = new StringBuffer(100);
