@@ -6,6 +6,7 @@ import gnu.kawa.reflect.CompileReflect;
 import gnu.kawa.reflect.Invoke;
 import gnu.kawa.reflect.ArrayGet;
 import gnu.kawa.reflect.ArraySet;
+import gnu.kawa.reflect.LazyType;
 import gnu.math.*;
 import gnu.text.Char;
 
@@ -16,24 +17,23 @@ import gnu.text.Char;
 
 public class CompilationHelpers
 {
+    public static boolean maybeLazy(Expression exp) {
+        if (exp instanceof QuoteExp)
+            return false;
+        return LazyType.maybeLazy(exp.getType());
+    }
+
   private static boolean nonNumeric(Expression exp)
   {
     if (exp instanceof QuoteExp)
       {
         Object value = ((QuoteExp) exp).getValue();
-        return ! (value instanceof Numeric || value instanceof Char
+        return ! (value instanceof Number || value instanceof Char
                   || value instanceof Symbol);
       }
     return false;
   }
 
-  static final ClassType typeList
-  /* #ifdef JAVA2 */
-  = ClassType.make("java.util.List");
-  /* #else */
-  // = ClassType.make("gnu.lists.Sequence");
-  /* #endif */
-
   public static Expression validateApplyToArgs
   (ApplyExp exp, InlineCalls visitor,
    Type required, Procedure applyToArgs)
@@ -45,13 +45,21 @@ public class CompilationHelpers
         Expression proc = args[0];
         if (! proc.getFlag(Expression.VALIDATED))
           {
-            if (proc instanceof LambdaExp)
+            Expression pval = proc;
+            if (proc instanceof ReferenceExp)
+              {
+                Declaration decl = ((ReferenceExp) proc).getBinding();
+                if (decl != null)
+                  pval = decl.getValue();
+              }
+            if (pval != null && pval.getClass() == LambdaExp.class)
               {
                 Expression[] rargs = new Expression[nargs];
                 System.arraycopy(args, 1, rargs, 0, nargs);
-                return visitor.visit(new ApplyExp(proc, rargs).setLine(exp), required);
+                exp.setFuncArgs(proc, rargs);
+                return visitor.visit(exp, required);
               }
-            proc = visitor.visit(proc, null);
+            proc = visitor.visit(proc, InlineCalls.typeForCalledFunction(proc));
             args[0] = proc;
           }
         Type ptype = proc.getType().getRealType();
@@ -61,34 +69,39 @@ public class CompilationHelpers
           {
             Expression[] rargs = new Expression[nargs];
             System.arraycopy(args, 1, rargs, 0, nargs);
-            ApplyExp nexp = new ApplyExp(proc, rargs);
-            nexp.setLine(exp);
-            return proc.validateApply(nexp, visitor, required, null);
+            exp.setFuncArgs(proc, rargs);
+            return proc.validateApply(exp, visitor, required, null);
           }
 
-        ClassType ctype;
+        ClassType ctype = ptype instanceof ClassType ? (ClassType) ptype
+	    : ptype instanceof ParameterizedType
+	    ? ((ParameterizedType) ptype).getRawType()
+	    : null;
         ApplyExp result = null;
         if (CompileReflect.checkKnownClass(ptype, comp) < 0)
           ; // This might be more cleanly handled at the type specifier. FIXME
         else if (ptype.isSubtype(Compilation.typeType)
                  || language.getTypeFor(proc,false) != null)
           {
-            result = new ApplyExp(Invoke.make, args);
+            result = exp.setFuncArgs(Invoke.make, args);
           }
         else if (ptype instanceof ArrayType)
           {
             Type elementType = ((ArrayType) ptype).getComponentType();
-            result = new ApplyExp(new ArrayGet(elementType), args);
+            exp.setFuncArgs(new ArrayGet(elementType), args);
+            result = exp;
           }
-        else if (ptype instanceof ClassType
-                 && (ctype = (ClassType) ptype).isSubclass(typeList)
+        else if (ctype != null
+		 && ctype.isSubclass(Compilation.typeList)
                  && nargs == 1)
           {
             // We search for a "get(int)" method, rather than just using
             // typeList.getDeclaredMethod("get", 1) to see if we make a
             // a virtual call rather than an interface call.
             Method get = ctype.getMethod("get", new Type[] { Type.intType  });
-            result = new ApplyExp(get, args);
+	    ParameterizedType prtype = ptype instanceof ParameterizedType ?
+		((ParameterizedType) ptype) : null;
+            result = exp.setFuncArgs(new PrimProcedure(get, 'V', language, prtype), args);
           }
         if (result != null)
           {
@@ -120,7 +133,7 @@ public class CompilationHelpers
             return new SetArrayExp(arg, (ArrayType) argType);
           }
         if (argType instanceof ClassType
-            && (ctype = (ClassType) argType).isSubclass(typeList))
+            && (ctype = (ClassType) argType).isSubclass(Compilation.typeList))
           {
             if (exp instanceof SetListExp)
               return exp;
@@ -159,7 +172,8 @@ public class CompilationHelpers
   {
     exp.visitArgs(visitor);
     Expression[] args = exp.getArgs();
-    if (nonNumeric(args[0]) || nonNumeric(args[1]))
+    if ((nonNumeric(args[0]) || nonNumeric(args[1]))
+        && ! maybeLazy(args[0]) && ! maybeLazy(args[1]))
       return new ApplyExp(((IsEqv) proc).isEq, args);
     return exp;
   }
