@@ -1,16 +1,21 @@
 package gnu.kawa.functions;
 import gnu.bytecode.*;
 import gnu.mapping.*;
+import gnu.math.*;
 import gnu.expr.*;
 import gnu.kawa.reflect.*;
 import gnu.kawa.lispexpr.LangObjType;
 import gnu.lists.LList;
 import kawa.standard.Scheme;
+import kawa.standard.SchemeCompilation;
 
 public class CompileMisc implements Inlineable
 {
   static final int CONVERT = 2;
   static final int NOT = 3;
+  static final int EQ = 4;
+  static final int NUMBER_COMPARE = 5;
+  static final int NUMBER_PREDICATE = 6;
   int code;
   Procedure proc;
 
@@ -30,6 +35,21 @@ public class CompileMisc implements Inlineable
     return new CompileMisc((Procedure) proc, NOT);
   }
 
+  public static CompileMisc forEq(Object proc)
+  {
+    return new CompileMisc((Procedure) proc, EQ);
+  }
+
+  public static CompileMisc forNumberCompare(Object proc)
+  {
+    return new CompileMisc((Procedure) proc, NUMBER_COMPARE);
+  }
+
+  public static CompileMisc forNumberPredicate(Object proc)
+  {
+    return new CompileMisc((Procedure) proc, NUMBER_PREDICATE);
+  }
+
   public void compile (ApplyExp exp, Compilation comp, Target target)
   {
     switch (code)
@@ -40,6 +60,15 @@ public class CompileMisc implements Inlineable
       case NOT:
         compileNot((Not) proc, exp, comp, target);
         return;
+      case EQ:
+        compileEq(exp.getArgs(), comp, target, ((IsEq) proc).language);
+        return;
+      case NUMBER_COMPARE:
+        compileNumberCompare((NumberCompare) proc, exp, comp, target);
+        return;
+      case NUMBER_PREDICATE:
+        compileNumberPredicate((NumberPredicate) proc, exp, comp, target);
+        return;
       default: throw new Error();
       }
   }
@@ -80,7 +109,7 @@ public class CompileMisc implements Inlineable
     return exp;
   }
 
-  public static Expression validateApplyNot
+  public static Expression validateApplySimpleBoolean
   (ApplyExp exp, InlineCalls visitor, Type required, Procedure proc)
   {
     exp.visitArgs(visitor);
@@ -288,6 +317,201 @@ public class CompileMisc implements Inlineable
       }
   }
 
+  public static void compileEq (Expression[] args, Compilation comp,
+			      Target target, Language language)
+  {
+    CodeAttr code = comp.getCode();
+    args[0].compile(comp, Target.pushObject);
+    args[1].compile(comp, Target.pushObject);
+    if (target instanceof ConditionalTarget)
+      {
+	ConditionalTarget ctarget = (ConditionalTarget) target;
+	if (ctarget.trueBranchComesFirst)
+	  code.emitGotoIfNE(ctarget.ifFalse);
+	else
+	  code.emitGotoIfEq(ctarget.ifTrue);
+	ctarget.emitGotoFirstBranch(code);
+      }
+    else
+      {
+	Type type;
+	code.emitIfEq();
+	if (target.getType() instanceof ClassType)
+	  {
+	    Object trueValue = language.booleanObject(true);
+	    Object falseValue = language.booleanObject(false);
+	    comp.compileConstant(trueValue, Target.pushObject);
+	    code.emitElse();
+	    comp.compileConstant(falseValue, Target.pushObject);
+	    if (trueValue instanceof Boolean && falseValue instanceof Boolean)
+	      type = Compilation.scmBooleanType;
+	    else
+	      type = Type.pointer_type;
+	  }
+	else
+	  {
+	    code.emitPushInt(1);
+	    code.emitElse();
+	    code.emitPushInt(0);
+	    type = language.getTypeFor(Boolean.TYPE);
+	  }
+	code.emitFi();
+	target.compileFromStack(comp, type);
+      }
+  }
+
+  public static void compileNumberCompare (NumberCompare proc, ApplyExp exp, Compilation comp, Target target)
+  {
+    Expression[] args = exp.getArgs();
+    if (args.length == 2)
+      {
+	Expression arg0 = args[0];
+	Expression arg1 = args[1];
+	int kind0 = classifyForNumCompare(arg0);
+	int kind1 = classifyForNumCompare(arg1);
+	CodeAttr code = comp.getCode();
+	if (kind0 > 0 && kind1 > 0
+            && kind0 <= Arithmetic.REALNUM_CODE && kind1 <= Arithmetic.REALNUM_CODE
+	    // Don't optimize if both operands are fractions. FIXME???
+	    && (kind0 != Arithmetic.RATNUM_CODE || kind1 != Arithmetic.RATNUM_CODE))
+	  {
+	    if (! (target instanceof ConditionalTarget))
+	      {
+		IfExp.compile(exp, QuoteExp.trueExp, QuoteExp.falseExp,
+			      comp, target);
+		return;
+	      }
+	    int mask = proc.flags;
+	    if (mask == NumberCompare.TRUE_IF_NEQ)
+	      mask = NumberCompare.TRUE_IF_GRT|NumberCompare.TRUE_IF_LSS;
+	    if (kind0 <= Arithmetic.INTNUM_CODE && kind1 <= Arithmetic.INTNUM_CODE
+		&& (kind0 > Arithmetic.LONG_CODE || kind1 > Arithmetic.LONG_CODE))
+	      {
+		Type[] ctypes = new Type[2];
+		ctypes[0] = Arithmetic.typeIntNum;
+		if (kind1 <= Arithmetic.LONG_CODE)
+		  {
+		    ctypes[1] = Type.longType;
+		  }
+		else if (kind0 <= Arithmetic.LONG_CODE
+			 // Simple check to avoid re-ordering side-effects.
+			 && (arg0 instanceof QuoteExp
+			     || arg1 instanceof QuoteExp
+			     || arg0 instanceof ReferenceExp
+			     || arg1 instanceof ReferenceExp))
+		  {
+		    ctypes[1] = Type.longType;
+		    args = new Expression[2];
+		    args[0] = arg1;
+		    args[1] = arg0;
+		    if (mask != NumberCompare.TRUE_IF_EQU && mask != NumberCompare.TRUE_IF_GRT+NumberCompare.TRUE_IF_LSS)
+		      mask ^= NumberCompare.TRUE_IF_GRT|NumberCompare.TRUE_IF_LSS;
+		  }
+		else
+		  ctypes[1] = Arithmetic.typeIntNum;
+		Method cmeth
+		  = Arithmetic.typeIntNum.getMethod("compare", ctypes);
+		PrimProcedure compare = new PrimProcedure(cmeth);
+		arg0 = new ApplyExp(compare, args);
+		arg1 = new QuoteExp(IntNum.zero());
+		kind0 = kind1 = Arithmetic.INT_CODE;
+	      }
+	    Type commonType;
+	    if (kind0 <= Arithmetic.INT_CODE && kind1 <= Arithmetic.INT_CODE)
+	      commonType = Type.intType;
+	    else if (kind0 <= Arithmetic.LONG_CODE && kind1 <= Arithmetic.LONG_CODE)
+	      commonType = Type.longType;
+	    else
+	      commonType = Type.doubleType;
+	    StackTarget subTarget = new StackTarget(commonType);
+	    ConditionalTarget ctarget = (ConditionalTarget) target;
+	    
+	    int opcode;
+	    if (arg0 instanceof QuoteExp && ! (arg1 instanceof QuoteExp))
+	      {
+		Expression tmp = arg1; arg1 = arg0; arg0 = tmp;
+		if (mask != NumberCompare.TRUE_IF_EQU && mask !=NumberCompare. TRUE_IF_GRT+NumberCompare.TRUE_IF_LSS)
+		  mask ^= NumberCompare.TRUE_IF_GRT|NumberCompare.TRUE_IF_LSS;
+	      }
+	    Label label1 = ctarget.trueBranchComesFirst ? ctarget.ifFalse : ctarget.ifTrue;
+	    if (ctarget.trueBranchComesFirst)
+	      mask ^= NumberCompare.TRUE_IF_GRT|NumberCompare.TRUE_IF_LSS|NumberCompare.TRUE_IF_EQU;
+	    switch (mask)
+	      {
+	      case NumberCompare.TRUE_IF_GRT:  opcode = 157 /*ifgt*/;  break;
+	      case NumberCompare.TRUE_IF_EQU:  opcode = 153 /*ifeq*/;  break;
+	      case NumberCompare.TRUE_IF_LSS:  opcode = 155 /*iflt*/;  break;
+	      case NumberCompare.TRUE_IF_GRT|NumberCompare.TRUE_IF_LSS:  opcode = 154 /*ifne*/;  break;
+	      case NumberCompare.TRUE_IF_GRT|NumberCompare.TRUE_IF_EQU:  opcode = 156 /*ifge*/;  break;
+	      case NumberCompare.TRUE_IF_LSS|NumberCompare.TRUE_IF_EQU:  opcode = 158 /*ifle*/;  break;
+	      default:
+		opcode = 0;
+	      }
+	    arg0.compile(comp, subTarget);
+	    Object value;
+	    if (kind0 <= Arithmetic.INT_CODE && kind1 <= Arithmetic.INT_CODE
+		&& arg1 instanceof QuoteExp
+		&& (value = ((QuoteExp) arg1).getValue()) instanceof IntNum
+		&& ((IntNum) value).isZero())
+	      {
+		code.emitGotoIfCompare1(label1, opcode);
+	      }
+	    else
+	      {
+		arg1.compile(comp, subTarget);
+		code.emitGotoIfCompare2(label1, opcode);
+	      }
+	    ctarget.emitGotoFirstBranch(code);
+	    return;
+	  }
+      }
+    ApplyExp.compile(exp, comp, target);
+  }
+
+  static int classifyForNumCompare (Expression exp)
+  {
+    Type type = exp.getType();
+    int kind = Arithmetic.classifyType(type);
+    Object value;
+    if (kind == Arithmetic.INTNUM_CODE && exp instanceof QuoteExp
+	&& (value = ((QuoteExp) exp).getValue()) instanceof IntNum)
+      {
+	int ilength = ((IntNum) value).intLength();
+	if (ilength < 32)
+	  return Arithmetic.INT_CODE;
+	if (ilength < 64)
+	  return Arithmetic.LONG_CODE;
+      }
+    return kind;
+  }
+
+  public static void compileNumberPredicate (NumberPredicate proc, ApplyExp exp, Compilation comp, Target target)
+  {
+    Expression[] args = exp.getArgs();
+    int op = proc.op;
+    if (args.length == 1 && (op == NumberPredicate.ODD || op == NumberPredicate.EVEN))
+      {
+	Expression arg0 = args[0];
+        int kind = Arithmetic.classifyType(arg0.getType());
+        if (kind <= Arithmetic.INTNUM_CODE)
+          {
+            PrimType wtype = Type.intType;
+            Target wtarget = StackTarget.getInstance(wtype);
+            CodeAttr code = comp.getCode();
+            if (op == NumberPredicate.EVEN)
+              code.emitPushInt(1);
+            arg0.compile(comp, wtarget);
+            code.emitPushInt(1);
+            code.emitAnd();
+            if (op == NumberPredicate.EVEN)
+              code.emitSub(Type.intType);
+            target.compileFromStack(comp, Type.booleanType);
+            return;
+          }
+      }
+   ApplyExp.compile(exp, comp, target);
+  }
+
   public static Expression validateApplyCallCC
   (ApplyExp exp, InlineCalls visitor, Type required, Procedure proc)
   {
@@ -342,7 +566,7 @@ public class CompileMisc implements Inlineable
     code.emitInvokeSpecial(typeContinuation.getDeclaredMethod("<init>", 1));
     code.emitStore(contVar);
     code.emitTryStart(false, target instanceof IgnoreTarget || target instanceof ConsumerTarget ? null : Type.objectType);
-    ApplyExp app = new ApplyExp(lambda, new Expression[] { new ReferenceExp(contDecl) });
+    ApplyExp app = new ApplyExp(lambda, new ReferenceExp(contDecl) );
     app.compile(comp, target);
     // Emit: cont.invoked = true
     if (code.reachableHere())
@@ -429,13 +653,22 @@ public class CompileMisc implements Inlineable
   {
     Map mproc = (Map) xproc;
     boolean collect = mproc.collect;
-    // FIXME: We should inline the list arguments first before inlining the
-    // procedure argument, for better type inference etc.
-    exp.visitArgs(visitor);
     Expression[] args = exp.getArgs();
     int nargs = args.length;
     if (nargs < 2)
       return exp;  // ERROR
+    
+    Type r = InlineCalls.typeForCalledFunction(args[0]);
+    if (r != null)
+      {
+        for (int i = 1; i < nargs;  i++)
+          visitor.visit(args[i], null);
+        // FIXME: We should inline the list arguments first before inlining the
+        // procedure argument, for better type inference etc.
+        visitor.visit(args[0], r);
+     }
+    else
+      exp.visitArgs(visitor);
 
     nargs--;
 
@@ -444,37 +677,29 @@ public class CompileMisc implements Inlineable
     // so inside loop, since that turns a "read" info a "call", which
     // may allow better inlining.
     boolean procSafeForMultipleEvaluation = ! proc.side_effects();
+    Compilation comp = visitor.getCompilation();
 
     // First an outer (let ((%proc PROC)) L2), where PROC is args[0].
-    Expression[] inits1 = new Expression[1];
-    inits1[0] = proc;
-    LetExp let1 = new LetExp(inits1);
-    Declaration procDecl
-      = let1.addDeclaration("%proc", Compilation.typeProcedure);
-    procDecl.noteValue(args[0]);
+    comp.letStart();
+    Declaration procDecl = comp.letVariable("%proc", Compilation.typeProcedure,
+                                            proc);
 
     // Then an inner L2=(let ((%loop (lambda (argi ...) ...))) (%loop ...))
-    Expression[] inits2 = new Expression[1];
-    LetExp let2 = new LetExp(inits2);
-    let1.setBody(let2);
+    comp.letStart();
     LambdaExp lexp = new LambdaExp(collect ? nargs + 1 : nargs);
-    inits2[0] = lexp;
-    Declaration loopDecl = let2.addDeclaration("%loop");
-    loopDecl.noteValue(lexp);
+    Declaration loopDecl = comp.letVariable("%loop", null, lexp);
+    comp.letEnter();
 
     // Finally an inner L3=(let ((parg1 (as <pair> arg1)) ...) ...)
-    Expression[] inits3 = new Expression[nargs];
-    LetExp let3 = new LetExp(inits3);
-
+    comp.letStart();
     Declaration[] largs = new Declaration[nargs];
     Declaration[] pargs = new Declaration[nargs];
     for (int i = 0;  i < nargs;  i++)
       {
 	String argName = "arg"+i;
 	largs[i] = lexp.addDeclaration(argName);
-	pargs[i] = let3.addDeclaration(argName, Compilation.typePair);
-	inits3[i] = new ReferenceExp(largs[i]);
-	pargs[i].noteValue(inits3[i]);
+	pargs[i] = comp.letVariable(argName, Compilation.typePair,
+                                    new ReferenceExp(largs[i]));
       }
     Declaration resultDecl = collect ? lexp.addDeclaration("result") : null;
     Expression[] doArgs = new Expression[1+nargs];
@@ -487,7 +712,8 @@ public class CompileMisc implements Inlineable
     if (! procSafeForMultipleEvaluation)
       proc = new ReferenceExp(procDecl);
     doArgs[0] = proc;
-    Expression doit = visitor.visitApplyOnly(new ApplyExp(new ReferenceExp(mproc.applyFieldDecl), doArgs), null);
+    Expression applyFunc = new ReferenceExp(SchemeCompilation.applyFieldDecl);
+    Expression doit = visitor.visitApplyOnly(new ApplyExp(applyFunc, doArgs), null);
     if (collect)
       {
 	Expression[] consArgs = new Expression[2];
@@ -497,9 +723,7 @@ public class CompileMisc implements Inlineable
 						 "make", consArgs);
       }
     Expression rec = visitor.visitApplyOnly(new ApplyExp(new ReferenceExp(loopDecl), recArgs), null);
-    lexp.body = collect ? rec : new BeginExp(doit, rec);
-    let3.setBody(lexp.body);
-    lexp.body = let3;
+    lexp.body = comp.letDone(collect ? rec : new BeginExp(doit, rec));
     Expression[] initArgs = new Expression[collect ? nargs + 1 : nargs];
     QuoteExp empty = new QuoteExp(LList.Empty);
     for (int i = nargs;  --i >= 0; )
@@ -525,11 +749,87 @@ public class CompileMisc implements Inlineable
 	body = Invoke.makeInvokeStatic(Compilation.scmListType,
 				       "reverseInPlace", reverseArgs);
       }
-    let2.setBody(body);
+    LetExp let2 = comp.letDone(body);
+    LetExp let1 = comp.letDone(let2);
 
     if (procSafeForMultipleEvaluation)
       return let2;
     else
       return let1;
   }
+
+  public static Expression validateApplyMakePromise
+  (ApplyExp exp, InlineCalls visitor, Type required, Procedure proc)
+  {
+    Expression[] args = exp.getArgs();
+    if (args.length == 1 && args[0] instanceof LambdaExp)
+      {
+        boolean forceValueIfPromise = proc == MakePromise.makeLazy;
+        Type bodyRequired;
+        if (required instanceof LazyType)
+          {
+            Type valueType = ((LazyType) required).getValueType();
+            bodyRequired = forceValueIfPromise ? LazyType.getLazyType(valueType)
+                : valueType;
+          }
+        else
+          bodyRequired = null;
+        LambdaExp lexp = (LambdaExp) args[0];
+        lexp.body = visitor.visit(lexp.body, bodyRequired);
+        args[0] = visitor.visit(lexp, null);
+        Type rtype = lexp.getReturnType();
+        if (forceValueIfPromise)
+          {
+            rtype = rtype instanceof LazyType
+                ? ((LazyType) rtype).getValueType()
+                : Type.objectType;
+          }
+        Type type = LazyType.getLazyType(rtype);
+        String mname = forceValueIfPromise ? "makePromiseLazy" : "makePromise";
+        Method meth = ClassType.make("gnu.kawa.functions.MakePromise")
+            .getDeclaredMethod(mname, 1);
+	PrimProcedure mproc
+	  = new PrimProcedure(meth);
+        mproc.setReturnType(type);
+        exp = new ApplyExp(mproc, args);
+        exp.setType(type);
+      }
+    else
+      {
+        exp.visitArgs(visitor);
+      }
+    return exp;
+  }
+
+    /** A hack to simplify inlining continuation calls.
+     */
+    static class CompileTimeContinuation extends ProcedureN implements Inlineable {
+        Target blockTarget;
+        ExitableBlock exitableBlock;
+
+        public Object applyN (Object[] args) throws Throwable {
+            throw new Error("internal error");
+        }
+
+        public void compile (ApplyExp exp, Compilation comp, Target target) {
+            CodeAttr code = comp.getCode();
+            Expression[] args = exp.getArgs();
+            int nargs = args.length;
+            boolean noStack = (blockTarget instanceof IgnoreTarget
+                               || blockTarget instanceof ConsumerTarget);
+            Type typeNeeded = noStack ? null : target.getType();
+            if (noStack || nargs == 1) {
+                for (int i = 0;  i < nargs;  i++)
+                    args[i].compileWithPosition(comp, blockTarget);
+            } else {
+                AppendValues app = AppendValues.appendValues;
+                app.compile(new ApplyExp(app, args), comp, blockTarget);
+            }
+            exitableBlock.exit();
+        }
+
+        public gnu.bytecode.Type getReturnType (Expression[] args) {
+            return Type.neverReturnsType;
+        }
+    }
 }
