@@ -36,20 +36,17 @@ public abstract class LispLanguage extends Language
   static { getNamedPartLocation.setProcedure(); }
 
   /** The default <code>ReadTable</code> for this language. */
-  public ReadTable defaultReadTable = createReadTable();
+  protected ReadTable defaultReadTable;
 
   /** Create a fresh <code>ReadTable</code> appropriate for this language. */
   public abstract ReadTable createReadTable ();
 
-  public Lexer getLexer(InPort inp, SourceMessages messages)
+  public LispReader getLexer(InPort inp, SourceMessages messages)
   {
     return new LispReader(inp, messages);
   }
 
-  public Compilation getCompilation (Lexer lexer, SourceMessages messages, NameLookup lexical)
-  {
-    return new Translator(this, messages, lexical);
-  }
+  public String getCompilationClass () { return "kawa.lang.Translator"; }
 
   public boolean parse (Compilation comp, int options)
     throws java.io.IOException, gnu.text.SyntaxException
@@ -76,6 +73,8 @@ public abstract class LispLanguage extends Language
                   return false;  // FIXME
                 break;
               }
+            if (lexer.peek() == ')')
+              lexer.fatal("An unexpected close paren was read.");
             tr.scanForm(sexp, mexp);
             if ((options & PARSE_ONE_LINE) != 0)
               {
@@ -98,8 +97,6 @@ public abstract class LispLanguage extends Language
                 return true;
               }
           }
-        if (lexer.peek() == ')')
-          lexer.fatal("An unexpected close paren was read.");
 
         // Must be done before any other module imports this module.
         tr.finishModule(mexp);
@@ -151,18 +148,7 @@ public abstract class LispLanguage extends Language
 
   protected void defSntxStFld(String name, String cname)
   {
-    defSntxStFld(name, cname, Compilation.mangleNameIfNeeded(name));
-  }
-
-  /** Combine a <body> consisting of a list of expression. */
-  public Expression makeBody(Expression[] exps)
-  {
-    return new BeginExp (exps);
-  }
-
-  public Expression makeApply (Expression func, Expression[] args)
-  {
-    return new ApplyExp(func, args);
+    defSntxStFld(name, cname, mangleNameIfNeeded(name));
   }
 
   public boolean selfEvaluatingSymbol (Object obj)
@@ -183,12 +169,4 @@ public abstract class LispLanguage extends Language
     return (Symbol) sym;
   }
 
-  /** If a symbol is lexically unbound, look for a default binding.
-   * The default implementation does nothing.
-   * @return null if no binidng, or an Expression.
-   */
-  public Expression checkDefaultBinding (Symbol name, Translator tr)
-  {
-    return null;
-  }
 }
