@@ -27,6 +27,27 @@ public class LispReader extends Lexer
 
   GeneralHashTable<Integer,Object> sharedStructureTable;
 
+    /** Bind value to index in sharingStructuretable.
+     * @param value The object being defined.
+     * @param sharingIndex Back-reference index.
+     *   I.e. the value N in a @code{#N=} form.  If negative, do nothing.
+     * @return The value unchanged.
+     */
+    public Object bindSharedObject(int sharingIndex, Object value) {
+	if (sharingIndex >= 0) {
+	    GeneralHashTable<Integer,Object> map = sharedStructureTable;
+	    if (map == null) {
+		map = new GeneralHashTable<Integer,Object>();
+		sharedStructureTable = map;
+	    }
+	    Integer key = Integer.valueOf(sharingIndex);
+	    if (map.get(key, this) != this)
+              error('w', "a duplicate #n= definition was read");
+	    map.put(key, value);
+	}
+	return value;
+    }
+
   /** Read a #|...|#-style comment (which may contain other nested comments).
     * Assumes the initial "#|" has already been read.
     */
@@ -61,41 +82,52 @@ public class LispReader extends Lexer
       } while (commentNesting > 0);
   }
 
+    boolean inQuasiSyntax;
+
+  char readCase = lookupReadCase();
+
   /** Get specification of how symbols should be case-folded.
-    * @return Either 'P' (means preserve case), 'U' (upcase),
+    * @return Either '\0' (unspecified - defaults to preserve case),
+    * 'P' (means preserve case), 'U' (upcase),
     * 'D' (downcase), or 'I' (invert case).
     */
-  static char getReadCase()
+  public char getReadCase () { return readCase; }
+  public void setReadCase(char readCase) { this.readCase = readCase; }
+
+  static char lookupReadCase()
   {
-    char read_case;
     try
       {
 	String read_case_string
 	  = Environment.getCurrent().get("symbol-read-case", "P").toString();
-	read_case = read_case_string.charAt(0);
-	if (read_case == 'P') ;
-	else if (read_case == 'u')
-	  read_case = 'U';
-	else if (read_case == 'd' || read_case == 'l' || read_case == 'L')
-	  read_case = 'D';
-	else if (read_case == 'i')
-	  read_case = 'I';
+        if (read_case_string.length() > 0)
+          {
+            char read_case = read_case_string.charAt(0);
+            if (read_case == 'P') ;
+            else if (read_case == 'u')
+              read_case = 'U';
+            else if (read_case == 'd' || read_case == 'l' || read_case == 'L')
+              read_case = 'D';
+            else if (read_case == 'i')
+              read_case = 'I';
+            return read_case;
+          }
       }
     catch (Exception ex)
       {
-	read_case = 'P';
       }
-    return read_case;
+    return '\0';
   }
 
-  public Object readValues (int ch,  ReadTable rtable)
+    public Object readValues (int ch,  ReadTable rtable, int sharingIndex)
       throws java.io.IOException, SyntaxException
   {
-    return readValues(ch, rtable.lookup(ch), rtable);
+    return readValues(ch, rtable.lookup(ch), rtable, sharingIndex);
   }
 
-  /** May return zero or multiple values. */
-  public Object readValues (int ch, ReadTableEntry entry, ReadTable rtable)
+  /** May return zero or multiple values.
+   * Returns no values if looking at whitespace or a comment. */
+  public Object readValues (int ch, ReadTableEntry entry, ReadTable rtable, int sharingIndex)
       throws java.io.IOException, SyntaxException
   {
     // Step numbers refer to steps in section 2.2 of the HyperSpec.
@@ -104,6 +136,7 @@ public class LispReader extends Lexer
 
     seenEscapes = false;
     int kind = entry.getKind();
+    Object result;
     switch (kind)
       {
       case ReadTable.ILLEGAL:
@@ -117,14 +150,14 @@ public class LispReader extends Lexer
 	return Values.empty;
       case ReadTable.TERMINATING_MACRO:
       case ReadTable.NON_TERMINATING_MACRO:
-	return entry.read(this, ch, -1);
+	  return entry.read(this, ch, -1, sharingIndex);
       case ReadTable.CONSTITUENT:
       case ReadTable.SINGLE_ESCAPE: // Step 5:
       case ReadTable.MULTIPLE_ESCAPE: // Step 6:
       default:  // 
-	break;
+	  result = readAndHandleToken(ch, startPos, rtable);
       }
-    return readAndHandleToken(ch, startPos, rtable);
+    return bindSharedObject(sharingIndex, result);
   }
 
   protected Object readAndHandleToken(int ch, int startPos, ReadTable rtable)
@@ -234,7 +267,7 @@ public class LispReader extends Lexer
         String uri = new String(tokenBuffer, lbrace, rbrace-lbrace);
         ch = read(); // skip ':' - previously peeked.
         ch = read();
-        Object rightOperand = readValues(ch, rtable.lookup(ch), rtable);
+        Object rightOperand = readValues(ch, rtable.lookup(ch), rtable, -1);
         if (! (rightOperand instanceof SimpleSymbol))
           error("expected identifier in symbol after '{URI}:'");
         // FIXME should allow "compound keyword" - for attribute names
@@ -360,9 +393,13 @@ public class LispReader extends Lexer
       }
   }
 
-  public Object readObject ()
-      throws java.io.IOException, SyntaxException
-  {
+    public Object readObject () throws java.io.IOException, SyntaxException {
+	return readObjectWithSharing(-1);
+    }
+
+    public Object readObjectWithSharing (int sharingIndex)
+	throws java.io.IOException, SyntaxException
+    {
     char saveReadState = ((InPort) port).readState;
     int startPos = tokenBufferLength;
     ((InPort) port).readState = ' ';
@@ -376,10 +413,11 @@ public class LispReader extends Lexer
 	    int ch = port.read();
 	    if (ch < 0)
 	      return Sequence.eofValue; // FIXME
-            Object value = readValues(ch, rtable);
+            Object value = readValues(ch, rtable, sharingIndex);
 	    if (value == Values.empty)
 	      continue;
-	    return handlePostfix(value, rtable, line, column);
+	    value = handlePostfix(value, rtable, line, column);
+	    return value;
 	  }
       }
     finally
@@ -392,7 +430,7 @@ public class LispReader extends Lexer
   protected boolean validPostfixLookupStart (int ch, ReadTable rtable)
       throws java.io.IOException
   {
-    if (ch < 0 || ch == ':' || ch == rtable.postfixLookupOperator)
+    if (ch < 0 || ch == rtable.postfixLookupOperator)
       return false;
     if (ch == ',')
       return true;
@@ -411,22 +449,34 @@ public class LispReader extends Lexer
     for (;;)
       {
         int ch = port.peek();
-        if (ch < 0 || ch != rtable.postfixLookupOperator)
-          break;
-        // A kludge to map PreOpWord to ($lookup$ Pre 'Word).
-        port.read();
-        int ch2 = port.peek();
-        if (! validPostfixLookupStart(ch2, rtable))
+        if (ch == '[' && rtable.defaultBracketMode == -2)
           {
-            unread();
-            break;
+            port.read();
+            Object lst = ReaderParens.readList(this, null, ch, 1, ']', -1);
+            value = PairWithPosition.make(value, lst,
+                                          port.getName(), line+1, column+1);
+            value = PairWithPosition.make(LispLanguage.bracket_apply_sym, value,
+                                          port.getName(), line+1, column+1);
           }
-        ch = port.read();
-        Object rightOperand = readValues(ch, rtable.lookup(ch), rtable);
-        value = LList.list2(value,
-                            LList.list2(rtable.makeSymbol(LispLanguage.quasiquote_sym), rightOperand));
-        value = PairWithPosition.make(LispLanguage.lookup_sym, value,
-                                      port.getName(), line+1, column+1);
+        else if (ch == rtable.postfixLookupOperator)
+          {
+            // A kludge to map PreOpWord to ($lookup$ Pre 'Word).
+            port.read();
+            int ch2 = port.peek();
+            if (! validPostfixLookupStart(ch2, rtable))
+              {
+                unread();
+                break;
+              }
+            ch = port.read();
+            Object rightOperand = readValues(ch, rtable.lookup(ch), rtable, -1);
+            value = LList.list2(value,
+                                LList.list2(rtable.makeSymbol(LispLanguage.quasiquote_sym), rightOperand));
+            value = PairWithPosition.make(LispLanguage.lookup_sym, value,
+                                          port.getName(), line+1, column+1);
+          }
+        else
+          break;
       }
     return value;
   }
@@ -1107,6 +1157,11 @@ public class LispReader extends Lexer
       return Pair.make(car, cdr);
   }
 
+  protected void setCar (Object pair, Object car)
+  {
+    ((Pair) pair).setCarBackdoor(car);
+  }
+
   protected void setCdr (Object pair, Object cdr)
   {
     ((Pair) pair).setCdrBackdoor(cdr);
@@ -1233,8 +1288,20 @@ public class LispReader extends Lexer
       return Special.undefined;
     if (name.equals("abstract"))
       return Special.abstractSpecial;
+    if (name.equals("native"))
+      return Special.nativeSpecial;
     if (name.equals("null"))
       return null;
+    if (name.equals("fold-case"))
+      {
+        reader.readCase = 'D';
+        return Values.empty;
+      }
+    if (name.equals("no-fold-case"))
+      {
+        reader.readCase = 'P';
+        return Values.empty;
+      }
     reader.error("unknown named constant #!"+name);
     return null;
   }
@@ -1262,7 +1329,7 @@ public class LispReader extends Lexer
         reader.error("invalid uniform vector syntax");
         return null;
       }
-    Object list = ReaderParens.readList(reader, '(', -1, ')');
+    Object list = ReaderParens.readList(reader, null, '(', -1, ')', -1);
     int len = LList.listLength(list, false);
     if (len < 0)
       {
