@@ -20,9 +20,9 @@ public class ReadTable extends RangeTable
   public static final int NON_TERMINATING_MACRO = 6;
 
   /** Default value to pass to setBracketMode() unless overridden. */
-  public static int defaultBracketMode = -1;
+  public static int defaultBracketMode = -2;
 
-  /** A character such that PreOpWord -> ($lookup$ Pre 'Word), if > 0. */
+  /** A character X such that PreXWord -> ($lookup$ Pre 'Word), if > 0. */
   public char postfixLookupOperator = (char) (-1);
 
   /** True if ":IDENTIFIER" should be treated as a keyword. */
@@ -36,6 +36,10 @@ public class ReadTable extends RangeTable
    */
   protected boolean hexEscapeAfterBackslash = true;
 
+    public static ReaderQuote unquoteEntry =
+        new ReaderQuote(makeSymbol(LispLanguage.unquote_sym),
+                        '@', makeSymbol(LispLanguage.unquotesplicing_sym));
+
   /** Set whether ":IDENTIFIER" should be treated as a keyword. */
   public void setInitialColonIsKeyword (boolean whenInitial)
   {
@@ -97,15 +101,9 @@ public class ReadTable extends RangeTable
 
     set('\'', new ReaderQuote(makeSymbol(LispLanguage.quote_sym)));
     set('`',  new ReaderQuote(makeSymbol(LispLanguage.quasiquote_sym)));
-    set(',',  new ReaderQuote(makeSymbol(LispLanguage.unquote_sym),
-                              '@', makeSymbol(LispLanguage.unquotesplicing_sym)));
-
-    if (false) // FUTURE
-      set('[',  ReaderParens.getInstance('[', ']', ReadTable.TERMINATING_MACRO,
-                                         LispLanguage.bracket_list_sym));
-    else
-      setBracketMode();  // Sets the entries for '[', ']', and '<'.
+    set(',',  unquoteEntry);
 
+    setBracketMode();  // Sets the entries for '[', ']', and '<'.
   }
 
   /** Create a new ReadTable and initialize it appropriately for Common Lisp. */
@@ -117,6 +115,8 @@ public class ReadTable extends RangeTable
   }
 
   /** Specify how '[' and ']' (and '<') are handled.
+   * The value -2 means {@code [a b c]} is {@code ($bracket-list$ a b c)}
+   * and {@code f[a b]} is {@code ($bracket-apply$ f a b)}.
    * The value -1 means that '[' and ']' are plain token constituents.
    * The value 0 means that '[' and ']' are equivalent to '(' and ')'.
    * The value 1 means that '[' and ']' are equivalent to '(' and ')', except
@@ -125,7 +125,13 @@ public class ReadTable extends RangeTable
    */
   public void setBracketMode(int mode)
   {
-    if (mode <= 0)
+    if (mode == -2)
+      {
+        set('[',  ReaderParens.getInstance('[', ']', ReadTable.TERMINATING_MACRO,
+                                           LispLanguage.bracket_list_sym));
+        set('<', new ReaderTypespec());
+      }
+    else if (mode <= 0)
       {
 	ReadTableEntry token = ReadTableEntry.getConstituentInstance();
 	set('<', token);
@@ -197,7 +203,18 @@ public class ReadTable extends RangeTable
       {
 	Language language = Language.getDefaultLanguage();
 	if (language instanceof LispLanguage)
-	  table = ((LispLanguage) language).defaultReadTable;
+          {
+            LispLanguage llanguage = (LispLanguage) language;
+            synchronized (llanguage)
+              {
+                table = llanguage.defaultReadTable;
+                if (table == null)
+                  {
+                    table = llanguage.createReadTable();
+                    llanguage.defaultReadTable = table;
+                  }
+              }
+          }
 	else
 	  table = ReadTable.createInitial();
 	current.set(table);
@@ -232,7 +249,7 @@ public class ReadTable extends RangeTable
     return entry;
   }
 
-  protected Object makeSymbol (String name)
+  static Object makeSymbol (String name)
   {
     return Namespace.EmptyNamespace.getSymbol(name.intern());
   }
