@@ -3,10 +3,16 @@ import gnu.mapping.*;
 import gnu.bytecode.*;
 import gnu.expr.*;
 import gnu.kawa.functions.CompileArith;
+import gnu.kawa.functions.CompilationHelpers;
+import gnu.kawa.lispexpr.LangObjType;
 import gnu.math.IntNum;
+import java.lang.reflect.Array;
 
 public class CompileInvoke
 {
+  public static final Method newConstVectorMethod =
+    Compilation.typeConstVector.getDeclaredMethod("<init>", new Type[] { Compilation.objArrayType });
+
   public static Expression validateApplyInvoke
   (ApplyExp exp, InlineCalls visitor, Type required, Procedure proc)
   {
@@ -15,11 +21,10 @@ public class CompileInvoke
     Compilation comp = visitor.getCompilation();
     Expression[] args = exp.getArgs();
     int nargs = args.length;
-    if (! comp.mustCompile
+    if (nargs == 0 || ((kind == 'V' || kind == '*') && nargs == 1))
+      {
         // This should never happen, as InlineCalls.visitApplyExp
         // checks the number of arguments before inline is called.
-        || nargs == 0 || ((kind == 'V' || kind == '*') && nargs == 1))
-      {
         exp.visitArgs(visitor);
         return exp;
       }
@@ -29,6 +34,8 @@ public class CompileInvoke
     Type type0 = (kind == 'V' || kind == '*' ? arg0.getType() : iproc.language.getTypeFor(arg0));
     if (type0 instanceof PairClassType && kind == 'N')
       type = ((PairClassType) type0).instanceType;
+    else if (type0 instanceof LangObjType && kind == 'N')
+      type = ((LangObjType) type0).getConstructorType();
     else if (type0 instanceof ObjectType)
       type = (ObjectType) type0;
     else
@@ -39,19 +46,19 @@ public class CompileInvoke
     if (kind == 'V' || kind == '*')      // Invoke virtual
       {
         margsLength = nargs - 1;
-        argsStartIndex = 2;
+        argsStartIndex = 2; // Skip receiver and method name.
         objIndex = 0;
       }
     else if (kind == 'N')                // make new
       {
         margsLength = nargs;
-        argsStartIndex = 0;
+        argsStartIndex = 0; // Include class specifier
         objIndex = -1;
       }
     else if (kind == 'S' || kind == 's') // Invoke static
       {
         margsLength = nargs - 2;
-        argsStartIndex = 2;
+        argsStartIndex = 2; // Skip class and method name.
         objIndex = -1;
       }
     else if (kind == 'P')                // Invoke special
@@ -59,6 +66,10 @@ public class CompileInvoke
         margsLength = nargs - 2;
         argsStartIndex = 3;
         objIndex = 1;
+        if (type0 == null)
+          comp.error('e', "unknown class for invoke-special", arg0);
+        else if (! (type instanceof ClassType) || type.isInterface())
+          comp.error('e', "invalid class for invoke-special", arg0);
       }
     else
       {
@@ -66,6 +77,9 @@ public class CompileInvoke
         return exp;
       }
 
+    if (kind == 'N' && type == LangObjType.constVectorType
+        && required instanceof ArrayType)
+      type = (ObjectType) required;
     if (kind == 'N' && type instanceof ArrayType)
       {
         ArrayType atype = (ArrayType) type;
@@ -86,14 +100,27 @@ public class CompileInvoke
         if (sizeArg == null)
           sizeArg = QuoteExp.getInstance(new Integer(args.length-1));
         sizeArg = visitor.visit(sizeArg, Type.intType);
+        Object constantValue = null;
+        if (visitor.processingAnnotations() && sizeArg instanceof QuoteExp)
+          {
+            try
+              {
+                int sz = ((Number) sizeArg.valueIfConstant()).intValue();
+                constantValue = Array.newInstance(elementType.getReflectClass(), sz);
+              }
+            catch (Throwable ex)
+              {
+                comp.error('e', "bad array size: "+ex.getMessage());
+              }
+          }
         ApplyExp alloc = new ApplyExp(new ArrayNew(elementType),
                                       new Expression[] { sizeArg } );
-        alloc.setType(atype);
+        alloc.setType(type);
         if (lengthSpecified && args.length == 3)
           return alloc;
-        LetExp let = new LetExp(new Expression[] { alloc });
-        Declaration adecl = let.addDeclaration((String) null, atype);
-        adecl.noteValue(alloc);
+        comp.letStart();
+        Declaration adecl = comp.letVariable((String) null, type, alloc);
+        adecl.setCanRead(true);
         BeginExp begin = new BeginExp();
         int index = 0;
         for (int i = lengthSpecified ? 3 : 1; i < args.length;  i++)
@@ -117,7 +144,21 @@ public class CompileInvoke
                       }
                   }
               }
+            Expression xa = arg;
             arg = visitor.visit(arg, elementType);
+            if (! (arg instanceof QuoteExp))
+              constantValue = null;
+            else if (constantValue != null)
+              {
+                try
+                  {
+                    Array.set(constantValue, index, arg.valueIfConstant());
+                  }
+                catch (Throwable ex)
+                  {
+                    constantValue = null;
+                  }
+              }
             begin.add(new ApplyExp(new ArraySet(elementType),
                                    new Expression[] {
                                      new ReferenceExp(adecl),
@@ -126,7 +167,9 @@ public class CompileInvoke
             index++;
           }
         begin.add(new ReferenceExp(adecl));
-        let.body = begin;
+        LetExp let = comp.letDone(begin);
+        if (constantValue != null)
+          return new QuoteExp(constantValue, type);
         return let;
       }
     else if (type != null && name != null)
@@ -138,7 +181,8 @@ public class CompileInvoke
               {
                 Expression[] xargs = new Expression[nargs-1];
                 System.arraycopy(args, 1, xargs, 0, nargs-1);
-                return visitor.visit(new ApplyExp(constructor, xargs), required);
+                Expression xapp = new ApplyExp(constructor, xargs);
+                return visitor.visit(xapp.setLine(exp), required);
               }
           }
         PrimProcedure[] methods;
@@ -147,117 +191,74 @@ public class CompileInvoke
           : comp.mainClass;
         ObjectType ctype = (ObjectType) type;
         int numCode;
+        int keywordStart = kind == 'N' ? hasKeywordArgument(1, args) : nargs;
+        int tailArgs = nargs - keywordStart;
         try
           {
             methods = getMethods(ctype, name, caller, iproc);
-            numCode = ClassMethods.selectApplicable(methods, margsLength);
+            numCode = ClassMethods.selectApplicable(methods,
+                                                    margsLength - tailArgs);
           }
         catch (Exception ex)
           {
             comp.error('w', "unknown class: " + type.getName());
             return exp;
           }
-        int index = -1;
-        Object[] slots;
-        int keywordStart;
-        if (kind == 'N'
-            && ((keywordStart = hasKeywordArgument(1, args)) < args.length
-                || (numCode <= 0
-                    // There is a default constructor.
-                    && (ClassMethods.selectApplicable(methods,
-                                                      new Type[] { Compilation.typeClassType })
-                        >> 32) == 1))
-            && ((slots = checkKeywords(ctype, args, keywordStart, caller))
-                .length * 2 == (args.length - keywordStart)
-                || ClassMethods.selectApplicable(ClassMethods.getMethods(ctype, "add", 'V', null, iproc.language), 2) > 0))
+
+        if (kind == 'N')
           {
-            StringBuffer errbuf = null;
-            for (int i = 0;  i < slots.length;  i++)
+            boolean usingConstVector = false;
+            if (type == LangObjType.constVectorType)
               {
-                if (slots[i] instanceof String)
+                Method defcons;
+                ClassType creq;
+                if (tailArgs == 0 && required instanceof ClassType
+                    && (creq = (ClassType) required).isSubclass(Compilation.typeList)
+                    && (defcons = creq.getDefaultConstructor()) != null)
                   {
-                    if (errbuf == null)
-                      {
-                        errbuf = new StringBuffer();
-                        errbuf.append("no field or setter ");
-                      }
-                    else
-                      errbuf.append(", ");
-                    errbuf.append('`');
-                    errbuf.append(slots[i]);
-                    errbuf.append('\'');
-                  }
-              }
-            if (errbuf != null)
-              {
-                errbuf.append(" in class ");
-                errbuf.append(type.getName());
-                comp.error('w', errbuf.toString());
-                return exp;
-              }
-            else
-              {
-                ApplyExp ae;
-                if (keywordStart < args.length)
-                  {
-                    Expression[] xargs = new Expression[keywordStart];
-                    System.arraycopy(args, 0, xargs, 0, keywordStart);
-                    ae = (ApplyExp) visitor.visit(new ApplyExp(exp.getFunction(), xargs), ctype);
+                    ctype = creq;
+                    type = ctype;
+                    usingConstVector = true;
+                    keywordStart = 1;
+                    numCode = MethodProc.NO_MATCH_TOO_MANY_ARGS;
+                    tailArgs = nargs-1;
+                    methods[0] = new PrimProcedure(defcons, iproc.language);
+                    arg0 = new QuoteExp(ctype.getReflectClass());
                   }
                 else
-                  ae = new ApplyExp(methods[0], new Expression[] { arg0 });
-                ae.setType(ctype);
-                Expression e = ae;
-                if (args.length > 0)
                   {
-                    for (int i = 0;  i < slots.length;  i++)
-                      {
-                        Object slot = slots[i];
-                        Type stype;
-                        if (slot instanceof Method)
-                          stype = ((Method) slot).getParameterTypes()[0];
-                        else if (slot instanceof Field)
-                          stype = ((Field) slot).getType();
-                        else
-                          stype = null;
-                        if (stype != null)
-                          stype = iproc.language.getLangTypeFor(stype);
-                        Expression arg = visitor.visit(args[keywordStart + 2 * i + 1], stype);
-                        Expression[] sargs
-                          = { ae, new QuoteExp(slot), arg};
-                        ae = new ApplyExp(SlotSet.setFieldReturnObject, sargs);
-                        ae.setType(ctype);
-                      }
-                    int sargs = keywordStart == args.length ? 1
-                      : 2 * slots.length + keywordStart;
-                    e = ae;
-                    if (sargs < args.length)
-                      {
-                        LetExp let = new LetExp(new Expression[] { e });
-                        Declaration adecl = let.addDeclaration((String) null, ctype);
-                        adecl.noteValue(e);
-                        BeginExp begin = new BeginExp();
-                        for (int i = sargs;  i < args.length;  i++)
-                          {
-                            Expression[]  iargs = {
-                              new ReferenceExp(adecl),
-                              QuoteExp.getInstance("add"),
-                              args[i]
-                            };
-                            begin.add(visitor.visit(new ApplyExp(Invoke.invoke,
-                                                                 iargs),
-                                                    null));
-                          }
-                        begin.add(new ReferenceExp(adecl));
-                        let.body = begin;
-                        e = let;
-                      }
+                    Expression[] xargs = new Expression[nargs];
+                    System.arraycopy(args, 1, xargs, 1, nargs-1);
+                    xargs[0] = new QuoteExp(Compilation.objArrayType);
+                    return visitor.visit(new ApplyExp(iproc,
+                                                      new Expression[] {
+                                                          new QuoteExp(Compilation.typeConstVector),
+                                                          new ApplyExp(exp.getFunction(), xargs).setLine(exp)}).setLine(exp), required);
                   }
-                return visitor.checkType(e.setLine(exp), required);
+              }
+
+            CompileBuildObject builder = CompileBuildObject.make(exp, visitor, required, keywordStart, ctype, caller);
+            if (tailArgs > 0 && numCode > 0) // Have keywords
+              return builder.build();
+            if (usingConstVector)
+              {
+                builder.setDefaultConstructor(methods[0]);  
+                return builder.build();
+              }
+            if (numCode == MethodProc.NO_MATCH_TOO_MANY_ARGS
+                && builder.hasDefaultConstructor()
+                && builder.hasAddChildMethod())
+              {
+                builder.keywordStart = 1;
+                return builder.build();
               }
           }
-        int okCount, maybeCount;
-        if (numCode >= 0)
+        int okCount = 0, maybeCount = 0;
+	if (kind == 'N' && tailArgs > 0)
+	  {
+	    comp.error('w', "args following keyword args but no 'add' method");
+	  }
+        else if (numCode >= 0)
           {
             for (int i = 1;  i < nargs; i++)
               {
@@ -302,11 +303,6 @@ public class CompileInvoke
             okCount = (int) (num >> 32);
             maybeCount = (int) num;
           }
-        else
-          {
-            okCount = 0;
-            maybeCount = 0;
-          }
         int nmethods = methods.length;
         if (okCount + maybeCount == 0 && kind == 'N')
           {
@@ -318,6 +314,7 @@ public class CompileInvoke
             okCount = (int) (num >> 32);
             maybeCount = (int) num;
           }
+        int index = -1;
         if (okCount + maybeCount == 0)
           {
             if (kind == 'P' || comp.warnInvokeUnknownMethod())
@@ -411,35 +408,6 @@ public class CompileInvoke
     return exp;
   }
 
-  /** Return an array if args (starting with start) is a set of
-   * (keyword, value)-value pairs. */
-  static Object[] checkKeywords(ObjectType type, Expression[] args,
-                                int start, ClassType caller)
-  {
-    int len = args.length;
-    int npairs = 0;
-    while (start + 2 * npairs + 1 < len
-           && args[start + 2 * npairs].valueIfConstant() instanceof Keyword)
-      npairs++;
-    Object[] fields = new Object[npairs];
-    for (int i = 0;  i < npairs;  i++)
-      {
-        Object value = args[start + 2 * i].valueIfConstant();
-        String name = ((Keyword) value).getName();
-        // Look for field name for a "set" method.
-        Member slot = SlotSet.lookupMember(type, name, caller);
-        if (slot == null)
-          {
-            // Look for for an "add" method.
-            // For example: (define b (JButton action-listener: ...))
-            // maps to: (define b ...) (b:addActionListener ...)
-            slot = type.getMethod(ClassExp.slotToMethodName("add", name), SlotSet.type1Array);
-          }
-        fields[i] = slot != null ? (Object) slot : (Object) name;
-      }
-    return fields;
-  }
-
   private static String getMethodName(Expression[] args, char kind)
   {
     if (kind == 'N')
