@@ -45,6 +45,7 @@ public class XSLT extends XQuery
   {
     Compilation.defaultCallConvention = Compilation.CALL_WITH_CONSUMER;
     ((XslTranslator) comp.lexer).parse(comp);
+    comp.setState(Compilation.BODY_PARSED);
 
     XQParser xqparser = new XQParser(null, comp.getMessages(), this);
     XQResolveNames resolver = new XQResolveNames(comp);
@@ -165,7 +166,8 @@ public class XSLT extends XQuery
     throws Throwable
   {
     CallContext ctx = CallContext.getInstance();
-    String[] args = kawa.repl.commandLineArgArray;
+    ApplicationMainSupport.processSetProperties();
+    String[] args = ApplicationMainSupport.commandLineArgArray;
     for (int i = 0;  i < args.length;  i++)
       {
 	String arg = args[i];
