@@ -4,9 +4,10 @@
 package gnu.lists;
 import java.io.*;
 
-/** Simple adjustable-length vector whose elements are Object references. */
+/** Simple adjustable-length vector whose elements are Object references.
+ */
 
-public class FVector extends SimpleVector
+public class FVector<E> extends SimpleVector<E>
   implements Externalizable, Consumable
   /* #ifdef JAVA2 */
   , Comparable
@@ -74,6 +75,7 @@ public class FVector extends SimpleVector
 
   public void setBufferLength(int length)
   {
+    checkCanWrite();
     int oldLength = data.length;
     if (oldLength != length)
       {
@@ -88,23 +90,25 @@ public class FVector extends SimpleVector
 
   public void shift(int srcStart, int dstStart, int count)
   {
+    checkCanWrite();
     System.arraycopy(data, srcStart, data, dstStart, count);
   }
 
-  public final Object getBuffer(int index)
+  public final E getBuffer(int index)
   {
-    return data[index];
+    return (E) data[index];
   }
 
-  public final Object get (int index)
+  public final E get (int index)
   {
     if (index >= size)
       throw new ArrayIndexOutOfBoundsException();
-    return data[index];
+    return (E) data[index];
   }
 
   public final Object setBuffer(int index, Object value)
   {
+    checkCanWrite();
     Object old = data[index];
     data[index] = value;
     return old;
@@ -112,6 +116,7 @@ public class FVector extends SimpleVector
 
   protected void clearBuffer(int start, int count)
   {
+    checkCanWrite();
     Object[] d = data;
     while (--count >= 0)
       d[start++] = null;
@@ -168,6 +173,7 @@ public class FVector extends SimpleVector
   // FIXME - bad name - setAll should take a Collection
   public final void setAll (Object new_value)
   {
+    checkCanWrite();
     Object[] d = data;
     for (int i = size; --i >= 0; )
       d[i] = new_value;
