@@ -77,6 +77,9 @@ public class Namespace
     return EmptyNamespace;
   }
 
+  /** Return Namespace with the given name (namespace-URI).
+   *  Create it if needed.
+   */
   public static Namespace valueOf (String name)
   {
     if (name == null)
@@ -93,6 +96,15 @@ public class Namespace
       }
   }
 
+    /** Return Namespace with the given name (namespace-URI), if it exists.
+     * Return null if no such namespace exists.
+     */
+    public static Namespace valueOfNoCreate(String name) {
+        if (name == null)
+            name = "";
+        return (Namespace) nsTable.get(name);
+    }
+
   public static Namespace valueOf (String uri, String prefix)
   {
     if (prefix == null || prefix.length() == 0)
@@ -221,7 +233,7 @@ public class Namespace
             if (this == EmptyNamespace)
               sym = new SimpleSymbol(key);
             else
-              sym = new Symbol(this, key);
+              sym = new Symbol(key, this);
             return add(sym, hash);
           }
 	else
