@@ -162,6 +162,11 @@ public class Char
   }
 
   public String toString ()
+  {
+    return toString(value);
+  }
+
+  public static String toString (int value)
   {
     StringBuffer buf = new StringBuffer();
     buf.append('\'');
@@ -280,7 +285,7 @@ class CharMap extends AbstractWeakHashTable<Char,Char>
 	 node != null;  node = node.next)
       {
         Char val = node.getValue();
-        if (val.intValue() == key)
+        if (val != null && val.intValue() == key)
           return val;
       }
     Char val = new Char(key);
