@@ -988,8 +988,12 @@ public class RunXQTS extends FilterConsumer
 
   boolean isSelected (String testName)
   {
+    // it's too slow - slow implementation of indexing of large sequence.
     if ("Constr-cont-document-3".equals(testName))
       return false;
+    // infinite recursion
+    if ("K2-InternalVariablesWithout-9".equals(testName))
+      return false;
     return selectedTest == null || selectedTest.equals(testName);
   }
 
