@@ -246,7 +246,7 @@ public class Shell
       {
 	for (;;)
 	  {
-	    int opts = Language.PARSE_FOR_EVAL|Language.PARSE_ONE_LINE;
+	    int opts = Language.PARSE_FOR_EVAL|Language.PARSE_ONE_LINE|Language.PARSE_INTERACTIVE_MODULE;
 	    try
 	      {
 		Compilation comp = language.parse(lexer, opts, null);
@@ -261,8 +261,6 @@ public class Shell
 		  break;
 		if (sawError)
 		  continue;
-		comp.getModule().setName("atInteractiveLevel$"
-					 + (++ModuleExp.interactiveCounter));
 
 		// Skip whitespace, in case (read-char) or similar is called:
 		int ch;
@@ -447,8 +445,7 @@ public class Shell
           }
         try
           {
-            CompiledModule cmodule = CompiledModule.make(clas, language);
-            cmodule.evalModule(Environment.getCurrent(), OutPort.outDefault());
+            runClass(clas, Environment.getCurrent());
             return true;
           }
         catch (Throwable ex)
@@ -459,6 +456,10 @@ public class Shell
       }
   }
 
+    public static void runClass(Class clas, Environment env) throws Throwable {
+        CompiledModule cmodule = CompiledModule.make(clas, Language.getDefaultLanguage());
+        cmodule.evalModule(env, OutPort.outDefault());
+    }
 
   public static final boolean runFile(InputStream fs, Path path,
                                       Environment env,
@@ -482,20 +483,21 @@ public class Shell
               {
                 SourceMessages messages = new SourceMessages();
                 URL url = path.toURL();
+                OutPort perr = OutPort.errDefault();
                 if (lineByLine)
                   {
                     boolean print = ModuleBody.getMainPrintValues();
                     Consumer out = (print ? getOutputConsumer(OutPort.outDefault())
                                     : new VoidConsumer());
                     Throwable ex
-                      = run(language, env, src, out, null, url, messages);
+                        = run(language, env, src, out, perr, url, messages);
                     if (ex != null)
                       throw ex;
                   }
                 else
                   {
                     cmodule = compileSource (src, env, url, language, messages);
-                    messages.printAll(OutPort.errDefault(), 20);
+                    messages.printAll(perr, 20);
                     if (cmodule == null)
                       return false;
                   }
