@@ -16,16 +16,29 @@ import gnu.kawa.functions.CompileNamedPart;
  * @author	Per Bothner
  */
 
-public class Quote extends Syntax
-{
-  public static final Quote plainQuote = new Quote("quote", false);
-  public static final Quote quasiQuote = new Quote("quasiquote", true);
+public class Quote extends Syntax {
+    public static final Quote plainQuote = new Quote("quote", false);
+    public static final Quote quasiQuote = new Quote("quasiquote", true);
 
-  public Quote (String name, boolean isQuasi)
-  {
-    super(name);
-    this.isQuasi = isQuasi;
-  }
+    public Quote (String name, boolean isQuasi) {
+        super(name);
+        this.isQuasi = isQuasi;
+    }
+
+    protected boolean matchesUnquote(Pair pair, SyntaxForm syntax,
+                                     Translator tr) {
+        return tr.matches(pair.getCar(), syntax, LispLanguage.unquote_sym);
+    }
+
+    protected boolean matchesUnquoteSplicing(Pair pair, SyntaxForm syntax,
+                                     Translator tr) {
+        return tr.matches(pair.getCar(), syntax, LispLanguage.unquotesplicing_sym);
+    }
+
+    protected boolean matchesQuasiQuote(Object form, SyntaxForm syntax,
+                                     Translator tr) {
+        return tr.matches(form, syntax, LispLanguage.quasiquote_sym);
+    }
 
   /** An initial value for 'depth' for plain (non-quasi) quote. */
   protected static final int QUOTE_DEPTH = -1;
@@ -123,26 +136,21 @@ public class Quote extends Syntax
               cdr = new ApplyExp(quoteType.getDeclaredMethod("makeSymbol", 2),
                                  new Expression[] { QuoteExp.getInstance(ns),
                                                     part2 });
-            else if (part1 instanceof ReferenceExp
+            else if (p1.getCar() instanceof SimpleSymbol
                      && part2 instanceof QuoteExp)
-              cdr = tr.getGlobalEnvironment().getSymbol(((ReferenceExp) part1).getName() + ':' + ((QuoteExp) part2).getValue().toString());
+              cdr = tr.getGlobalEnvironment().getSymbol(p1.getCar().toString() + ':' + ((QuoteExp) part2).getValue().toString());
             else if ((combinedName = CompileNamedPart.combineName(part1, part2)) != null)
               cdr = tr.getGlobalEnvironment().getSymbol(combinedName);
             else
-              {
-                Object save = tr.pushPositionOf(pair);
-                tr.error('e', "'"+p1.getCar()+"' is not a valid prefix");
-                tr.popPositionOf(save);
-                cdr = sym;
-              }
+              cdr = pair;
             break;
           }
         else if (depth < 0)
           {
           }
-        else if (tr.matches(pair.getCar(), syntax, LispLanguage.quasiquote_sym))
+        else if (matchesQuasiQuote(pair.getCar(), syntax, tr))
           depth++;
-        else if (tr.matches(pair.getCar(), syntax, LispLanguage.unquote_sym))
+        else if (matchesUnquote(pair, syntax, tr))
           {
             depth--;
             Pair pair_cdr;
@@ -156,7 +164,7 @@ public class Quote extends Syntax
                 break;
               }
           }
-        else if (tr.matches(pair.getCar(), syntax, LispLanguage.unquotesplicing_sym))
+        else if (matchesUnquoteSplicing(pair, syntax, tr))
           return tr.syntaxError ("invalid used of " + pair.getCar() +
 				 " in quasiquote template");
         if (depth == 1 && pair.getCar() instanceof Pair)
@@ -171,10 +179,10 @@ public class Quote extends Syntax
             int splicing = -1;
             if (form instanceof Pair)
               {
-                Object op = ((Pair) form).getCar();
-                if (tr.matches(op, subsyntax, LispLanguage.unquote_sym))
+                Pair pform = (Pair) form;
+                if (matchesUnquote(pform, subsyntax, tr))
                   splicing = 0;
-                else if (tr.matches(op, subsyntax, LispLanguage.unquotesplicing_sym))
+                else if (matchesUnquoteSplicing(pform, subsyntax, tr))
                   splicing = 1;
               }
             if (splicing >= 0)
@@ -221,10 +229,18 @@ public class Quote extends Syntax
             rest = pair.getCdr();
             if (rest instanceof Pair)
               {
-                pair = (Pair) rest;
-                continue;
+                IdentityHashMap map = (IdentityHashMap) seen;
+                Object old = map.get(rest);
+                if (old == null)
+                  {
+                    map.put(rest, WORKING);
+                    pair = (Pair) rest;
+                    continue;
+                  }
               }
             cdr = expand(rest, depth, syntax, seen, tr);
+            if (cdr == rest)
+              return list;
             break;
           }
         cdr = expand (pair.getCdr(), depth, syntax, seen, tr);
@@ -286,8 +302,10 @@ public class Quote extends Syntax
     return result;
   }
 
+  // Note in 'seen' map that datum is currently being expanded.
   private static final Object WORKING = new String("(working)");
-  private static final Object CYCLE = new String("(cycle)");
+  // Note in 'seen' map that datum is used multiple times, partly in cycle.
+  private static final Object SHARED = new String("(shared)");
 
   /** Backquote-expand a template.
    * @param template the quasiquoted template to expand
@@ -306,15 +324,16 @@ public class Quote extends Syntax
     Object old = map.get(template);
     if (old == WORKING)
       {
-        map.put(template, CYCLE);
-        return old;
+        map.put(template, SHARED);
+        return template;
       }
-    else if (old == CYCLE)
+    else if (old == SHARED)
       {
-        return old;
+        return template;
       }
     else if (old != null)
       return old;
+    map.put(template, WORKING);
     /* #endif */
     Object result;
     if (template instanceof Pair)
@@ -342,8 +361,7 @@ public class Quote extends Syntax
 	    int element_depth = depth;
 	    Pair pair;
 	    if (element instanceof Pair && depth > QUOTE_DEPTH
-		&& tr.matches((pair = (Pair)element).getCar(), syntax,
-			      LispLanguage.unquotesplicing_sym)
+		&& matchesUnquoteSplicing((pair = (Pair)element), syntax, tr)
 		&& --element_depth == 0)
 	      {
 		Pair pair_cdr;
@@ -370,7 +388,7 @@ public class Quote extends Syntax
 	if (max_state == 0)
 	  result = vector;
 	else if (max_state == 1)
-	  result = new FVector (buffer);
+	  result = new ConstVector<Object>(buffer);
 	else
 	  {
 	    Expression[] args = new Expression[n];
@@ -384,7 +402,7 @@ public class Quote extends Syntax
 		  {
 		    Object[] arg1 = new Object[1];
 		    arg1[0] = buffer[i];
-		    args[i] = leaf(new FVector (arg1), tr);
+		    args[i] = leaf(new ConstVector<Object>(arg1), tr);
 		  }
 		else
 		  {
@@ -402,7 +420,7 @@ public class Quote extends Syntax
     else
       result = template;
     /* #ifdef use:java.util.IdentityHashMap */ 
-    if (template != result && map.get(template) == CYCLE)
+    if (template != result && map.get(template) == SHARED)
       tr.error('e', "cycle in non-literal data");
     map.put(template, result);
     /* #endif */
@@ -474,11 +492,11 @@ public class Quote extends Syntax
 
   static final Method vectorAppendMethod
   = ClassType.make("kawa.standard.vector_append")
-    .getDeclaredMethod("apply$V", 1);
+    .getDeclaredMethod("appendToConstVector", 1);
   static final ClassType quoteType = ClassType.make("kawa.lang.Quote");
   static final Method consXMethod = quoteType.getDeclaredMethod("consX$V", 1);
   static final Method appendMethod = quoteType.getDeclaredMethod("append$V", 1);
   static final Method makePairMethod = Compilation.typePair.getDeclaredMethod("make", 2);
-  static final Method makeVectorMethod = ClassType.make("gnu.lists.FVector")
+  static final Method makeVectorMethod = ClassType.make("gnu.lists.ConstVector")
     .getDeclaredMethod("make", 1);
 }
