@@ -4,6 +4,7 @@ import gnu.expr.*;
 import gnu.kawa.reflect.*;
 import gnu.bytecode.Type;
 import gnu.bytecode.ClassType;
+import gnu.bytecode.Field;
 import gnu.bytecode.ArrayClassLoader;
 import gnu.bytecode.ZipLoader;
 import gnu.text.SourceMessages;
@@ -40,12 +41,20 @@ public class Translator extends Compilation
 
   public Declaration templateScopeDecl;
 
+  /** A "mark" created for the current macro application.
+   * This is (more-or-less) the mark specified by the syntax-case
+   * specification (in r6rs-lib), applied to the output of a transformer.
+   * However, instead of "applying" a mark to the transformer output,
+   * we remember in the TemlateScope an object unique to the application.
+   */
+  Object currentMacroMark = null;
+
   /** A variable to hold the matched values for syntax-case
    * pattern variables. */
   public Declaration matchArray;
 
   /** A stack of aliases pushed by <code>pushRenamedAlias</code>. */
-  Stack renamedAliasStack;
+  private Stack<Declaration> renamedAliasStack;
 
   public Stack formStack = new Stack();
   public int firstForm;
@@ -413,7 +422,12 @@ public class Translator extends Compilation
         // FIXME don't copy the args array in makeExp ...
         return CompileNamedPart.makeExp(part1, part2);
       }
-    return ((LispLanguage) getLanguage()).makeApply(func, args);
+    return makeApply(func, args);
+  }
+
+  public ApplyExp makeApply (Expression func, Expression[] args)
+  { 
+    return new ApplyExp(func, args);
   }
 
   public Namespace namespaceResolvePrefix (Expression context)
@@ -553,14 +567,6 @@ public class Translator extends Compilation
       formStack.add(rewrite(exp, false));
   }
 
-  /**
-   * Re-write a Scheme expression in S-expression format into internal form.
-   */
-  public Expression rewrite (Object exp)
-  {
-    return rewrite(exp, false);
-  }
-
   public Object namespaceResolve (Object name)
   {
     if (! (name instanceof SimpleSymbol))
@@ -584,10 +590,27 @@ public class Translator extends Compilation
     return name;
   }
 
+  /**
+   * Re-write a Scheme expression in S-expression format into internal form.
+   */
+  public Expression rewrite (Object exp)
+  {
+    return rewrite(exp, 'N');
+  }
+
   /**
    * Re-write a Scheme expression in S-expression format into internal form.
    */
   public Expression rewrite (Object exp, boolean function)
+  {
+    return rewrite(exp, function ? 'F' : 'N');
+  }
+
+  /** Re-write a Scheme expression in S-expression format into internal form.
+   * @param mode either 'N' (normal), 'F' (function application context),
+   *  or 'M' (macro-checking).
+   */
+  public Expression rewrite (Object exp, char mode)
   {
     if (exp instanceof SyntaxForm)
       {
@@ -596,7 +619,7 @@ public class Translator extends Compilation
 	try
 	  {
 	    setCurrentScope(sf.getScope());
-	    Expression s = rewrite(sf.getDatum(), function);
+	    Expression s = rewrite(sf.getDatum(), mode);
 	    return s;
 	  }
 	finally
@@ -604,6 +627,7 @@ public class Translator extends Compilation
 	    setCurrentScope(save_scope);
 	  }
       }
+    boolean function = mode != 'N';
     if (exp instanceof PairWithPosition)
       return rewrite_with_position (exp, function, (PairWithPosition) exp);
     else if (exp instanceof Pair)
@@ -644,10 +668,10 @@ public class Translator extends Compilation
                          && caller.nameDecl.isStatic()));
                 if (part == null)
                   {
-                    char mode = contextStatic ? 'S' : 'V';
                     PrimProcedure[] methods
                       = ClassMethods.getMethods(ctype, dname,
-                                                mode, ctype, language);
+                                                contextStatic ? 'S' : 'V',
+                                                ctype, language);
                     if (methods.length == 0)
                       continue;
                   }
@@ -721,10 +745,10 @@ public class Translator extends Compilation
                         // A kludge - we get a bunch of testsuite failures
                         // if we don't inline $lookup$.  FIXME.
                         && (decl != getNamedPartDecl
-                            && ! ("objectSyntax".equals(floc.getMemberName())
-                                  // Another kludge to support "object" as a
-                                  // type specifier.
-                                  && "kawa.standard.object".equals(floc. getDeclaringClass().getName()))))
+                            // Another kludge to support "object" as a
+                            // type specifier.
+                            && ! isObjectSyntax(floc.getDeclaringClass(),
+                                                floc.getMemberName())))
                       decl = null;
                     else if (immediate)
                       {
@@ -758,10 +782,9 @@ public class Translator extends Compilation
                     decl = null;
                   }
               }
-            else if (loc == null || ! loc.isBound())
+            else if (mode != 'M' && (loc == null || ! loc.isBound()))
               {
-                Expression e
-                  = ((LispLanguage) getLanguage()).checkDefaultBinding(symbol, this);
+                Expression e = checkDefaultBinding(symbol, this);
                 if (e != null)
                   return e;
               }
@@ -780,8 +803,10 @@ public class Translator extends Compilation
             // A special kludge to deal with the overloading between the
             // object macro and object as being equivalent to java.lang.Object.
             // A cleaner solution would be to use an identifier macro.
-            if (! function
-                && decl.getConstantValue() instanceof kawa.standard.object)
+            Field dfield = decl.field;
+            if (! function && dfield != null
+                && isObjectSyntax(dfield.getDeclaringClass(),
+                                  dfield.getName()))
               return QuoteExp.getInstance(Object.class);
 
             if (decl.getContext() instanceof PatternScope)
@@ -801,10 +826,21 @@ public class Translator extends Compilation
       return (Expression) exp;
     else if (exp == Special.abstractSpecial)
       return QuoteExp.abstractExp;
+    else if (exp == Special.nativeSpecial)
+      return QuoteExp.nativeExp;
     else
       return QuoteExp.getInstance(Quote.quote(exp, this), this);
   }
 
+  /** If a symbol is lexically unbound, look for a default binding.
+   * The default implementation does nothing.
+   * @return null if no binidng, or an Expression.
+   */
+  public Expression checkDefaultBinding (Symbol name, Translator tr)
+  {
+    return null;
+  }
+
   public static void setLine(Expression exp, Object location)
   {
     if (location instanceof SourceLocator)
@@ -835,7 +871,7 @@ public class Translator extends Compilation
 	|| positionPair.getLineNumber() != getLineNumber()
 	|| positionPair.getColumnNumber() != getColumnNumber())
       {
-        saved = new PairWithPosition(this, Special.eof, positionPair);
+        saved = new PairWithPosition(this, this, positionPair);
       }
     else
       saved = positionPair;
@@ -853,7 +889,7 @@ public class Translator extends Compilation
       return;
     setLine(saved);
     positionPair = (PairWithPosition) saved;
-    if (positionPair.getCar() == Special.eof)
+    if (positionPair.getCar() == this)
       positionPair = (PairWithPosition) positionPair.getCdr();
   }
 
@@ -870,11 +906,16 @@ public class Translator extends Compilation
 
   /** Extract a type from the car of a pair. */
   public Type exp2Type(Pair typeSpecPair)
+  {
+    return exp2Type(typeSpecPair, null, null);
+  }
+
+  public Type exp2Type(Pair typeSpecPair, Declaration decl, SyntaxForm syntax)
   {
     Object saved = pushPositionOf(typeSpecPair);
     try
       {
-	Expression texp = rewrite_car(typeSpecPair, false);
+	Expression texp = rewrite_car(typeSpecPair, syntax);
         texp = InlineCalls.inlineCalls(texp, this);
 	if (texp instanceof ErrorExp)
 	  return null;
@@ -901,9 +942,11 @@ public class Translator extends Compilation
 	     else
 	       error('e',
 		 "invalid type spec (must be \"type\" or 'type or <type>)");
-	     return Type.pointer_type;
+	     type = Type.pointer_type;
 	   }
-	 return type;
+        if (decl != null)
+          decl.setType(texp, type);
+        return type;
       }
     finally
       {
@@ -1036,7 +1079,7 @@ public class Translator extends Compilation
             
             if (obj instanceof Symbol && ! selfEvaluatingSymbol(obj))
               {
-                Expression func = rewrite(obj, true);
+                Expression func = rewrite(obj, 'M');
                 if (func instanceof ReferenceExp)
                   {
                     Declaration decl = ((ReferenceExp) func).getBinding();
@@ -1051,7 +1094,7 @@ public class Translator extends Compilation
                   }
               }
             // Recognize deferred begin created in scanBody for pendingForms.
-            // A seemingly-cleaned (obj instanceof Syntax) causes problems
+            // A seemingly-cleaner (obj instanceof Syntax) causes problems
             // with some Syntax forms, such as define.
             else if (obj == kawa.standard.begin.begin)
               syntax = (Syntax) obj;
@@ -1136,6 +1179,8 @@ public class Translator extends Compilation
                 if (pair.getCar() != pendingForm)
                   pair = makePair(pair, pendingForm, pair.getCdr());
                 pendingForm = new Pair(kawa.standard.begin.begin, pair);
+                if (list != LList.Empty)
+                  formStack.add(new Pair(kawa.standard.begin.begin, list));
                 return LList.Empty;
               }
 	    int fsize = formStack.size();
@@ -1180,7 +1225,7 @@ public class Translator extends Compilation
     // NOTE we have both a rewrite_body and a rewriteBody.
     // This is confusing, at the least.  FIXME.
     Object saved = pushPositionOf(exp);
-    LetExp defs = new LetExp(null);
+    LetExp defs = new LetExp();
     int first = formStack.size();
     defs.outer = current_scope;
     current_scope = defs;
@@ -1189,20 +1234,21 @@ public class Translator extends Compilation
         LList list = scanBody(exp, defs, true);
 	if (list.isEmpty())
 	  formStack.add(syntaxError ("body with no expressions"));
-	int ndecls = defs.countNonDynamicDecls();
-	if (ndecls != 0)
-	  {
-	    Expression[] inits = new Expression[ndecls];
-	    for (int i = ndecls;  --i >= 0; )
-	      inits[i] = QuoteExp.undefined_exp;
-	    defs.inits = inits;
-	  }
+        int ndecls = 0;
+        for (Declaration decl = defs.firstDecl(); decl != null; decl = decl.nextDecl())
+          {
+            if (! decl.getFlag(Declaration.IS_DYNAMIC))
+              {
+                ndecls++;
+                decl.setInitValue(QuoteExp.undefined_exp);
+              }
+          }
         rewriteBody(list);
 	Expression body = makeBody(first, null);
 	setLineOf(body);
 	if (ndecls == 0)
 	  return body;
-	defs.body = body;
+	defs.setBody(body);
 	setLineOf(defs);
 	return defs;
       }
@@ -1251,10 +1297,21 @@ public class Translator extends Compilation
 	  return new ApplyExp(gnu.kawa.functions.AppendValues.appendValues,
 			      exps);
 	else
-	  return ((LispLanguage) getLanguage()).makeBody(exps);
+	  return makeBody(exps);
       }
   }
 
+  public boolean appendBodyValues () { return false; }
+
+  /** Combine a <body> consisting of a list of expression. */
+  public Expression makeBody(Expression[] exps)
+  {
+    if (appendBodyValues())
+      return new ApplyExp(gnu.kawa.functions.AppendValues.appendValues, exps);
+    else
+      return new BeginExp (exps);
+  }
+
   /** Storage used by noteAccess and processAccesses. */
   Vector notedAccess;
 
@@ -1328,7 +1385,7 @@ public class Translator extends Compilation
 		    decl.setPrivate(false);
 		  }
 	      }
-	    else
+	    else if (! kawa.standard.IfFeature.isProvide(decl))
 	      decl.setPrivate(true);
 	  }
 	if (moduleStatic)
@@ -1448,22 +1505,21 @@ public class Translator extends Compilation
    * So that such references can resolve to <code>decl</code>, we
    * create an alias in <code>templateScope</code> that points
    * to <code>decl</code>.  We record that we did this in the
-   * <code> renamedLiasStack</code>, so we can remove the alias later.
+   * <code>renamedAliasStack</code>, so we can remove the alias later.
    */
   public void pushRenamedAlias (Declaration alias)
   {
     Declaration decl = getOriginalRef(alias).getBinding();
     ScopeExp templateScope = alias.context;
     decl.setSymbol(null);
-    Declaration old = templateScope.lookup(decl.getSymbol());
+    Declaration old = templateScope.lookup(alias.getSymbol());
     if (old != null)
       templateScope.remove(old);
     templateScope.addDeclaration(alias);
     if (renamedAliasStack == null)
-      renamedAliasStack = new Stack();
+      renamedAliasStack = new Stack<Declaration>();
     renamedAliasStack.push(old);
     renamedAliasStack.push(alias);
-    renamedAliasStack.push(templateScope);
   }
 
   /** Remove one or more aliases created by <code>pushRenamedAlias</code>. */
@@ -1471,14 +1527,14 @@ public class Translator extends Compilation
   {
     while (--count >= 0)
       {
-	ScopeExp templateScope = (ScopeExp) renamedAliasStack.pop();
 	Declaration alias = (Declaration) renamedAliasStack.pop();
+        ScopeExp templateScope = alias.getContext();
 	Declaration decl = getOriginalRef(alias).getBinding();
 	decl.setSymbol(alias.getSymbol());
 	templateScope.remove(alias);
-	Object old = renamedAliasStack.pop();
+	Declaration old = renamedAliasStack.pop();
 	if (old != null)
-	  templateScope.addDeclaration((Declaration) old);
+	  templateScope.addDeclaration(old);
       }
   }
 
@@ -1495,4 +1551,10 @@ public class Translator extends Compilation
     push(decl);
     return decl;
   }
+
+  static boolean isObjectSyntax (ClassType declaringClass, String fieldName)
+  {
+    return "objectSyntax".equals(fieldName)
+      &&  "kawa.standard.object".equals(declaringClass.getName());
+  }
 }
