@@ -12,13 +12,13 @@
 (require <kawa.lib.reflection>)
 
 (define (pair? x) :: <boolean>
-  (<pair>:instance? x))
+  (<pair>:instance? (gnu.mapping.Promise:force x)))
 
 (define (cons car cdr) :: <pair>
   (<pair>:new car cdr))
 
 (define (null? x) :: <boolean>
-  (eq? x '()))
+  (eq? (gnu.mapping.Promise:force x) '()))
 
 (define (set-car! (p <pair>) x)
   (set! p:car x))
@@ -42,25 +42,25 @@
        ((_ fname slots)
        #`(define-procedure fname
 	   setter: (lambda (arg value)
-		     ,(syntax-case #'slots ()
+		     #,(syntax-case #'slots ()
 			((first1 . rest1)
 			 #`(set!
 			    (field 
-			     ,(let loop ((f #'rest1))
+			     #,(let loop ((f #'rest1))
 				(syntax-case f ()
 				  (() #'arg)
 				  ((first . rest)
-				   #`(field (as <pair> ,(loop #'rest))
-					    ',(syntax-object->datum #'first)))))
+				   #`(field (as <pair> #,(loop #'rest))
+					    '#,(syntax-object->datum #'first)))))
 			     'first1)
 			    value))))
 	   (lambda (arg) name: 'fname
-		   ,(let loop ((f #'slots))
+		   #,(let loop ((f #'slots))
 		      (syntax-case f ()
 			(() #'arg)
 			((first . rest)
-			 #`(field (as <pair> ,(loop #'rest))
-				  ',(syntax-object->datum #'first)))))))))))
+			 #`(field (as <pair> #,(loop #'rest))
+				  '#,(syntax-object->datum #'first)))))))))))
 
 (define-cxr caar (car car))
 (define-cxr cadr (car cdr))
@@ -101,7 +101,15 @@
 	  (loop (cdr pair) (cons (car pair) result))))))
 
 (define (list-tail list (count :: <int>))
-  (<list>:listTail list count))
+  (let loop ((lst list))
+    (set! count (- count 1))
+    (cond ((< count 0)
+           lst)
+          (else
+           (let ((flst (gnu.mapping.Promise:force lst)))
+             (if (gnu.lists.Pair? flst)
+                 (loop ((as gnu.lists.Pair flst):getCdr))
+                 (primitive-throw (java.lang.IndexOutOfBoundsException "List is too short."))))))))
 
 (define (list-ref list (index :: <int>))
   (car (list-tail list index)))
@@ -154,10 +162,10 @@
 	      (lp (cdr list)))))))
 
 ;;;  The optional test argument is an srfi-1 extension.
-(define (assoc x list #!optional (test :: <procedure> equal?))
+(define (assoc key list #!optional (test :: <procedure> equal?))
   (let lp ((list list))
     (if (eq? list '())
 	 #f
 	(let ((pair :: <pair> (car list)))
-	  (if (test pair:car x) pair
+	  (if (test key pair:car) pair
 	      (lp (cdr list)))))))
