@@ -65,8 +65,8 @@
   (gnu.text.Char:print (char->integer ch) port))
 
 ;; SRFI-6
-(define (open-input-string (str :: <string>)) :: <input-port>
-  (make <gnu.mapping.CharArrayInPort> str))
+(define (open-input-string (str ::string)) ::string-input-port
+  (gnu.mapping.CharArrayInPort str))
 
 (define (open-output-string) :: <string-output-port>
   (<string-output-port>))
@@ -75,8 +75,7 @@
   (<gnu.lists.FString> (output-port:toCharArray)))
 
 (define (call-with-input-string (str :: <string>) (proc :: <procedure>))
-  (let* ((port :: <gnu.mapping.CharArrayInPort>
-	  (make <gnu.mapping.CharArrayInPort> str))
+  (let* ((port ::string-input-port (gnu.mapping.CharArrayInPort str))
 	 (result (proc port)))
     (close-input-port port)
     result))
@@ -104,9 +103,24 @@
 (define (char-ready? #!optional (port (current-input-port)))
   (invoke-static <kawa.standard.char_ready_p> 'ready port))
 
-(define (write value #!optional (out (current-output-port))) :: <void>
-  (*:format (kawa.standard.Scheme:.writeFormat) value out))
+(define (write value #!optional (out ::output-port (current-output-port))) ::void
+  (if *print-circle*
+      (write-with-shared-structure value out)
+      (*:format (kawa.standard.Scheme:.writeFormat) value out)))
 
+(define (write-with-shared-structure 
+	 value #!optional (out ::output-port (current-output-port))) ::void
+  (let ((pretty-out (out:getPrettyWriter)))
+    (pretty-out:initialiseIDHash)
+    (pretty-out:setSharing #t)
+    (try-finally
+     (*:format (kawa.standard.Scheme:.sharedWriteFormat) value out)
+     (pretty-out:setSharing #f))
+    (pretty-out:clearIDHash)
+    (pretty-out:writeEndOfExpression)
+    (pretty-out:resolveBackReferences)
+    (pretty-out:flush)))
+    
 (define (display value #!optional (out (current-output-port))) :: <void>
   (*:format (kawa.standard.Scheme:.displayFormat) value out))
 
