@@ -5,6 +5,7 @@ import gnu.mapping.*;
 import gnu.expr.*;
 import gnu.text.SourceMessages;
 import gnu.text.SyntaxException;
+import gnu.text.Path;
 import gnu.lists.*;
 import java.util.ArrayList;
 import java.util.Vector;
@@ -124,27 +125,58 @@ public class repl extends Procedure0or1
     /* Set homeDirectory;  if first time called, run ~/.kawarc.scm. */
     if (homeDirectory == null)
       {
-	File initFile = null;
-	homeDirectory = System.getProperty ("user.home");
-	Object scmHomeDirectory;
-	if (homeDirectory != null)
-	  {
-	    scmHomeDirectory = new FString (homeDirectory);
-	    String file_separator = System.getProperty("file.separator");
-	    String kawarc_name =
-	      "/".equals(file_separator) ? ".kawarc.scm"
-	      : "kawarc.scm";
-	    initFile = new File(homeDirectory, kawarc_name);
-	  }
-	else
-	  scmHomeDirectory = Boolean.FALSE;
-	Environment.getCurrent().put("home-directory", scmHomeDirectory);
+	File initFile = getInitFile();
 	if (initFile != null && initFile.exists())
 	  if (! Shell.runFileOrClass(initFile.getPath(), true, 0))
-            System.exit(-1);
+	    System.exit(-1);
       }
   }
 
+  static File getInitFile()
+  {
+    File initFile = null;
+    homeDirectory = System.getProperty ("user.home");
+    Object scmHomeDirectory;
+    if (homeDirectory != null)
+      {
+	scmHomeDirectory = new FString (homeDirectory);
+	String file_separator = System.getProperty("file.separator");
+	String kawarc_name = new String();
+	if (Language.getDefaultLanguage().getName().equals("Emacs-Lisp"))
+	    kawarc_name = ".jemacs";
+	else
+	    kawarc_name =
+	  "/".equals(file_separator) ? ".kawarc.scm"
+	  : "kawarc.scm";
+	initFile = new File(homeDirectory, kawarc_name);
+      }
+    else
+      scmHomeDirectory = Boolean.FALSE;
+    Environment.getCurrent().put("home-directory", scmHomeDirectory);
+    return initFile;
+  }
+
+  /** Evaluate init file, if any, returning error message on failure. */
+  public static String evalInitFileWithErrorMessage ()
+  {
+    File initFile = getInitFile();
+    if (initFile != null && initFile.exists())
+      {
+        try 
+          {
+            Path path = Path.valueOf(initFile.getPath());
+            InputStream fs = path.openInputStream();
+            Environment env = Environment.getCurrent();
+            Shell.runFile(fs, path, env, true, 0);
+          }
+        catch (Throwable e)
+          {
+            return "An error occurred while loading '" + initFile +"' : " + e;
+          }
+      }
+    return null;
+  }
+
   public static void setArgs (String[] args, int arg_start)
   {
     ApplicationMainSupport.setArgs(args, arg_start);
@@ -367,7 +399,7 @@ public class repl extends Procedure0or1
             ModuleManager manager = ModuleManager.getInstance();
 	    manager.setCompilationDirectory(args[iArg]);
 	  }
-        else if (arg.equals("--target") || arg.equals("target"))
+        else if (arg.equals("--target") || arg.equals("-target"))
           {
 	    iArg++;
 	    if (iArg == maxArg)
@@ -590,6 +622,8 @@ public class repl extends Procedure0or1
 	  {
             SourceMessages.debugStackTraceOnWarning = true;
 	  }
+        else if (arg.equals("--diagnostic-strip-directories"))
+          SourceMessages.stripDirectoriesDefault = true;
 	else if (arg.equals("--module-nonstatic")
                  || arg.equals("--no-module-static"))
 	  {
@@ -648,7 +682,7 @@ public class repl extends Procedure0or1
 	    System.out.print("Kawa ");
 	    System.out.print(Version.getVersion());
 	    System.out.println();
-	    System.out.println("Copyright (C) 2009 Per Bothner");
+	    System.out.println("Copyright (C) 2011 Per Bothner");
 	    something_done = true;
 	  }
 	else if (arg.length () > 0 && arg.charAt(0) == '-')
@@ -668,7 +702,7 @@ public class repl extends Procedure0or1
 	    else
 	      {
 		// See if arg is a valid Compilation option, and if so set it.
-		int eq = name.indexOf("=");
+		int eq = name.indexOf('=');
 		String opt_value;
 		if (eq < 0)
 		  opt_value = null;
@@ -681,19 +715,26 @@ public class repl extends Procedure0or1
 		// Convert "--no-xxx" to "--xxx=no":
 		boolean startsWithNo
 		  = name.startsWith("no-") && name.length() > 3;
-		if (opt_value == null && startsWithNo)
+		boolean addedNo = false;
+                if (opt_value == null && startsWithNo)
 		  {
 		    opt_value = "no";
 		    name = name.substring(3);
+                    addedNo = true;
 		  }
 
 		String msg = Compilation.options.set(name, opt_value);
 		if (msg != null)
 		  {
-		    // It wasn't a valid Complation option.
-		    if (startsWithNo && msg == gnu.text.Options.UNKNOWN)
-		      msg = "both '--no-' prefix and '="+
-			opt_value+"' specified";
+		    if (msg == gnu.text.Options.UNKNOWN)
+                      {
+                        if (addedNo)
+                          msg = "unknown option '"+name+"'";
+                        // It wasn't a valid Compilation option.
+                        else if (startsWithNo)
+                          msg = "both '--no-' prefix and '="+
+                            opt_value+"' specified";
+                      }
 		    if (msg == gnu.text.Options.UNKNOWN)
 		      {
 			bad_option(arg);
@@ -739,22 +780,22 @@ public class repl extends Procedure0or1
                 break; // Kludge to shut up compiler.
               }
             
-            comp
-              = language.parse(fstream, messages,
-                               defaultParseOptions);
-
+            ModuleInfo minfo = manager.findWithSourcePath(arg);
+    
             if (compilationTopname != null)
               {
                 String cname
                   = Compilation.mangleNameIfNeeded(compilationTopname);
-                ClassType ctype = new ClassType(cname);
-                ModuleExp mexp = comp.getModule();
-                mexp.setType(ctype);
-                mexp.setName(compilationTopname);
-                comp.mainClass = ctype;
+                if (Compilation.classPrefixDefault != null)
+                  cname = Compilation.classPrefixDefault + cname;
+                minfo.setClassName(cname);
               }
-            
-            infos[i-iArg] = manager.find(comp);
+
+            comp
+              = language.parse(fstream, messages,
+                               defaultParseOptions, minfo);
+            comp.getModule().classFor(comp);
+            infos[i-iArg] = minfo;
             comps[i-iArg] = comp;
 
           }
@@ -800,6 +841,8 @@ public class repl extends Procedure0or1
 
   static void internalError (Throwable ex, Compilation comp, Object arg)
   {
+    try { comp.getMessages().checkErrors(System.err, 50); }
+    catch (Throwable e) { }
     StringBuffer sbuf = new StringBuffer();
     if (comp != null)
       {
