@@ -1,6 +1,7 @@
 package kawa.standard;
 import kawa.lang.*;
 import gnu.expr.*;
+import gnu.mapping.Symbol;
 import gnu.mapping.SimpleSymbol;
 
 public class IfFeature
@@ -12,6 +13,7 @@ public class IfFeature
 	SyntaxForm sf = (SyntaxForm) form;
 	form = sf.getDatum();
       }
+    form = ((Translator) Compilation.getCurrent()).namespaceResolve(form);
     if (form instanceof String || form instanceof SimpleSymbol)
       return hasFeature(form.toString());
     return false;  // FIXME - return error
@@ -76,12 +78,42 @@ public class IfFeature
         if (name == "in-servlet")
           return (mflags & ModuleContext.IN_SERVLET) != 0;
       }
+    /* #ifdef JAVA6 */
+    if (name == "java-6")
+      return true;
+    /* #endif */
 
-    String provide_name = ("%provide%"+name).intern();
-    Compilation comp = Compilation.getCurrent();
-    Declaration decl = comp.lookup(provide_name, -1);
+    /* #ifdef JAVA7 */
+    // if (name == "java-7")
+    //  return true;
+    /* #endif */
+
+    String classExistsPrefix = "class-exists:";
+    if (name.startsWith(classExistsPrefix))
+      {
+        name = name.substring(classExistsPrefix.length());
+        try
+          {
+            Class.forName(name, false, IfFeature.class.getClassLoader());
+            return true;
+          }
+        catch (ClassNotFoundException ex)
+          {
+            return false;
+          }
+      }
+
+    Symbol provide_symbol = Symbol.valueOf(PROVIDE_PREFIX+name);
+    Declaration decl = Compilation.getCurrent().lookup(provide_symbol, -1);
     if (decl!=null && ! decl.getFlag(Declaration.IS_UNKNOWN))
       return true;
     return false;
   }
+
+  public static final String PROVIDE_PREFIX = "%provide%";
+
+  public static boolean isProvide (Declaration decl)
+  {
+    return decl.getName().startsWith(PROVIDE_PREFIX);
+  }
 }
