@@ -14,9 +14,8 @@ import gnu.lists.AbstractFormat;
 import gnu.kawa.functions.*;
 import gnu.kawa.reflect.ClassMethods;
 import gnu.kawa.reflect.StaticFieldLocation;
+import gnu.kawa.reflect.LazyType;
 import gnu.math.DFloNum;
-import gnu.kawa.xml.XmlNamespace;
-import gnu.math.Unit;
 import gnu.kawa.servlet.HttpRequestContext;
 
 public class Scheme extends LispLanguage
@@ -24,10 +23,22 @@ public class Scheme extends LispLanguage
   public static final Environment nullEnvironment;
   public static final Environment r4Environment;
   public static final Environment r5Environment;
+  public static final Environment r6Environment;
   protected static final SimpleEnvironment kawaEnvironment;
 
   public static LangPrimType booleanType;
+
+  public static final int FOLLOW_R5RS = 5;
+  public static final int FOLLOW_R6RS = 6;
+  public static final int FOLLOW_R7RS = 7;
+
   public static final Scheme instance;
+  private static Scheme r5rsInstance;
+  private static Scheme r6rsInstance;
+  private static Scheme r7rsInstance;
+  int standardToFollow;
+
+  public int getStandardToFollow() { return standardToFollow; }
 
   public static final gnu.kawa.reflect.InstanceOf instanceOf;
   public static final Not not;
@@ -47,28 +58,28 @@ public class Scheme extends LispLanguage
 
   public static final Apply apply;
   public static final ApplyToArgs applyToArgs;
-  static final Declaration applyFieldDecl;
+
+  private static final String[] uniformVectorTags =
+    {"s8", "s16", "s32", "s64", "u8", "u16", "u32", "u64", "f32", "f64" };
 
   static {
     // (null-environment)
     nullEnvironment = Environment.make("null-environment");
     r4Environment = Environment.make("r4rs-environment", nullEnvironment);
     r5Environment = Environment.make("r5rs-environment", r4Environment);
-    kawaEnvironment = Environment.make("kawa-environment", r5Environment);
+    r6Environment = Environment.make("r6rs-environment", r5Environment);
+    kawaEnvironment = Environment.make("kawa-environment", r6Environment);
 
     instance = new Scheme(kawaEnvironment);
     instanceOf = new gnu.kawa.reflect.InstanceOf(instance, "instance?");
     not = new Not(instance, "not");
     applyToArgs = new ApplyToArgs("apply-to-args", instance);
-    applyFieldDecl
-      = Declaration.getDeclarationFromStatic("kawa.standard.Scheme",
-                                             "applyToArgs");
     apply = new Apply("apply", applyToArgs);
     isEq = new gnu.kawa.functions.IsEq(instance, "eq?");
     isEqv = new gnu.kawa.functions.IsEqv(instance, "eqv?", isEq);
     isEqual = new gnu.kawa.functions.IsEqual(instance, "equal?");
-    map = new gnu.kawa.functions.Map(true, applyToArgs, applyFieldDecl, isEq);
-    forEach = new gnu.kawa.functions.Map(false, applyToArgs, applyFieldDecl, isEq);
+    map = new gnu.kawa.functions.Map(true, applyToArgs, isEq);
+    forEach = new gnu.kawa.functions.Map(false, applyToArgs, isEq);
     numEqu = NumberCompare.make(instance, "=",
                                 NumberCompare.TRUE_IF_EQU);
     numGrt = NumberCompare.make(instance, ">",
@@ -103,6 +114,34 @@ public class Scheme extends LispLanguage
     return instance;
   }
 
+  private static Scheme newStandardInstance (int standardToFollow)
+  {
+    Scheme instance = new Scheme(kawaEnvironment);
+    instance.standardToFollow = standardToFollow;
+    return instance;
+  }
+
+  public static synchronized Scheme getR5rsInstance()
+  {
+    if (r5rsInstance == null)
+      r5rsInstance = newStandardInstance(FOLLOW_R5RS);
+    return r5rsInstance;
+  }
+
+  public static synchronized Scheme getR6rsInstance()
+  {
+    if (r6rsInstance == null)
+      r6rsInstance = newStandardInstance(FOLLOW_R6RS);
+    return r6rsInstance;
+  }
+
+  public static synchronized Scheme getR7rsInstance()
+  {
+    if (r7rsInstance == null)
+      r7rsInstance = newStandardInstance(FOLLOW_R7RS);
+    return r7rsInstance;
+  }
+
   public static Environment builtin ()
   {
     return kawaEnvironment;
@@ -115,6 +154,7 @@ public class Scheme extends LispLanguage
       environ.addLocation(LispLanguage.lookup_sym, null, getNamedPartLocation);
 
       defSntxStFld("lambda", "kawa.standard.SchemeCompilation", "lambda");
+      defSntxStFld("$bracket-apply$", "gnu.kawa.lispexpr.BracketApply", "instance");
 
       //-- Section 4.1  -- complete
       defSntxStFld(LispLanguage.quote_sym, "kawa.lang.Quote", "plainQuote");
@@ -136,8 +176,8 @@ public class Scheme extends LispLanguage
 
       defSntxStFld("begin", "kawa.standard.begin", "begin");
       defSntxStFld("do", "kawa.lib.std_syntax");
+      defSntxStFld("lazy", "kawa.lib.std_syntax");
       defSntxStFld("delay", "kawa.lib.std_syntax");
-      defProcStFld("%make-promise", "kawa.lib.std_syntax");
       defSntxStFld("quasiquote", "kawa.lang.Quote", "quasiQuote");
 
       //-- Section 5  -- complete [except for internal definitions]
@@ -372,8 +412,14 @@ public class Scheme extends LispLanguage
       defProcStFld("for-each", "kawa.standard.Scheme", "forEach");
       defProcStFld("call-with-current-continuation",
                    "gnu.kawa.functions.CallCC", "callcc");
-      defProcStFld("call/cc", "kawa.standard.callcc", "callcc");
+      defProcStFld("call/cc", "gnu.kawa.functions.CallCC", "callcc");
       defProcStFld("force", "kawa.lib.misc");
+      defProcStFld("force*", "kawa.lib.misc");
+      defProcStFld("eager", "kawa.lib.misc");
+      defProcStFld("promise-set-value!", "kawa.lib.misc");
+      defProcStFld("promise-set-alias!", "kawa.lib.misc");
+      defProcStFld("promise-set-exception!", "kawa.lib.misc");
+      defProcStFld("promise-set-thunk!", "kawa.lib.misc");
 
       //-- Section 6.10  -- complete
       defProcStFld("call-with-input-file", "kawa.lib.ports");
@@ -395,6 +441,7 @@ public class Scheme extends LispLanguage
       defProcStFld("eof-object?", "kawa.lib.ports");
       defProcStFld("char-ready?", "kawa.lib.ports");
       defProcStFld("write", "kawa.lib.ports");
+      defProcStFld("write-with-shared-structure", "kawa.lib.ports");
       defProcStFld("display", "kawa.lib.ports");
       defProcStFld("print-as-xml", "gnu.xquery.lang.XQuery", "writeFormat");
       defProcStFld("write-char", "kawa.lib.ports");
@@ -437,7 +484,7 @@ public class Scheme extends LispLanguage
       defSntxStFld("let*-values", "kawa.lib.syntax");
       defSntxStFld("case-lambda", "kawa.lib.syntax");
       defSntxStFld("receive", "kawa.lib.syntax");
-      defProcStFld("eval", "kawa.lang.Eval");
+      defProcStFld("eval", "kawa.lib.std_syntax");
       defProcStFld("repl", "kawa.standard.SchemeCompilation", "repl");
       defProcStFld("scheme-report-environment", "kawa.lib.misc");
       defProcStFld("null-environment", "kawa.lib.misc");
@@ -445,6 +492,22 @@ public class Scheme extends LispLanguage
       defProcStFld("dynamic-wind", "kawa.lib.misc");
 
       r5Environment.setLocked();
+      environ = r6Environment;
+
+      defProcStFld("vector-map", "kawa.lib.vectors");
+      defProcStFld("vector-for-each", "kawa.lib.vectors");
+      defProcStFld("string-for-each", "gnu.kawa.slib.srfi13");
+      defProcStFld("real-valued?", "kawa.lib.numbers");
+      defProcStFld("rational-valued?", "kawa.lib.numbers");
+      defProcStFld("integer-valued?", "kawa.lib.numbers");
+      defProcStFld("finite?", "kawa.lib.numbers");
+      defProcStFld("infinite?", "kawa.lib.numbers");
+      defProcStFld("nan?", "kawa.lib.numbers");
+      defProcStFld("exact-integer-sqrt", "kawa.lib.numbers");
+      // TODO Some of the bindings made below in kawaEnvironment
+      // should be moved here instead.
+
+      r6Environment.setLocked();
       environ = kawaEnvironment;
  
       defSntxStFld("define-private", "kawa.lib.prim_syntax");
@@ -481,7 +544,7 @@ public class Scheme extends LispLanguage
       defProcStFld("logop", "kawa.lib.numbers");
       defProcStFld("bitwise-bit-set?", "kawa.lib.numbers");
       defProcStFld("logbit?", "kawa.lib.numbers",
-                   Compilation.mangleNameIfNeeded("bitwise-bit-set?"));
+                   Language.mangleNameIfNeeded("bitwise-bit-set?"));
       defProcStFld("logtest", "kawa.lib.numbers");
       defProcStFld("bitwise-bit-count", "kawa.lib.numbers");
       defProcStFld("logcount", "kawa.lib.numbers");
@@ -489,7 +552,7 @@ public class Scheme extends LispLanguage
       defProcStFld("bitwise-copy-bit-field", "kawa.lib.numbers");
       defProcStFld("bitwise-bit-field", "kawa.lib.numbers");
       defProcStFld("bit-extract", "kawa.lib.numbers",
-                   Compilation.mangleNameIfNeeded("bitwise-bit-field"));
+                   Language.mangleNameIfNeeded("bitwise-bit-field"));
       defProcStFld("bitwise-length", "kawa.lib.numbers");
       defProcStFld("integer-length", "kawa.lib.numbers", "bitwise$Mnlength");
       defProcStFld("bitwise-first-bit-set", "kawa.lib.numbers");
@@ -558,16 +621,21 @@ public class Scheme extends LispLanguage
       defSntxStFld("syntax-case", "kawa.standard.syntax_case", "syntax_case");
       defSntxStFld("%define-syntax", "kawa.standard.define_syntax",
                    "define_syntax");
+      defSntxStFld("define-rewrite-syntax", "kawa.standard.define_syntax",
+                   "define_rewrite_syntax");
       defSntxStFld("syntax", "kawa.standard.syntax", "syntax");
       defSntxStFld("quasisyntax", "kawa.standard.syntax", "quasiSyntax");
+      defProcStFld("syntax->datum", "kawa.lib.std_syntax");
       defProcStFld("syntax-object->datum", "kawa.lib.std_syntax");
       defProcStFld("datum->syntax-object", "kawa.lib.std_syntax");
+      defProcStFld("datum->syntax", "kawa.lib.std_syntax");
       defProcStFld("syntax->expression", "kawa.lib.prim_syntax");
       defProcStFld("syntax-body->expression", "kawa.lib.prim_syntax");
       defProcStFld("generate-temporaries", "kawa.lib.std_syntax");
       defSntxStFld("with-syntax", "kawa.lib.std_syntax");
       defProcStFld("identifier?", "kawa.lib.std_syntax");
       defProcStFld("free-identifier=?", "kawa.lib.std_syntax");
+      defProcStFld("bound-identifier=?", "kawa.lib.std_syntax");
       defProcStFld("syntax-source", "kawa.lib.std_syntax");
       defProcStFld("syntax-line", "kawa.lib.std_syntax");
       defProcStFld("syntax-column", "kawa.lib.std_syntax");
@@ -588,7 +656,7 @@ public class Scheme extends LispLanguage
       defProcStFld("create-directory", "kawa.lib.files");
       defProcStFld("->pathname", "kawa.lib.files");
       define("port-char-encoding", Boolean.TRUE);
-      define("symbol-read-case", "P");
+      define("symbol-read-case", "");
 
       defProcStFld("system", "kawa.lib.system");
       defProcStFld("make-process", "kawa.lib.system");
@@ -709,90 +777,29 @@ public class Scheme extends LispLanguage
       defProcStFld("array-set!", "gnu.kawa.functions.ArraySet", "arraySet");
       defProcStFld("share-array", "kawa.lib.arrays");
 
-      defProcStFld("s8vector?", "kawa.lib.uniform");
-      defProcStFld("make-s8vector", "kawa.lib.uniform");
-      defProcStFld("s8vector", "kawa.lib.uniform");
-      defProcStFld("s8vector-length", "kawa.lib.uniform");
-      defProcStFld("s8vector-ref", "kawa.lib.uniform");
-      defProcStFld("s8vector-set!", "kawa.lib.uniform");
-      defProcStFld("s8vector->list", "kawa.lib.uniform");
-      defProcStFld("list->s8vector", "kawa.lib.uniform");
-      defProcStFld("u8vector?", "kawa.lib.uniform");
-      defProcStFld("make-u8vector", "kawa.lib.uniform");
-      defProcStFld("u8vector", "kawa.lib.uniform");
-      defProcStFld("u8vector-length", "kawa.lib.uniform");
-      defProcStFld("u8vector-ref", "kawa.lib.uniform");
-      defProcStFld("u8vector-set!", "kawa.lib.uniform");
-      defProcStFld("u8vector->list", "kawa.lib.uniform");
-      defProcStFld("list->u8vector", "kawa.lib.uniform");
+      for (int i = uniformVectorTags.length;  --i >= 0; )
+        {
+          String tag = uniformVectorTags[i];
+          defAliasStFld(tag+"vector", "gnu.kawa.lispexpr.LangObjType",
+                        tag+"vectorType");
+          defProcStFld(tag+"vector?", "kawa.lib.uniform");
+          defProcStFld("make-"+tag+"vector", "kawa.lib.uniform");
+          defProcStFld(tag+"vector-length", "kawa.lib.uniform");
+          defProcStFld(tag+"vector-ref", "kawa.lib.uniform");
+          defProcStFld(tag+"vector-set!", "kawa.lib.uniform");
+          defProcStFld(tag+"vector->list", "kawa.lib.uniform");
+          defProcStFld("list->"+tag+"vector", "kawa.lib.uniform");
+        }
 
-      defProcStFld("s16vector?", "kawa.lib.uniform");
-      defProcStFld("make-s16vector", "kawa.lib.uniform");
-      defProcStFld("s16vector", "kawa.lib.uniform");
-      defProcStFld("s16vector-length", "kawa.lib.uniform");
-      defProcStFld("s16vector-ref", "kawa.lib.uniform");
-      defProcStFld("s16vector-set!", "kawa.lib.uniform");
-      defProcStFld("s16vector->list", "kawa.lib.uniform");
-      defProcStFld("list->s16vector", "kawa.lib.uniform");
-      defProcStFld("u16vector?", "kawa.lib.uniform");
-      defProcStFld("make-u16vector", "kawa.lib.uniform");
-      defProcStFld("u16vector", "kawa.lib.uniform");
-      defProcStFld("u16vector-length", "kawa.lib.uniform");
-      defProcStFld("u16vector-ref", "kawa.lib.uniform");
-      defProcStFld("u16vector-set!", "kawa.lib.uniform");
-      defProcStFld("u16vector->list", "kawa.lib.uniform");
-      defProcStFld("list->u16vector", "kawa.lib.uniform");
-
-      defProcStFld("s32vector?", "kawa.lib.uniform");
-      defProcStFld("make-s32vector", "kawa.lib.uniform");
-      defProcStFld("s32vector", "kawa.lib.uniform");
-      defProcStFld("s32vector-length", "kawa.lib.uniform");
-      defProcStFld("s32vector-ref", "kawa.lib.uniform");
-      defProcStFld("s32vector-set!", "kawa.lib.uniform");
-      defProcStFld("s32vector->list", "kawa.lib.uniform");
-      defProcStFld("list->s32vector", "kawa.lib.uniform");
-      defProcStFld("u32vector?", "kawa.lib.uniform");
-      defProcStFld("make-u32vector", "kawa.lib.uniform");
-      defProcStFld("u32vector", "kawa.lib.uniform");
-      defProcStFld("u32vector-length", "kawa.lib.uniform");
-      defProcStFld("u32vector-ref", "kawa.lib.uniform");
-      defProcStFld("u32vector-set!", "kawa.lib.uniform");
-      defProcStFld("u32vector->list", "kawa.lib.uniform");
-      defProcStFld("list->u32vector", "kawa.lib.uniform");
-
-      defProcStFld("s64vector?", "kawa.lib.uniform");
-      defProcStFld("make-s64vector", "kawa.lib.uniform");
-      defProcStFld("s64vector", "kawa.lib.uniform");
-      defProcStFld("s64vector-length", "kawa.lib.uniform");
-      defProcStFld("s64vector-ref", "kawa.lib.uniform");
-      defProcStFld("s64vector-set!", "kawa.lib.uniform");
-      defProcStFld("s64vector->list", "kawa.lib.uniform");
-      defProcStFld("list->s64vector", "kawa.lib.uniform");
-      defProcStFld("u64vector?", "kawa.lib.uniform");
-      defProcStFld("make-u64vector", "kawa.lib.uniform");
-      defProcStFld("u64vector", "kawa.lib.uniform");
-      defProcStFld("u64vector-length", "kawa.lib.uniform");
-      defProcStFld("u64vector-ref", "kawa.lib.uniform");
-      defProcStFld("u64vector-set!", "kawa.lib.uniform");
-      defProcStFld("u64vector->list", "kawa.lib.uniform");
-      defProcStFld("list->u64vector", "kawa.lib.uniform");
-
-      defProcStFld("f32vector?", "kawa.lib.uniform");
-      defProcStFld("make-f32vector", "kawa.lib.uniform");
-      defProcStFld("f32vector", "kawa.lib.uniform");
-      defProcStFld("f32vector-length", "kawa.lib.uniform");
-      defProcStFld("f32vector-ref", "kawa.lib.uniform");
-      defProcStFld("f32vector-set!", "kawa.lib.uniform");
-      defProcStFld("f32vector->list", "kawa.lib.uniform");
-      defProcStFld("list->f32vector", "kawa.lib.uniform");
-      defProcStFld("f64vector?", "kawa.lib.uniform");
-      defProcStFld("make-f64vector", "kawa.lib.uniform");
-      defProcStFld("f64vector", "kawa.lib.uniform");
-      defProcStFld("f64vector-length", "kawa.lib.uniform");
-      defProcStFld("f64vector-ref", "kawa.lib.uniform");
-      defProcStFld("f64vector-set!", "kawa.lib.uniform");
-      defProcStFld("f64vector->list", "kawa.lib.uniform");
-      defProcStFld("list->f64vector", "kawa.lib.uniform");
+      defAliasStFld("bytevector", "gnu.kawa.lispexpr.LangObjType",
+                    "u8vectorType");
+      defProcStFld("make-bytevector", "kawa.lib.bytevector");
+      defProcStFld("bytevector-length", "kawa.lib.bytevector");
+      defProcStFld("bytevector-u8-ref", "kawa.lib.bytevector");
+      defProcStFld("bytevector-u8-set!", "kawa.lib.bytevector");
+      defProcStFld("bytevector-copy", "kawa.lib.bytevector");
+      defProcStFld("utf8->string", "kawa.lib.bytevector");
+      defProcStFld("string->utf8", "kawa.lib.bytevector");
 
       defSntxStFld("cut", "gnu.kawa.slib.cut");
       defSntxStFld("cute", "gnu.kawa.slib.cut");
@@ -808,6 +815,9 @@ public class Scheme extends LispLanguage
                     "gnu.text.PrettyWriter", "lineLengthLoc");
       defAliasStFld("*print-miser-width*",
                     "gnu.text.PrettyWriter", "miserWidthLoc");
+      defAliasStFld("*print-circle*", "gnu.text.PrettyWriter", "isSharing");
+      defAliasStFld("*print-xml-indent*",
+                    "gnu.xml.XMLPrinter", "indentLoc");
       defAliasStFld("html", "gnu.kawa.xml.XmlNamespace", "HTML");
       defAliasStFld("unit", "kawa.standard.Scheme", "unitNamespace");
 
@@ -816,6 +826,7 @@ public class Scheme extends LispLanguage
       defAliasStFld("URI", "gnu.kawa.lispexpr.LangObjType", "URIType");
       defProcStFld("resolve-uri", "kawa.lib.files");
 
+      defAliasStFld("$bracket-list$", "gnu.kawa.lispexpr.LangObjType", "constVectorType");
       defAliasStFld("vector", "gnu.kawa.lispexpr.LangObjType", "vectorType");
       defAliasStFld("string", "gnu.kawa.lispexpr.LangObjType", "stringType");
       defAliasStFld("list", "gnu.kawa.lispexpr.LangObjType", "listType");
@@ -837,6 +848,8 @@ public class Scheme extends LispLanguage
       defProcStFld("path-fragment", "kawa.lib.files");
       defProcStFld("path-query", "kawa.lib.files");
 
+      defProcStFld("annotation", "gnu.kawa.reflect.MakeAnnotation", "instance");
+
       kawaEnvironment.setLocked();
   }
 
@@ -853,11 +866,23 @@ public class Scheme extends LispLanguage
 
   public String getName()
   {
-    return "Scheme";
+    switch (standardToFollow)
+      {
+      case FOLLOW_R5RS:
+        return "Scheme-r5rs";
+      case FOLLOW_R6RS:
+        return "Scheme-r6rs";
+      case FOLLOW_R7RS:
+        return "Scheme-r7rs";
+      default:
+        return "Scheme";
+      }
   }
 
-  /** Evalutate Scheme expressions from string.
-   * @param string the string constaining Scheme expressions
+  public String getCompilationClass () { return "kawa.standard.SchemeCompilation"; }
+
+  /** Evaluate Scheme expressions from string.
+   * @param string the string containing Scheme expressions
    * @param env the Environment to evaluate the string in
    * @return result of last expression, or Language.voidObject if none. */
   public static Object eval (String string, Environment env)
@@ -865,7 +890,7 @@ public class Scheme extends LispLanguage
     return eval (new CharArrayInPort(string), env);
   }
 
-  /** Evalutate Scheme expressions from stream.
+  /** Evaluate Scheme expressions from stream.
    * @param port the port to read Scheme expressions from
    * @param env the Environment to evaluate the string in
    * @return result of last expression, or Language.voidObject if none. */
@@ -876,7 +901,7 @@ public class Scheme extends LispLanguage
       {
 	LispReader lexer = (LispReader)
 	  Language.getDefaultLanguage().getLexer(port, messages);
-	Object body = ReaderParens.readList(lexer, 0, 1, -1);
+	Object body = ReaderParens.readList(lexer, null, 0, 1, -1, -1);
         if (messages.seenErrors())
           throw new gnu.text.SyntaxException(messages);
 	return Eval.evalBody(body, env, messages);
@@ -907,7 +932,7 @@ public class Scheme extends LispLanguage
       }
   }
 
-  /** Evalutate Scheme expressions from an "S expression."
+  /** Evaluate Scheme expressions from an "S expression."
    * @param sexpr the S expression to evaluate
    * @param env the Environment to evaluate the string in
    * @return result of the expression. */
@@ -932,12 +957,28 @@ public class Scheme extends LispLanguage
   }
 
   public static final AbstractFormat writeFormat = new DisplayFormat(true, 'S');
+  public static final AbstractFormat sharedWriteFormat = new DisplayFormat(true, 'S');
+  // WRITE checks for circular references by default, DISPLAY does not.
+  static { ((DisplayFormat) sharedWriteFormat).checkSharing = true; }
   public static final AbstractFormat displayFormat = new DisplayFormat(false, 'S');
+  
+  @Override
   public AbstractFormat getFormat(boolean readable)
   {
     return readable ? writeFormat : displayFormat;
   }
 
+  @Override
+  public LispReader getLexer(InPort inp, SourceMessages messages)
+  {
+    LispReader reader = super.getLexer(inp, messages);
+    if (reader.getReadCase() == '\0'
+        && standardToFollow == FOLLOW_R5RS)
+      reader.setReadCase('D');
+    return reader;
+  }
+
+  @Override
   public int getNamespaceOf (Declaration decl)
   {
     return FUNCTION_NAMESPACE+VALUE_NAMESPACE;
@@ -991,28 +1032,29 @@ public class Scheme extends LispLanguage
 	types.put ("character", ClassType.make("gnu.text.Char"));
 	types.put ("vector", LangObjType.vectorType);
 	types.put ("string", LangObjType.stringType);
+        types.put ("empty-list", ClassType.make("gnu.lists.EmptyList"));
 	types.put ("list", LangObjType.listType);
 	types.put ("function", ClassType.make("gnu.mapping.Procedure"));
-	types.put ("procedure", ClassType.make("gnu.mapping.Procedure"));
+	types.put ("procedure", LangObjType.procedureType);
 	types.put ("input-port", ClassType.make("gnu.mapping.InPort"));
 	types.put ("output-port", ClassType.make("gnu.mapping.OutPort"));
 	types.put ("string-output-port",
                    ClassType.make("gnu.mapping.CharArrayOutPort"));
+	types.put ("string-input-port",
+                   ClassType.make("gnu.mapping.CharArrayInPort"));
 	types.put ("record", ClassType.make("kawa.lang.Record"));
 	types.put ("type", LangObjType.typeType);
 	types.put ("class-type", LangObjType.typeClassType);
 	types.put ("class", LangObjType.typeClass);
+	types.put ("promise", LangObjType.promiseType);
+
+        for (int i = uniformVectorTags.length;  --i >= 0; )
+          {
+            String tag = uniformVectorTags[i];
+            String cname = "gnu.lists."+tag.toUpperCase()+"Vector";
+            types.put(tag+"vector", ClassType.make(cname));
+          }
 
-        types.put ("s8vector", ClassType.make("gnu.lists.S8Vector"));
-        types.put ("u8vector", ClassType.make("gnu.lists.U8Vector"));
-        types.put ("s16vector", ClassType.make("gnu.lists.S16Vector"));
-        types.put ("u16vector", ClassType.make("gnu.lists.U16Vector"));
-        types.put ("s32vector", ClassType.make("gnu.lists.S32Vector"));
-        types.put ("u32vector", ClassType.make("gnu.lists.U32Vector"));
-        types.put ("s64vector", ClassType.make("gnu.lists.S64Vector"));
-        types.put ("u64vector", ClassType.make("gnu.lists.U64Vector"));
-        types.put ("f32vector", ClassType.make("gnu.lists.F32Vector"));
-        types.put ("f64vector", ClassType.make("gnu.lists.F64Vector"));
         types.put ("document", ClassType.make("gnu.kawa.xml.KDocument"));
         types.put ("readtable", ClassType.make("gnu.kawa.lispexpr.ReadTable"));
       }
@@ -1080,6 +1122,12 @@ public class Scheme extends LispLanguage
   public String formatType (Type type)
   {
     // FIXME synchronize
+    if (type instanceof LazyType)
+      {
+        LazyType ltype = (LazyType) type;
+        return formatType(ltype.getRawType())
+          +'['+formatType(ltype.getValueType())+']';
+      }
     if (typeToStringMap == null)
       {
         typeToStringMap = new HashMap<Type,String>();
@@ -1134,311 +1182,39 @@ public class Scheme extends LispLanguage
   public static final Namespace unitNamespace =
     Namespace.valueOf("http://kawa.gnu.org/unit", "unit");
 
-  /** If a symbol is lexically unbound, look for a default binding.
-   * If the symbol is the name of an existing Java type/class,
-   * return that Class.
-   * Handles both with and without (semi-deprecated) angle-brackets:
-   * {@code <java.lang.Integer>} and {@code java.lang.Integer}.
-   * Also handles arrays, such as {@code java.lang.String[]}.
-   */
-  public Expression checkDefaultBinding (Symbol symbol, Translator tr)
-  {
-    Namespace namespace = symbol.getNamespace();
-    String local = symbol.getLocalPart();
-    if (namespace instanceof XmlNamespace)
-      return QuoteExp.getInstance(((XmlNamespace) namespace).get(local));
-    if (namespace.getName() == unitNamespace.getName())
-      {
-        Object val = Unit.lookup(local);
-        if (val != null)
-          return QuoteExp.getInstance(val);
-      }
-    String name = symbol.toString();
-    int len = name.length();
-    if (len == 0)
-      return null;
-    if (len > 1 && name.charAt(len-1) == '?')
-      {
-        int llen = local.length();
-        if (llen > 1)
-          {
-            String tlocal = local.substring(0, llen-1).intern();
-            Symbol tsymbol = namespace.getSymbol(tlocal);
-            Expression texp = tr.rewrite(tsymbol, false);
-            if (texp instanceof ReferenceExp)
-              {
-                Declaration decl = ((ReferenceExp) texp).getBinding();
-                if (decl == null || decl.getFlag(Declaration.IS_UNKNOWN))
-                  texp = null;
-              }
-            else if (! (texp instanceof QuoteExp))
-              texp = null;
-            if (texp != null)
-              {
-                LambdaExp lexp = new LambdaExp(1);
-                lexp.setSymbol(symbol);
-                Declaration param = lexp.addDeclaration((Object) null);
-                lexp.body = new ApplyExp(instanceOf,
-                                         new Expression[] {
-                                           new ReferenceExp(param), texp, });
-                return lexp;
-              }
-          }
-      }
-    char ch0 = name.charAt(0);
-
-    // Look for quantities.
-    if (ch0 == '-' || ch0 == '+' || Character.digit(ch0, 10) >= 0)
-      {
-        // 1: initial + or -1 seen.
-        // 2: digits seen
-        // 3: '.' seen
-        // 4: fraction seen
-        // 5: [eE][=+]?[0-9]+ seen
-        int state = 0;
-        int i = 0;
-        for (;  i < len;  i++)
-          {
-            char ch = name.charAt(i);
-            if (Character.digit(ch, 10) >= 0)
-              state = state < 3 ? 2 : state < 5 ? 4 : 5;
-            else if ((ch == '+' || ch == '-') && state == 0)
-              state = 1;
-            else if (ch == '.' && state < 3)
-              state = 3;
-            else if ((ch == 'e' || ch == 'E') && (state == 2 || state == 4)
-                     && i + 1 < len)
-              {
-                int j = i+1;
-                char next = name.charAt(j);
-                if ((next == '-' || next == '+') && ++j < len)
-                  next = name.charAt(j);
-                if (Character.digit(next, 10) < 0)
-                  break;
-                state = 5;
-                i = j+1;
-              }
-            else
-              break;
-          }
-      tryQuantity:
-        if (i < len && state > 1)
-          {
-            DFloNum num = new DFloNum(name.substring(0,i));
-            boolean div = false;
-            Vector vec = new Vector();
-            for (; i < len ;)
-              {
-                char ch = name.charAt(i++);
-                if (ch == '*')
-                  {
-                    if (i == len) break tryQuantity;
-                    ch = name.charAt(i++);
-                  }
-                else if (ch == '/')
-                  {
-                    if (i == len || div) break tryQuantity; 
-                    div = true;
-                    ch = name.charAt(i++);
-                  }
-                int unitStart = i-1;
-                int unitEnd;
-                for (;;)
-                  {
-                    if (! Character.isLetter(ch))
-                      {
-                        unitEnd = i - 1;
-                        if (unitEnd == unitStart)
-                          break tryQuantity;
-                        break;
-                      }
-                    if (i == len)
-                      {
-                        unitEnd = i;
-                        ch = '1';
-                        break;
-                      }
-                    ch = name.charAt(i++);
-                  }
-                vec.addElement(name.substring(unitStart, unitEnd));
-                boolean expRequired = false;
-                if (ch == '^')
-                  {
-                    expRequired = true;
-                    if (i == len) break tryQuantity; 
-                    ch = name.charAt(i++);
-                  }
-                boolean neg = div;
-                if (ch == '+')
-                  {
-                    expRequired = true;
-                    if (i == len) break tryQuantity; 
-                    ch = name.charAt(i++);
-                  }
-                else if (ch == '-')
-                  {
-                    expRequired = true;
-                    if (i == len) break tryQuantity; 
-                    ch = name.charAt(i++);
-                    neg = ! neg;
-                  }
-                int nexp = 0;
-                int exp = 0;
-                for (;;)
-                  {
-                    int dig = Character.digit(ch, 10);
-                    if (dig <= 0)
-                      {
-                        i--;
-                        break;
-                      }
-                    exp = 10 * exp + dig;
-                    nexp++;
-                    if (i == len)
-                      break;
-                    ch = name.charAt(i++);
-                  }
-                if (nexp == 0)
-                  {
-                    exp = 1;
-                    if (expRequired)
-                      break tryQuantity;
-                  }
-                if (neg)
-                  exp = - exp;
-                vec.addElement(IntNum.make(exp));
-              }
-            if (i == len)
-              {
-                int nunits = vec.size() >> 1;
-                Expression[] units = new Expression[nunits];
-                for (i = 0;  i < nunits;  i++)
-                  {
-                    String uname = (String) vec.elementAt(2*i);
-                    Symbol usym = unitNamespace.getSymbol(uname.intern());
-                    Expression uref = tr.rewrite(usym);
-                    IntNum uexp = (IntNum) vec.elementAt(2*i+1);
-                    if (uexp.longValue() != 1)
-                      uref = new ApplyExp(expt.expt,
-                                          new Expression[] { uref , QuoteExp.getInstance(uexp) });
-                    units[i] = uref;
-                  }
-                Expression unit;
-                if (nunits == 1)
-                  unit = units[0];
-                else
-                  unit = new ApplyExp(MultiplyOp.$St, units);
-                return new ApplyExp(MultiplyOp.$St,
-                                    new Expression[] { QuoteExp.getInstance(num),
-                                                       unit });
-              }
-          }
-      }
-
-    boolean sawAngle;
-    if (len > 2 && ch0 == '<' && name.charAt(len-1) == '>')
-      {
-        name = name.substring(1, len-1);
-        len -= 2;
-        sawAngle = true;
-      }
-    else
-      sawAngle = false;
-    int rank = 0;
-    while (len > 2 && name.charAt(len-2) == '[' && name.charAt(len-1) == ']')
-      {
-        len -= 2;
-        rank++;
-      }
-
-    String cname = name;
-    if (rank != 0)
-      cname = name.substring(0, len);
-    try
-      { 
-        Class clas;
-        Type type = getNamedType(cname);
-        if (rank > 0 && (! sawAngle || type == null))
-          {
-            Symbol tsymbol = namespace.getSymbol(cname.intern());
-            Expression texp = tr.rewrite(tsymbol, false);
-            texp = InlineCalls.inlineCalls(texp, tr);
-            if (! (texp instanceof ErrorExp))
-              type = tr.getLanguage().getTypeFor(texp);
-          }
-        if (type != null)
-          {
-            // Somewhat inconsistent: Types named by getNamedType are Type,
-            // while standard type/classes are Class.  FIXME.
-            while (--rank >= 0)
-              type = gnu.bytecode.ArrayType.make(type);
-            return QuoteExp.getInstance(type);
-          }
-        else
-          {
-            type = Type.lookupType(cname);
-            if (type instanceof gnu.bytecode.PrimType)
-              clas = type.getReflectClass();
-            else
-              {
-                if (cname.indexOf('.') < 0)
-                  cname = (tr.classPrefix
-                           + Compilation.mangleNameIfNeeded(cname));
-                clas = ClassType.getContextClass(cname);
-              }
-          }
-        if (clas != null)
-          {
-            if (rank > 0)
-              {
-                type = Type.make(clas);
-                while (--rank >= 0)
-                  type = gnu.bytecode.ArrayType.make(type);
-                clas = type.getReflectClass();
-              }
-            return QuoteExp.getInstance(clas);
-          }
-      }
-    catch (ClassNotFoundException ex)
-      {
-        Package pack = gnu.bytecode.ArrayClassLoader.getContextPackage(name);
-        if (pack != null)
-          return QuoteExp.getInstance(pack);
-      }
-    catch (Throwable ex)
-      {
-      }
-    return null;
-  }
-
-  public Expression makeApply (Expression func, Expression[] args)
-  {
-    Expression[] exps = new Expression[args.length+1];
-    exps[0] = func;
-    System.arraycopy(args, 0, exps, 1, args.length);
-    return new ApplyExp(new ReferenceExp(applyFieldDecl), exps);
-  }
-
   public Symbol asSymbol (String ident)
   {
     return Namespace.EmptyNamespace.getSymbol(ident);
   }
 
+  /** Should the values of body/block be appended as multiple values?
+   * Otherwise, just return the result of the final expression.
+   */
+  public boolean appendBodyValues () { return false; }
+
   public ReadTable createReadTable ()
   {
     ReadTable tab = ReadTable.createInitial();
-    tab.postfixLookupOperator = ':';
+    int std =  standardToFollow;
     ReaderDispatch dispatchTable = (ReaderDispatch) tab.lookup('#');
-    dispatchTable.set('\'', new ReaderQuote(asSymbol("syntax")));
-    dispatchTable.set('`', new ReaderQuote(asSymbol("quasisyntax")));
-    dispatchTable.set(',', ReaderDispatchMisc.getInstance());
+    ReaderDispatchSyntaxQuote sentry = new ReaderDispatchSyntaxQuote();
+    dispatchTable.set('\'', sentry);
+    dispatchTable.set('`', sentry);
+    dispatchTable.set(',', sentry);
     tab.putReaderCtorFld("path", "gnu.kawa.lispexpr.LangObjType", "pathType");
     tab.putReaderCtorFld("filepath", "gnu.kawa.lispexpr.LangObjType", "filepathType");
     tab.putReaderCtorFld("URI", "gnu.kawa.lispexpr.LangObjType", "URIType");
     tab.putReaderCtor("symbol", ClassType.make("gnu.mapping.Symbol"));
     tab.putReaderCtor("namespace", ClassType.make("gnu.mapping.Namespace"));
     tab.putReaderCtorFld("duration", "kawa.lib.numbers", "duration");
-    tab.setFinalColonIsKeyword(true);
+    if (std == FOLLOW_R5RS || std == FOLLOW_R6RS || std == FOLLOW_R7RS)
+      {
+      }
+    else
+      {
+        tab.postfixLookupOperator = ':';
+        tab.setFinalColonIsKeyword(true);
+      }
     return tab;
   }
 
