@@ -29,7 +29,9 @@ public class let extends Syntax
       return tr.syntaxError("bindings not a proper list");
       
     Expression[] inits = new Expression[decl_count];
-    LetExp let = new LetExp (inits);
+    Declaration[] decls = new Declaration[decl_count];
+    // Used to check for duplicate definitions.
+    SimpleEnvironment dupenv = new SimpleEnvironment();
     Stack renamedAliases = null;
     int renamedAliasesCount = 0;
     SyntaxForm syntaxRest = null;
@@ -66,8 +68,14 @@ public class let extends Syntax
         name = tr.namespaceResolve(name);
 	if (! (name instanceof Symbol))
 	  return tr.syntaxError("variable "+name+" in let binding is not a symbol: "+obj);
-
-	Declaration decl = let.addDeclaration(name);
+	Declaration decl = new Declaration(name);
+        Translator.setLine(decl, binding);
+        Symbol sym = (Symbol) name;
+        Object old = dupenv.get(sym, templateScope, null);
+        if (old != null)
+          ScopeExp.duplicateDeclarationError((Declaration) old, decl, tr);
+        dupenv.put(sym, templateScope, decl);
+        decls[i] = decl;
         decl.setFlag(Declaration.IS_SINGLE_VALUE);
 	if (templateScope != null)
 	  {
@@ -112,25 +120,26 @@ public class let extends Syntax
 	  }
 	else if (binding_cdr instanceof Pair)
 	  {
-	    decl.setType(tr.exp2Type(binding));
+            decl.setType(tr.exp2Type(binding, null, syntax));
 	    decl.setFlag(Declaration.TYPE_SPECIFIED);
 	    init = (Pair) binding_cdr;
 	  }
 	else
 	  return tr.syntaxError("let binding for '"+name+"' is improper list");
-	inits[i] = tr.rewrite_car (init, syntax);
+        inits[i] = tr.rewrite_car (init, syntax);
 	if (init.getCdr() != LList.Empty)
 	  return tr.syntaxError("junk after declaration of "+name);
-	decl.noteValue (inits[i]);
 	bindings = bind_pair.getCdr();
       }
 
     for (int i = renamedAliasesCount;  --i >= 0; )
       tr.pushRenamedAlias((Declaration) renamedAliases.pop());
 
-    tr.push(let);
-    let.body = tr.rewrite_body(body);
-    tr.pop(let);
+    tr.letStart();
+    for (int i = 0; i < decl_count;  i++)
+      tr.letVariable(decls[i], inits[i]);
+    tr.letEnter();
+    LetExp let = tr.letDone(tr.rewrite_body(body));
     tr.popRenamedAlias(renamedAliasesCount);
 
     return let;
