@@ -81,6 +81,7 @@ public class require extends Syntax
     map("srfi-2", SLIB_PREFIX + "srfi2");
     map("and-let*", SLIB_PREFIX + "srfi2");
     map("srfi-13", SLIB_PREFIX + "srfi13");
+    map("srfi-14", SLIB_PREFIX + "srfi14");
     map("string-lib", SLIB_PREFIX + "srfi13");
     map("srfi-34", SLIB_PREFIX + "srfi34");
     map("srfi-35", SLIB_PREFIX + "conditions");
@@ -88,6 +89,11 @@ public class require extends Syntax
     map("conditions", SLIB_PREFIX + "conditions");
     map("srfi-37", SLIB_PREFIX + "srfi37");
     map("args-fold", SLIB_PREFIX + "srfi37");
+    map("srfi-41", SLIB_PREFIX + "Streams");
+    map("srfi-41-streams", SLIB_PREFIX + "Streams");
+    map("srfi-41-streams-type", SLIB_PREFIX + "StreamsType");
+    map("srfi-41-streams-primitive", SLIB_PREFIX + "StreamsPrimitive");
+    map("srfi-41-streams-derived", SLIB_PREFIX + "StreamsDerived");
     map("srfi-64", SLIB_PREFIX + "testing");
     map("testing", SLIB_PREFIX + "testing");
     map("srfi-69", SLIB_PREFIX + "srfi69");
@@ -100,6 +106,7 @@ public class require extends Syntax
     map("gui", SLIB_PREFIX + "gui");
     map("swing-gui", SLIB_PREFIX + "swing");
     map("android-defs", "gnu.kawa.android.defs");
+    map("javafx-defs", "gnu.kawa.javafx.defs");
     map("syntax-utils", SLIB_PREFIX + "syntaxutils");
   }
 
@@ -167,7 +174,12 @@ public class require extends Syntax
     else if (name instanceof Symbol && ! tr.selfEvaluatingSymbol(name))
       {
         type = tr.getLanguage().getTypeFor(tr.rewrite(name, false));
-        if (type instanceof ClassType && args.getCdr() instanceof Pair)
+        String cname = null;
+        if (type instanceof ClassType)
+          cname = type.getName();
+        else if (type == null && name instanceof SimpleSymbol)
+          cname = ((SimpleSymbol) name).getName();
+        if (cname != null && args.getCdr() instanceof Pair)
           {
             name = ((Pair) args.getCdr()).getCar();
             if
@@ -184,7 +196,7 @@ public class require extends Syntax
                     tr.error('e', "malformed URL: "+sourceName);
                     return false;
                   }
-                return importDefinitions(type.getName(), info, null,
+                return importDefinitions(cname, info, null,
                                          forms, defs, tr);
               }
 	  }
@@ -456,7 +468,8 @@ public class require extends Syntax
             ReferenceExp aref = (ReferenceExp) adecl.getValue();
             Declaration xdecl = ((ReferenceExp) fval).getBinding();
             aref.setBinding(xdecl);
-            if (xdecl.needsContext())
+            // xdecl can be null on an error.
+            if (xdecl != null && xdecl.needsContext())
               {
                 String iname
                   = (xdecl.field.getDeclaringClass().getName().replace('.', '$')
