@@ -6,7 +6,7 @@
     ((import-class fqcn)
      (let* ((cls :: java.lang.Class (eval (syntax fqcn)))
 	    (name (string->symbol (java.lang.Class:getSimpleName cls))))
-       #`(define-alias ,(datum->syntax-object form name) fqcn)))))
+       #`(define-alias #,(datum->syntax-object form name) fqcn)))))
 
 (import-class java.util.Date)
 
@@ -48,6 +48,9 @@
   ((f (y :: <int>)) :: <int>
    (if (equal? hyphenated-field? "yes") (+ (g) b y) 999))
 
+  ((withVarArg a #!rest b)
+   (format "a:~a b:~b" a b))
+
   ((asText (i ::int))::string (java.lang.Integer:toString i))
   ((asText (t ::java.lang.CharSequence))::string (t:toString))
   ((test-asText)
