@@ -6,7 +6,8 @@
 
 (define-class <ClsE> (<ClsC> <ClsD>)
   (e :: <int> init-form: 39)
-  ((f (y :: <int>)) :: <int> (+ 100 xx  e y)))
+  ((f (y :: int)) ::int  throws: (java.lang.Exception)
+   (+ 100 xx  e y)))
 
 (define-class-using-syntax-rules <SimpleC> <SimpleB>
   (d :: <int> init-form: 23 init-keyword: d:)
