@@ -1,4 +1,4 @@
-(test-init "Miscellaneous" 204)
+(test-init "Miscellaneous" 210)
 
 ;;; DSSSL spec example 11
 (test '(3 4 5 6) (lambda x x) 3 4 5 6)
@@ -63,6 +63,12 @@
 (test foo: string->keyword "foo")
 (test ||: string->keyword "")
 
+;; Test keyword parameter with primitive type.
+(define (fun-with-keys1 #!key (code ::int 400) (message ::string "brrp"))
+   (format "code: ~a message: ~a." code message))
+(test "code: 400 message: brrp." 'test-fun-with-keys1 (fun-with-keys1))
+(test "code: 200 message: brrp." 'test-fun-with-keys1 (fun-with-keys1 code: 200))
+
 (test "Hello" symbol->string 'H\x65;llo)
 
 ;;; DSSSL spec example 45
@@ -137,7 +143,7 @@
       (string-set! cr-test-string i #\Linefeed)))
 (call-with-input-string
  cr-test-string
- (lambda (iport)
+ (lambda (iport ::input-port)
    (iport:setConvertCR #t)
    (test 1 input-port-column-number iport)
    (test 1 input-port-line-number iport)
@@ -960,3 +966,67 @@
      (loop))
    (e java.lang.Exception 
       (e:printStackTrace))))
+
+;; Savannah bug #32656: ArrayIndexOutOfBoundsException in mergeLocalType
+(test 2 'savannah-32656
+      (letrec ((f (lambda (x)
+                    (case x
+                      ((0) (f x))
+                      ((1) (g x))
+                      ((2) (h x)))))
+               (g (lambda (x)
+                    (case x
+                      ((0) (f x))
+                      ((1) (g x))
+                      ((2) (h x)))))
+               (h (lambda (x)
+                    (case x
+                      ((0) (f x))
+                      ((1) (g x))
+                      ((2) ;(h x)
+                       x)))))
+        (f 2)))
+
+;; Savannah bug #32657: Verification error with JDK7
+(begin
+  (define (foo-savannah-32657) ()
+    (let ((x (bar-savannah-32657))
+          (fail (lambda () (error "fail"))))
+      (if (instance? x <pair>)
+          (let ((y :: <pair> x))
+            (let ((z (y:getCar)))
+              (if (eq? (y:getCdr) '())
+                  z
+                  (fail))))
+          (fail))))
+  (define (bar-savannah-32657) ::<list>
+    (list 1))
+  (test 1 'savannah-32657 (foo-savannah-32657)))
+
+;; Savannah bug ##32678: set! and endless loop
+(define (foo-savannah-32678 x)
+  (let ((fail 0)
+	(result #!null))
+    (if (instance? x pair)
+	(set! result (do () (#f)))
+	(set! fail -1))
+    (if (= fail 0) 
+	result 
+	#f)))
+(test #f foo-savannah-32678 123)
+
+;; Testcase simplified from slime/config/swank-kawa.scm
+(define-syntax mif
+  (syntax-rules ()
+    ((mif ((p . ps) value) then)
+     (let ((fail (lambda () (error "mlet failed")))
+           (tmp value))
+       (if (instance? tmp <pair>)
+           (let* ((tmp :: <pair> tmp))
+             then)
+           (fail))))))
+(define (dispatch-events)
+  (let ((tmp0 '(a b c)))
+    (mif ((c . event) tmp0)
+            1234)))
+(test 1234 dispatch-events)
