@@ -1,4 +1,4 @@
-(test-init "Objects" 135)
+(test-init "Objects" 138)
 
 ;; Force procedure to be applied without being inlined:
 (define-syntax force-eval
@@ -193,6 +193,11 @@
 	(list v2 (get3-mod0-v2))))
 (test 25 'test-mod2-v5  mod2-v5)
 
+;; Test for Savannah bug #34004: Nullpointer exception in compiler
+(test 1 check-thunk)
+
+(test '(1 2) 'counter-test-result counter-test-result)
+
 (define ts1 (make <MyTimestamp> 10 1))
 (define ts2 (make <MyTimestamp> 10 2))
 (test #t < (my-compare ts1 ts2) 0)
@@ -247,6 +252,7 @@
 (test #t slot-ref (as <SimpleA> obj1) 'happy)
 (set! (field obj1 'happy) #f)
 (test #f field obj1 'happy)
+(test "a:5 b:(10 15)" 'with-var-arg (obj1:withVarArg 5 10 15))
 
 (define obj2 (make <SimpleB>))
 (test 4 field obj2 'a)
